@@ -1,5 +1,6 @@
 /*
- * Copyright (C) 1996-2000 Michael R. Elkins <me@cs.hmc.edu>
+ * Copyright (C) 1996-2000 Michael R. Elkins <me@mutt.org>
+ * Copyright (C) 2000 Thomas Roessler <roessler@does-not-exist.org>
  * 
  *     This program is free software; you can redistribute it and/or modify
  *     it under the terms of the GNU General Public License as published by
@@ -25,6 +26,11 @@
 #include "copy.h"
 #include "mx.h"
 #include "pager.h"
+#include "mutt_crypt.h"
+#include "mutt_idna.h"
+#include <sys/types.h>
+#include <sys/stat.h>
+#include <fcntl.h>
 
 #ifdef USE_IMAP
 #include "imap.h"
@@ -34,22 +40,6 @@
 #include "buffy.h"
 #endif
 
-
-
-#ifdef HAVE_PGP
-#include "pgp.h"
-#endif
-
-
-
-
-
-
-
-
-
-
-
 #include <errno.h>
 #include <unistd.h>
 #include <stdlib.h>
@@ -69,26 +59,30 @@ int mutt_display_message (HEADER *cur)
   char tempfile[_POSIX_PATH_MAX], buf[LONG_STRING];
   int rc = 0, builtin = 0;
   int cmflags = M_CM_DECODE | M_CM_DISPLAY | M_CM_CHARCONV;
-  FILE *fpout;
+  FILE *fpout = NULL;
+  FILE *fpfilterout = NULL;
+  pid_t filterpid = -1;
+  int res;
 
   snprintf (buf, sizeof (buf), "%s/%s", TYPE (cur->content),
 	    cur->content->subtype);
 
   mutt_parse_mime_message (Context, cur);
+  mutt_message_hook (Context, cur, M_MESSAGEHOOK);
 
-
-#ifdef HAVE_PGP
-  /* see if PGP is needed for this message.  if so, we should exit curses */
-  if (cur->pgp)
+  /* see if crytpo is needed for this message.  if so, we should exit curses */
+  if (WithCrypto && cur->security)
   {
-    if (cur->pgp & PGPENCRYPT)
+    if (cur->security & ENCRYPT)
     {
-      if (!pgp_valid_passphrase ())
+      if (cur->security & APPLICATION_SMIME)
+	crypt_smime_getkeys (cur->env);
+      if(!crypt_valid_passphrase(cur->security))
 	return 0;
 
       cmflags |= M_CM_VERIFY;
     }
-    else if (cur->pgp & PGPSIGN)
+    else if (cur->security & SIGN)
     {
       /* find out whether or not the verify signature */
       if (query_quadoption (OPT_VERIFYSIG, _("Verify PGP signature?")) == M_YES)
@@ -98,15 +92,20 @@ int mutt_display_message (HEADER *cur)
     }
   }
   
-  if ((cmflags & M_CM_VERIFY) || (cur->pgp & PGPENCRYPT))
+  if (cmflags & M_CM_VERIFY || cur->security & ENCRYPT)
   {
-    if (cur->env->from)
-      pgp_invoke_getkeys (cur->env->from);
+    if (cur->security & APPLICATION_PGP)
+    {
+      if (cur->env->from)
+        crypt_pgp_invoke_getkeys (cur->env->from);
+      
+      crypt_invoke_message (APPLICATION_PGP);
+    }
 
-    mutt_message _("Invoking PGP...");
+    if (cur->security & APPLICATION_SMIME)
+      crypt_invoke_message (APPLICATION_SMIME);
   }
 
-#endif
 
   mutt_mktemp (tempfile);
   if ((fpout = safe_fopen (tempfile, "w")) == NULL)
@@ -115,6 +114,22 @@ int mutt_display_message (HEADER *cur)
     return (0);
   }
 
+  if (DisplayFilter && *DisplayFilter) 
+  {
+    fpfilterout = fpout;
+    fpout = NULL;
+    /* mutt_endwin (NULL); */
+    filterpid = mutt_create_filter_fd (DisplayFilter, &fpout, NULL, NULL,
+				       -1, fileno(fpfilterout), -1);
+    if (filterpid < 0)
+    {
+      mutt_error (_("Cannot create display filter"));
+      safe_fclose (&fpfilterout);
+      unlink (tempfile);
+      return 0;
+    }
+  }
+
   if (!Pager || mutt_strcmp (Pager, "builtin") == 0)
     builtin = 1;
   else
@@ -124,34 +139,66 @@ int mutt_display_message (HEADER *cur)
     fputs ("\n\n", fpout);
   }
 
-  if (mutt_copy_message (fpout, Context, cur, cmflags,
-			 (option (OPTWEED) ? (CH_WEED | CH_REORDER) : 0) | CH_DECODE | CH_FROM) == -1)
+  res = mutt_copy_message (fpout, Context, cur, cmflags,
+       	(option (OPTWEED) ? (CH_WEED | CH_REORDER) : 0) | CH_DECODE | CH_FROM);
+  if ((safe_fclose (&fpout) != 0 && errno != EPIPE) || res == -1)
   {
-    fclose (fpout);
-    unlink (tempfile);
+    mutt_error (_("Could not copy message"));
+    if (fpfilterout != NULL)
+    {
+      mutt_wait_filter (filterpid);
+      safe_fclose (&fpfilterout);
+    }
+    mutt_unlink (tempfile);
     return 0;
   }
 
-  if (fclose (fpout) != 0 && errno != EPIPE)
-  {
-    mutt_perror ("fclose");
-    mutt_unlink (tempfile);
-    return (0);
-  }
+  if (fpfilterout != NULL && mutt_wait_filter (filterpid) != 0)
+    mutt_any_key_to_continue (NULL);
 
-#ifdef HAVE_PGP
-  /* update PGP information for this message */
-  cur->pgp |= pgp_query (cur->content);
-#endif
+  safe_fclose (&fpfilterout);	/* XXX - check result? */
+
+  
+  if (WithCrypto)
+  {
+    /* update crypto information for this message */
+    cur->security |= crypt_query (cur->content);
+  
+    /* Remove color cache for this message, in case there
+       are color patterns for both ~g and ~V */
+    cur->pair = 0;
+  }
 
   if (builtin)
   {
     pager_t info;
 
-#ifdef HAVE_PGP
-    if (cur->pgp & PGPGOODSIGN)
-      mutt_message _("PGP signature successfully verified.");
-#endif
+    if (WithCrypto 
+        && (cur->security & APPLICATION_SMIME) && (cmflags & M_CM_VERIFY))
+    {
+      if (cur->security & GOODSIGN)
+      {
+	if (!crypt_smime_verify_sender(cur))
+	  mutt_message ( _("S/MIME signature successfully verified."));
+	else
+	  mutt_error ( _("S/MIME certificate owner does not match sender."));
+      }
+      else if (cur->security & PARTSIGN)
+	mutt_message (_("Warning: Part of this message has not been signed."));
+      else if (cur->security & SIGN || cur->security & BADSIGN)
+	mutt_error ( _("S/MIME signature could NOT be verified."));
+    }
+
+    if (WithCrypto 
+        && (cur->security & APPLICATION_PGP) && (cmflags & M_CM_VERIFY))
+    {
+      if (cur->security & GOODSIGN)
+	mutt_message (_("PGP signature successfully verified."));
+      else if (cur->security & PARTSIGN)
+	mutt_message (_("Warning: Part of this message has not been signed."));
+      else
+	mutt_message (_("PGP signature could NOT be verified."));
+    }
 
     /* Invoke the builtin pager */
     memset (&info, 0, sizeof (pager_t));
@@ -163,7 +210,7 @@ int mutt_display_message (HEADER *cur)
   {
     int r;
 
-    endwin ();
+    mutt_endwin (NULL);
     snprintf (buf, sizeof (buf), "%s %s", NONULL(Pager), tempfile);
     if ((r = mutt_system (buf)) == -1)
       mutt_error (_("Error running \"%s\"!"), buf);
@@ -188,6 +235,7 @@ void ci_bounce_message (HEADER *h, int *redraw)
   char prompt[SHORT_STRING];
   char buf[HUGE_STRING] = { 0 };
   ADDRESS *adr = NULL;
+  char *err = NULL;
   int rc;
 
   if(h)
@@ -214,25 +262,49 @@ void ci_bounce_message (HEADER *h, int *redraw)
 
   adr = mutt_expand_aliases (adr);
 
-  buf[0] = 0;
-  rfc822_write_address (buf, sizeof (buf), adr);
-
-  snprintf (prompt, (COLS > sizeof(prompt) ? sizeof(prompt) : COLS) - 13, 
-           (h ? _("Bounce message to %s") : _("Bounce messages to %s")), buf);
-  strcat(prompt, "...?");
-  if (mutt_yesorno (prompt, 1) != 1)
+  if (mutt_addrlist_to_idna (adr, &err) < 0)
   {
+    mutt_error (_("Bad IDN: '%s'"), err);
+    FREE (&err);
     rfc822_free_address (&adr);
-    CLEARLINE (LINES-1);
     return;
   }
 
-  mutt_bounce_message (NULL, h, adr);
+  buf[0] = 0;
+  rfc822_write_address (buf, sizeof (buf), adr, 1);
+
+#define extra_space (15 + 7 + 2)
+  snprintf (prompt, sizeof (prompt),
+           (h ? _("Bounce message to %s") : _("Bounce messages to %s")), buf);
+
+  if (mutt_strwidth (prompt) > COLS - extra_space)
+  {
+    mutt_format_string (prompt, sizeof (prompt),
+			0, COLS-extra_space, 0, 0,
+			prompt, sizeof (prompt), 0);
+    strncat (prompt, "...?", sizeof (prompt));
+  }
+  else
+    strncat (prompt, "?", sizeof (prompt));
+
+  if (query_quadoption (OPT_BOUNCE, prompt) != M_YES)
+  {
+    rfc822_free_address (&adr);
+    CLEARLINE (LINES - 1);
+    mutt_message (h ? _("Message not bounced.") : _("Messages not bounced."));
+    return;
+  }
+
+  CLEARLINE (LINES - 1);
+  
+  rc = mutt_bounce_message (NULL, h, adr);
   rfc822_free_address (&adr);
-  mutt_message (h ? _("Message bounced.") : _("Messages bounced."));
+  /* If no error, or background, display message. */
+  if ((rc == 0) || (rc == S_BKG))
+    mutt_message (h ? _("Message bounced.") : _("Messages bounced."));
 }
 
-static void pipe_print_set_flags (int decode, int *cmflags, int *chflags)
+static void pipe_set_flags (int decode, int print, int *cmflags, int *chflags)
 {
   if (decode)
   {
@@ -245,95 +317,108 @@ static void pipe_print_set_flags (int decode, int *cmflags, int *chflags)
       *cmflags |= M_CM_WEED;
     }
   }
+  
+  if (print)
+    *cmflags |= M_CM_PRINTING;
+  
 }
 
-void mutt_pipe_message_to_state (HEADER *h, STATE *s)
+void pipe_msg (HEADER *h, FILE *fp, int decode, int print)
 {
   int cmflags = 0;
   int chflags = CH_FROM;
   
-  pipe_print_set_flags (option (OPTPIPEDECODE), &cmflags, &chflags);
-  
-  if (option (OPTPIPEDECODE))
+  pipe_set_flags (decode, print, &cmflags, &chflags);
+
+  if (WithCrypto && decode && h->security & ENCRYPT)
+  {
+    if(!crypt_valid_passphrase(h->security))
+      return;
+    endwin ();
+  }
+
+  if (decode)
     mutt_parse_mime_message (Context, h);
 
-  mutt_copy_message (s->fpout, Context, h,
-		     cmflags, chflags);
+  mutt_copy_message (fp, Context, h, cmflags, chflags);
 }
 
-int mutt_pipe_message (HEADER *h)
+
+/* the following code is shared between printing and piping */
+
+static int _mutt_pipe_message (HEADER *h, char *cmd,
+			       int decode,
+			       int print,
+			       int split,
+			       char *sep)
 {
-  STATE s;
-  char buffer[LONG_STRING];
-  int i, rc = 0; 
+  
+  int i, rc = 0;
   pid_t thepid;
+  FILE *fpout;
+  
+/*   mutt_endwin (NULL); 
 
-  buffer[0] = 0;
-  if (mutt_get_field (_("Pipe to command: "), buffer, sizeof (buffer), M_CMD)
-      != 0 || !buffer[0])
-    return 0;
-  mutt_expand_path (buffer, sizeof (buffer));
-
-  memset (&s, 0, sizeof (s));
-
-  endwin ();
+     is this really needed here ? 
+     it makes the screen flicker on pgp and s/mime messages,
+     before asking for a passphrase...
+                                     Oliver Ehli */
   if (h)
   {
 
+    mutt_message_hook (Context, h, M_MESSAGEHOOK);
 
-
-#ifdef HAVE_PGP
-    if (option (OPTPIPEDECODE))
+    if (WithCrypto && decode)
     {
       mutt_parse_mime_message (Context, h);
-      if(h->pgp & PGPENCRYPT && !pgp_valid_passphrase())
+      if(h->security & ENCRYPT && !crypt_valid_passphrase(h->security))
 	return 1;
     }
-    endwin ();
-#endif
+    mutt_endwin (NULL);
 
-
-
-    thepid = mutt_create_filter (buffer, &s.fpout, NULL, NULL);
-    mutt_pipe_message_to_state (h, &s);
-    fclose (s.fpout);
+    if ((thepid = mutt_create_filter (cmd, &fpout, NULL, NULL)) < 0)
+    {
+      mutt_perror _("Can't create filter process");
+      return 1;
+    }
+      
+    pipe_msg (h, fpout, decode, print);
+    fclose (fpout);
     rc = mutt_wait_filter (thepid);
   }
   else
   { /* handle tagged messages */
 
-
-
-#ifdef HAVE_PGP
-
-    if(option(OPTPIPEDECODE))
+    if (WithCrypto && decode)
     {
       for (i = 0; i < Context->vcount; i++)
 	if(Context->hdrs[Context->v2r[i]]->tagged)
 	{
+	  mutt_message_hook (Context, Context->hdrs[Context->v2r[i]], M_MESSAGEHOOK);
 	  mutt_parse_mime_message(Context, Context->hdrs[Context->v2r[i]]);
-	  if (Context->hdrs[Context->v2r[i]]->pgp & PGPENCRYPT &&
-	      !pgp_valid_passphrase())
+	  if (Context->hdrs[Context->v2r[i]]->security & ENCRYPT &&
+	      !crypt_valid_passphrase(Context->hdrs[Context->v2r[i]]->security))
 	    return 1;
 	}
     }
-#endif
     
-
-
-    if (option (OPTPIPESPLIT))
+    if (split)
     {
       for (i = 0; i < Context->vcount; i++)
       {
         if (Context->hdrs[Context->v2r[i]]->tagged)
         {
-	  endwin ();
-	  thepid = mutt_create_filter (buffer, &(s.fpout), NULL, NULL);
-          mutt_pipe_message_to_state (Context->hdrs[Context->v2r[i]], &s);
+	  mutt_message_hook (Context, Context->hdrs[Context->v2r[i]], M_MESSAGEHOOK);
+	  mutt_endwin (NULL);
+	  if ((thepid = mutt_create_filter (cmd, &fpout, NULL, NULL)) < 0)
+	  {
+	    mutt_perror _("Can't create filter process");
+	    return 1;
+	  }
+          pipe_msg (Context->hdrs[Context->v2r[i]], fpout, decode, print);
           /* add the message separator */
-          if (PipeSep)
-	    state_puts (PipeSep, &s);
-	  fclose (s.fpout);
+          if (sep)  fputs (sep, fpout);
+	  safe_fclose (&fpout);
 	  if (mutt_wait_filter (thepid) != 0)
 	    rc = 1;
         }
@@ -341,19 +426,23 @@ int mutt_pipe_message (HEADER *h)
     }
     else
     {
-      endwin ();
-      thepid = mutt_create_filter (buffer, &(s.fpout), NULL, NULL);
+      mutt_endwin (NULL);
+      if ((thepid = mutt_create_filter (cmd, &fpout, NULL, NULL)) < 0)
+      {
+	mutt_perror _("Can't create filter process");
+	return 1;
+      }
       for (i = 0; i < Context->vcount; i++)
       {
         if (Context->hdrs[Context->v2r[i]]->tagged)
         {
-          mutt_pipe_message_to_state (Context->hdrs[Context->v2r[i]], &s);
+	  mutt_message_hook (Context, Context->hdrs[Context->v2r[i]], M_MESSAGEHOOK);
+          pipe_msg (Context->hdrs[Context->v2r[i]], fpout, decode, print);
           /* add the message separator */
-          if (PipeSep)
-	    state_puts (PipeSep, &s);
+          if (sep) fputs (sep, fpout);
         }
       }
-      fclose (s.fpout);
+      safe_fclose (&fpout);
       if (mutt_wait_filter (thepid) != 0)
 	rc = 1;
     }
@@ -361,9 +450,52 @@ int mutt_pipe_message (HEADER *h)
 
   if (rc || option (OPTWAITKEY))
     mutt_any_key_to_continue (NULL);
-  return 1;
+  return rc;
 }
 
+void mutt_pipe_message (HEADER *h)
+{
+  char buffer[LONG_STRING];
+
+  buffer[0] = 0;
+  if (mutt_get_field (_("Pipe to command: "), buffer, sizeof (buffer), M_CMD)
+      != 0 || !buffer[0])
+    return;
+
+  mutt_expand_path (buffer, sizeof (buffer));
+  _mutt_pipe_message (h, buffer,
+		      option (OPTPIPEDECODE),
+		      0, 
+		      option (OPTPIPESPLIT),
+		      PipeSep);
+}
+
+void mutt_print_message (HEADER *h)
+{
+
+  if (quadoption (OPT_PRINT) && (!PrintCmd || !*PrintCmd))
+  {
+    mutt_message (_("No printing command has been defined."));
+    return;
+  }
+  
+  if (query_quadoption (OPT_PRINT,
+			h ? _("Print message?") : _("Print tagged messages?"))
+		  	!= M_YES)
+    return;
+
+  if (_mutt_pipe_message (h, PrintCmd,
+			  option (OPTPRINTDECODE),
+			  1,
+			  option (OPTPRINTSPLIT),
+			  "\f") == 0)
+    mutt_message (h ? _("Message printed") : _("Messages printed"));
+  else
+    mutt_message (h ? _("Message could not be printed") :
+		  _("Messages could not be printed"));
+}
+
+
 int mutt_select_sort (int reverse)
 {
   int method = Sort; /* save the current method in case of abort */
@@ -431,7 +563,7 @@ void mutt_shell_escape (void)
     if(buf[0])
     {
       CLEARLINE (LINES-1);
-      endwin ();
+      mutt_endwin (NULL);
       fflush (stdout);
       if (mutt_system (buf) != 0 || option (OPTWAITKEY))
 	mutt_any_key_to_continue (NULL);
@@ -445,8 +577,6 @@ void mutt_enter_command (void)
   BUFFER err, token;
   char buffer[LONG_STRING], errbuf[SHORT_STRING];
   int r;
-  int old_strictthreads = option (OPTSTRICTTHREADS);
-  int old_sortre	= option (OPTSORTRE);
 
   buffer[0] = 0;
   if (mutt_get_field (":", buffer, sizeof (buffer), M_COMMAND) != 0 || !buffer[0])
@@ -466,9 +596,6 @@ void mutt_enter_command (void)
     else
       mutt_error ("%s", errbuf);
   }
-  if (option (OPTSTRICTTHREADS) != old_strictthreads ||
-      option (OPTSORTRE) != old_sortre)
-    set_option (OPTNEEDRESORT);
 }
 
 void mutt_display_address (ENVELOPE *env)
@@ -480,9 +607,16 @@ void mutt_display_address (ENVELOPE *env)
   adr = mutt_get_address (env, &pfx);
 
   if (!adr) return;
-
+  
+  /* 
+   * Note: We don't convert IDNA to local representation this time.
+   * That is intentional, so the user has an opportunity to copy &
+   * paste the on-the-wire form of the address to other, IDN-unable
+   * software. 
+   */
+  
   buf[0] = 0;
-  rfc822_write_address (buf, sizeof (buf), adr);
+  rfc822_write_address (buf, sizeof (buf), adr, 0);
   mutt_message ("%s: %s", pfx, buf);
 }
 
@@ -491,37 +625,44 @@ static void set_copy_flags (HEADER *hdr, int decode, int decrypt, int *cmflags,
   *cmflags = 0;
   *chflags = CH_UPDATE_LEN;
   
-#ifdef HAVE_PGP
-  if (!decode && decrypt && (hdr->pgp & PGPENCRYPT))
+  if (WithCrypto && !decode && decrypt && (hdr->security & ENCRYPT))
   {
-    if (mutt_is_multipart_encrypted(hdr->content))
+    if ((WithCrypto & APPLICATION_PGP)
+        && mutt_is_multipart_encrypted(hdr->content))
     {
       *chflags = CH_NONEWLINE | CH_XMIT | CH_MIME;
       *cmflags = M_CM_DECODE_PGP;
     }
-    else if (mutt_is_application_pgp(hdr->content) & PGPENCRYPT)
+    else if ((WithCrypto & APPLICATION_PGP)
+              && mutt_is_application_pgp (hdr->content) & ENCRYPT)
       decode = 1;
+    else if ((WithCrypto & APPLICATION_SMIME)
+             && mutt_is_application_smime(hdr->content) & ENCRYPT)
+    {
+      *chflags = CH_NONEWLINE | CH_XMIT | CH_MIME;
+      *cmflags = M_CM_DECODE_SMIME;
+    }
   }
-#endif
 
   if (decode)
   {
     *chflags = CH_XMIT | CH_MIME | CH_TXTPLAIN;
     *cmflags = M_CM_DECODE | M_CM_CHARCONV;
-  }
 
-  /* respect $weed only if decode doesn't kick in
-   * for decrypt.
-   */
+    if (!decrypt)	/* If decode doesn't kick in for decrypt, */
+    {
+      *chflags |= CH_DECODE;	/* then decode RFC 2047 headers, */
 
-  if (decode && !decrypt && option (OPTWEED))
-  {
-    *chflags |= CH_WEED;
-    *cmflags |= M_CM_WEED;
+      if (option (OPTWEED))
+      {
+	*chflags |= CH_WEED;	/* and respect $weed. */
+	*cmflags |= M_CM_WEED;
+      }
+    }
   }
 }
 
-static void _mutt_save_message (HEADER *h, CONTEXT *ctx, int delete, int decode, int decrypt)
+void _mutt_save_message (HEADER *h, CONTEXT *ctx, int delete, int decode, int decrypt)
 {
   int cmflags, chflags;
   
@@ -543,9 +684,7 @@ int mutt_save_message (HEADER *h, int delete,
 		       int decode, int decrypt, int *redraw)
 {
   int i, need_buffy_cleanup;
-#ifdef HAVE_PGP
-  int need_passphrase = 0;
-#endif
+  int need_passphrase = 0, app=0;
   char prompt[SHORT_STRING], buf[_POSIX_PATH_MAX];
   CONTEXT ctx;
   struct stat st;
@@ -557,16 +696,24 @@ int mutt_save_message (HEADER *h, int delete,
 
   *redraw = 0;
 
-  snprintf (prompt, sizeof (prompt), _("%s%s to mailbox"),
-	    decode ? (delete ? _("Decode-save") : _("Decode-copy")) :
-	    (decrypt ? (delete ? _("Decrypt-save") : _("Decrypt-copy")):
-	     (delete ? _("Save") : _("Copy"))), h ? "" : _(" tagged"));
   
+  snprintf (prompt, sizeof (prompt),
+	    decode  ? (delete ? _("Decode-save%s to mailbox") :
+		       _("Decode-copy%s to mailbox")) :
+	    (decrypt ? (delete ? _("Decrypt-save%s to mailbox") :
+			_("Decrypt-copy%s to mailbox")) :
+	     (delete ? _("Save%s to mailbox") : _("Copy%s to mailbox"))),
+	    h ? "" : _(" tagged"));
+  
+
   if (h)
   {
-#ifdef HAVE_PGP
-    need_passphrase = h->pgp & PGPENCRYPT;
-#endif
+    if (WithCrypto)
+    {
+      need_passphrase = h->security & ENCRYPT;
+      app = h->security;
+    }
+    mutt_message_hook (Context, h, M_MESSAGEHOOK);
     mutt_default_save (buf, sizeof (buf), h);
   }
   else
@@ -582,12 +729,16 @@ int mutt_save_message (HEADER *h, int delete,
       }
     }
 
+
     if (h)
     {
+      mutt_message_hook (Context, h, M_MESSAGEHOOK);
       mutt_default_save (buf, sizeof (buf), h);
-#ifdef HAVE_PGP
-      need_passphrase |= h->pgp & PGPENCRYPT;
-#endif
+      if (WithCrypto)
+      {
+        need_passphrase = h->security & ENCRYPT;
+        app = h->security;
+      }
       h = NULL;
     }
   }
@@ -618,16 +769,12 @@ int mutt_save_message (HEADER *h, int delete,
   mutt_expand_path (buf, sizeof (buf));
 
   /* check to make sure that this file is really the one the user wants */
-  if (!mutt_save_confirm (buf, &st))
-  {
-    CLEARLINE (LINES-1);
-    return (-1);
-  }
+  if (mutt_save_confirm (buf, &st) != 0)
+    return -1;
 
-#ifdef HAVE_PGP
-  if(need_passphrase && (decode || decrypt) && !pgp_valid_passphrase())
+  if (WithCrypto && need_passphrase && (decode || decrypt)
+      && !crypt_valid_passphrase(app))
     return -1;
-#endif
   
   mutt_message (_("Copying to %s..."), buf);
   
@@ -656,12 +803,15 @@ int mutt_save_message (HEADER *h, int delete,
       for (i = 0; i < Context->vcount; i++)
       {
 	if (Context->hdrs[Context->v2r[i]]->tagged)
+	{
+	  mutt_message_hook (Context, Context->hdrs[Context->v2r[i]], M_MESSAGEHOOK);
 	  _mutt_save_message(Context->hdrs[Context->v2r[i]],
 			     &ctx, delete, decode, decrypt);
+	}
       }
     }
 
-    need_buffy_cleanup = (ctx.magic == M_MBOX || ctx.magic == M_MMDF || ctx.magic == M_KENDRA);
+    need_buffy_cleanup = (ctx.magic == M_MBOX || ctx.magic == M_MMDF);
 
     mx_close_mailbox (&ctx, NULL);
 
@@ -691,79 +841,30 @@ int mutt_save_message (HEADER *h, int delete,
   return -1;
 }
 
-/* XXX - merge this with mutt_pipe_message_to_state? */
-
-static void print_msg (FILE *fp, CONTEXT *ctx, HEADER *h)
-{
-  int cmflags = 0;
-  int chflags = CH_FROM;
-
-  pipe_print_set_flags (option (OPTPRINTDECODE), &cmflags, &chflags);
-
-#ifdef HAVE_PGP
-  if (option (OPTPRINTDECODE) && (h->pgp & PGPENCRYPT))
-  {
-    if (!pgp_valid_passphrase ())
-      return;
-    endwin ();
-  }
-#endif
-
-  if (option (OPTPRINTDECODE))
-    mutt_parse_mime_message (ctx, h);
-
-  mutt_copy_message (fp, ctx, h, cmflags, chflags);
-}
-
-void mutt_print_message (HEADER *h)
-{
-  int i, count = 0;
-  pid_t thepid;
-  FILE *fp;
-
-  if (query_quadoption (OPT_PRINT,
-			h ? _("Print message?") : _("Print tagged messages?"))
-		  	!= M_YES)
-    return;
-  endwin ();
-  if ((thepid = mutt_create_filter (NONULL(PrintCmd), &fp, NULL, NULL)) == -1)
-    return;
-  if (h)
-  {
-    print_msg (fp, Context, h);
-    count++;
-  }
-  else
-  {
-    for (i = 0 ; i < Context->vcount ; i++)
-    {
-      if (Context->hdrs[Context->v2r[i]]->tagged)
-      {
-	print_msg (fp, Context, Context->hdrs[Context->v2r[i]]);
-	/* add a formfeed */
-	fputc ('\f', fp);
-	count++;
-      }
-    }
-  }
-  fclose (fp);
-  if (mutt_wait_filter (thepid) || option (OPTWAITKEY))
-    mutt_any_key_to_continue (NULL);
-  mutt_message ((count > 1) ? _("Message printed") : _("Messages printed"));
-}
 
 void mutt_version (void)
 {
   mutt_message ("Mutt %s (%s)", MUTT_VERSION, ReleaseDate);
 }
 
-void mutt_edit_content_type (HEADER *h, BODY *b)
+void mutt_edit_content_type (HEADER *h, BODY *b, FILE *fp)
 {
   char buf[LONG_STRING];
+  char obuf[LONG_STRING];
   char tmp[STRING];
   PARAMETER *p;
+
+  char charset[STRING];
+  char *cp;
+
+  short charset_changed = 0;
+  short type_changed = 0;
   
+  cp = mutt_get_parameter ("charset", b->parameter);
+  strfcpy (charset, NONULL (cp), sizeof (charset));
+
   snprintf (buf, sizeof (buf), "%s/%s", TYPE (b), b->subtype);
+  strfcpy (obuf, buf, sizeof (obuf));
   if (b->parameter)
   {
     size_t l;
@@ -786,10 +887,34 @@ void mutt_edit_content_type (HEADER *h, BODY *b)
   FREE (&b->subtype);
   
   mutt_parse_content_type (buf, b);
+
   
-  mutt_message ("Content-Type changed to %s/%s.", TYPE (b),
-		NONULL (b->subtype));
+  snprintf (tmp, sizeof (tmp), "%s/%s", TYPE (b), NONULL (b->subtype));
+  type_changed = ascii_strcasecmp (tmp, obuf);
+  charset_changed = ascii_strcasecmp (charset, mutt_get_parameter ("charset", b->parameter));
+
+  /* if in send mode, check for conversion - current setting is default. */
+
+  if (!h && b->type == TYPETEXT && charset_changed)
+  {
+    int r;
+    snprintf (tmp, sizeof (tmp), _("Convert to %s upon sending?"),
+	      mutt_get_parameter ("charset", b->parameter));
+    if ((r = mutt_yesorno (tmp, !b->noconv)) != -1)
+      b->noconv = (r == M_NO);
+  }
+
+  /* inform the user */
   
+  if (type_changed)
+    mutt_message (_("Content-Type changed to %s."), tmp);
+  else if (b->type == TYPETEXT && charset_changed)
+    mutt_message (_("Character set changed to %s; %s."),
+		  mutt_get_parameter ("charset", b->parameter),
+		  b->noconv ? _("not converting") : _("converting"));
+
+  b->force_charset |= charset_changed ? 1 : 0;
+
   if (!is_multipart (b) && b->parts)
     mutt_free_body (&b->parts);
   if (!mutt_is_message_type (b->type, b->subtype) && b->hdr)
@@ -798,13 +923,52 @@ void mutt_edit_content_type (HEADER *h, BODY *b)
     mutt_free_header (&b->hdr);
   }
 
-#ifdef HAVE_PGP
-  if (h)
+  if (fp && (is_multipart (b) || mutt_is_message_type (b->type, b->subtype)))
+    mutt_parse_part (fp, b);
+  
+  if (WithCrypto && h)
   {
     if (h->content == b)
-      h->pgp = 0;
-    h->pgp |= pgp_query (b);
-  }
-#endif /* HAVE_PGP */
+      h->security  = 0;
 
+    h->security |= crypt_query (b);
+  }
 }
+
+
+static int _mutt_check_traditional_pgp (HEADER *h, int *redraw)
+{
+  MESSAGE *msg;
+  int rv = 0;
+  
+  mutt_parse_mime_message (Context, h);
+  if ((msg = mx_open_message (Context, h->msgno)) == NULL)
+    return 0;
+  if (crypt_pgp_check_traditional (msg->fp, h->content, 0))
+  {
+    h->security = crypt_query (h->content);
+    *redraw |= REDRAW_FULL;
+    rv = 1;
+  }
+  
+  mx_close_message (&msg);
+  return rv;
+}
+
+int mutt_check_traditional_pgp (HEADER *h, int *redraw)
+{
+  int i;
+  int rv = 0;
+  if (h)
+    rv = _mutt_check_traditional_pgp (h, redraw);
+  else
+  {
+    for (i = 0; i < Context->vcount; i++)
+      if (Context->hdrs[Context->v2r[i]]->tagged)
+	rv = _mutt_check_traditional_pgp (Context->hdrs[Context->v2r[i]], redraw)
+	  || rv;
+  }
+  return rv;
+}
+
+
