@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 1996-8 Michael R. Elkins <me@cs.hmc.edu>
+ * Copyright (C) 1996-2000 Michael R. Elkins <me@mutt.org>
  * 
  *     This program is free software; you can redistribute it and/or modify
  *     it under the terms of the GNU General Public License as published by
@@ -13,11 +13,12 @@
  * 
  *     You should have received a copy of the GNU General Public License
  *     along with this program; if not, write to the Free Software
- *     Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
+ *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
  */ 
 
 #include "mutt.h"
 #include "mutt_curses.h"
+#include "mutt_idna.h"
 #include "mutt_menu.h"
 #include "rfc1524.h"
 #include "mime.h"
@@ -25,14 +26,22 @@
 #include "mapping.h"
 #include "mailbox.h"
 #include "sort.h"
+#include "charset.h"
 
+#ifdef MIXMASTER
+#include "remailer.h"
+#endif
+
+#include <errno.h>
 #include <string.h>
 #include <sys/stat.h>
 #include <sys/wait.h>
 #include <unistd.h>
 #include <stdlib.h>
 
-#define CHECK_COUNT if (idxlen == 0) { mutt_error _("There are no attachments."); break; }
+static const char* There_are_no_attachments = N_("There are no attachments.");
+
+#define CHECK_COUNT if (idxlen == 0) { mutt_error _(There_are_no_attachments); break; }
 
 
 
@@ -46,19 +55,18 @@ enum
   HDR_REPLYTO,
   HDR_FCC,
 
-
-#ifdef _PGPPATH
-  HDR_PGP,
-  HDR_PGPSIGINFO,
+#ifdef MIXMASTER
+  HDR_MIX,
 #endif
-  
-  
+
+  HDR_CRYPT,
+  HDR_CRYPTINFO,
 
   HDR_ATTACH  = (HDR_FCC + 5) /* where to start printing the attachments */
 };
 
 #define HDR_XOFFSET 10
-#define TITLE_FMT "%10s"
+#define TITLE_FMT "%10s" /* Used for Prompts, which are ASCII */
 #define W (COLS - HDR_XOFFSET)
 
 static char *Prompts[] =
@@ -93,128 +101,239 @@ static void snd_entry (char *b, size_t blen, MUTTMENU *menu, int num)
 
 
 
-#ifdef _PGPPATH
-#include "pgp.h"
+#include "mutt_crypt.h"
 
-static void redraw_pgp_lines (int pgp)
+static void redraw_crypt_lines (HEADER *msg)
 {
-  mvaddstr (HDR_PGP, 0,     "     PGP: ");
-  if ((pgp & (PGPENCRYPT | PGPSIGN)) == (PGPENCRYPT | PGPSIGN))
+  int off = 0;
+
+  if ((WithCrypto & APPLICATION_PGP) && (WithCrypto & APPLICATION_SMIME))
+  {     
+    if (!msg->security)
+      mvaddstr (HDR_CRYPT, 0,     "Security: ");
+    else if (msg->security & APPLICATION_SMIME)
+      mvaddstr (HDR_CRYPT, 0,     "  S/MIME: ");
+    else if (msg->security & APPLICATION_PGP)
+      mvaddstr (HDR_CRYPT, 0,     "     PGP: ");
+  }
+  else if ((WithCrypto & APPLICATION_SMIME))
+    mvaddstr (HDR_CRYPT, 0,     "  S/MIME: ");
+  else if ((WithCrypto & APPLICATION_PGP))
+    mvaddstr (HDR_CRYPT, 0,     "     PGP: ");
+  else
+    return;
+
+  if ((msg->security & (ENCRYPT | SIGN)) == (ENCRYPT | SIGN))
     addstr (_("Sign, Encrypt"));
-  else if (pgp & PGPENCRYPT)
+  else if (msg->security & ENCRYPT)
     addstr (_("Encrypt"));
-  else if (pgp & PGPSIGN)
+  else if (msg->security & SIGN)
     addstr (_("Sign"));
   else
     addstr (_("Clear"));
   clrtoeol ();
 
-  move (HDR_PGPSIGINFO, 0);
+  move (HDR_CRYPTINFO, 0);
   clrtoeol ();
-  if (pgp & PGPSIGN)
-  {
+  if ((WithCrypto & APPLICATION_PGP)
+      && msg->security & APPLICATION_PGP  && msg->security & SIGN)
     printw ("%s%s", _(" sign as: "), PgpSignAs ? PgpSignAs : _("<default>"));
-    mvprintw (HDR_PGPSIGINFO, 40, "%s%s", _("MIC algorithm: "),
-	      NONULL(PgpSignMicalg));
+
+  if ((WithCrypto & APPLICATION_SMIME)
+      && msg->security & APPLICATION_SMIME  && msg->security & SIGN) {
+      printw ("%s%s", _(" sign as: "), SmimeDefaultKey ? SmimeDefaultKey : _("<default>"));
+  }
+
+  if ((WithCrypto & APPLICATION_SMIME)
+       && msg->security & APPLICATION_SMIME  && (msg->security & ENCRYPT)) {
+      mvprintw (HDR_CRYPTINFO, 40, "%s%s", _("Encrypt with: "),
+		NONULL(SmimeCryptAlg));
+      off = 20;
   }
 }
 
-static int pgp_send_menu (int bits, int *redraw)
+
+
+static int pgp_send_menu (HEADER *msg, int *redraw)
 {
-  char *p;
-  char *micalg = NULL;
+  pgp_key_t p;
   char input_signas[SHORT_STRING];
-  char input_micalg[SHORT_STRING];
-  KEYINFO *secring;
 
-  struct pgp_vinfo *pgp = pgp_get_vinfo(PGP_SIGN);
+  if (!(WithCrypto & APPLICATION_PGP))
+    return msg->security;
 
-  switch (mutt_multi_choice (_("(e)ncrypt, (s)ign, sign (a)s, (b)oth, select (m)ic algorithm, or (f)orget it? "),
-			     _("esabmf")))
+  switch (mutt_multi_choice (_("PGP (e)ncrypt, (s)ign, sign (a)s, (b)oth, or (f)orget it? "),
+			     _("esabf")))
   {
   case 1: /* (e)ncrypt */
-    bits |= PGPENCRYPT;
+    msg->security |= ENCRYPT;
     break;
 
   case 2: /* (s)ign */
-    bits |= PGPSIGN;
+    msg->security |= SIGN;
     break;
 
   case 3: /* sign (a)s */
     unset_option(OPTPGPCHECKTRUST);
 
-    if(pgp)
+    if ((p = crypt_pgp_ask_for_key (_("Sign as: "), NULL,
+                                    KEYFLAG_CANSIGN, PGP_PUBRING)))
     {
-      if(!(secring = pgp->read_secring(pgp)))
-      {
-	mutt_error _("Can't open your secret key ring!");
-	bits &= ~PGPSIGN;
-      }
-      else 
-      {
-	if ((p = pgp_ask_for_key (pgp, secring, _("Sign as: "), 
-				  NULL, KEYFLAG_CANSIGN, &micalg)))
-	{
-	  snprintf (input_signas, sizeof (input_signas), "0x%s", p);
-	  safe_free((void **) &PgpSignAs);
-	  PgpSignAs = safe_strdup(input_signas);
-	  safe_free((void **) &PgpSignMicalg);
-	  PgpSignMicalg = micalg;	/* micalg is malloc()ed by pgp_ask_for_key */
-	  pgp_void_passphrase ();	/* probably need a different passphrase */
-	  safe_free ((void **) &p);
-	  bits |= PGPSIGN;
-	}
-
-	pgp_close_keydb(&secring);
-	*redraw = REDRAW_FULL;
-      }
+      snprintf (input_signas, sizeof (input_signas), "0x%s",
+                crypt_pgp_keyid (p));
+      mutt_str_replace (&PgpSignAs, input_signas);
+      crypt_pgp_free_key (&p);
+      
+      msg->security |= SIGN;
+	
+      crypt_pgp_void_passphrase ();  /* probably need a different passphrase */
     }
     else
     {
-      bits &= ~PGPSIGN;
-      mutt_error _("An unkown PGP version was defined for signing.");
+      msg->security &= ~SIGN;
     }
+
+    *redraw = REDRAW_FULL;
     break;
 
   case 4: /* (b)oth */
-    bits = PGPENCRYPT | PGPSIGN;
+    msg->security = ENCRYPT | SIGN;
     break;
 
-  case 5: /* select (m)ic algorithm */
-    if(!(bits & PGPSIGN))
-      mutt_error _("This doesn't make sense if you don't want to sign the message.");
-    else
-    {
-      /* Copy the existing MIC algorithm into place */
-      strfcpy(input_micalg, NONULL(PgpSignMicalg), sizeof(input_micalg));
+  case 5: /* (f)orget it */
+    msg->security = 0;
+    break;
+  }
 
-      if(mutt_get_field (_("MIC algorithm: "), input_micalg, sizeof(input_micalg), 0) == 0)
-      {
-	if(mutt_strcasecmp(input_micalg, "pgp-md5") && mutt_strcasecmp(input_micalg, "pgp-sha1")
-	   && mutt_strcasecmp(input_micalg, "pgp-rmd160"))
-	{
-	  mutt_error _("Unknown MIC algorithm, valid ones are: pgp-md5, pgp-sha1, pgp-rmd160");
-	}
-	else 
-	{
-	  safe_free((void **) &PgpSignMicalg);
-	  PgpSignMicalg = safe_strdup(input_micalg);
-	}
-      }
+  if (msg->security && msg->security != APPLICATION_PGP)
+    msg->security |= APPLICATION_PGP;
+  else
+    msg->security = 0;
+
+  if(*redraw)
+      redraw_crypt_lines (msg);
+  return (msg->security);
+}
+
+
+
+static int smime_send_menu (HEADER *msg, int *redraw)
+{
+  char *p;
+
+  if (!(WithCrypto & APPLICATION_SMIME))
+    return msg->security;
+
+  switch (mutt_multi_choice (_("S/MIME (e)ncrypt, (s)ign, encrypt (w)ith, sign (a)s, (b)oth, or (f)orget it? "),
+			     _("eswabf")))
+  {
+  case 1: /* (e)ncrypt */
+    msg->security |= ENCRYPT;
+    break;
+
+  case 3: /* encrypt (w)ith */
+    msg->security |= ENCRYPT;
+    switch (mutt_multi_choice (_("1: DES, 2: Triple-DES, 3: RC2-40,"
+				 " 4: RC2-64, 5: RC2-128, or (f)orget it? "),
+			       _("12345f"))) {
+    case 1:
+	mutt_str_replace (&SmimeCryptAlg, "des");
+	break;
+    case 2:
+	mutt_str_replace (&SmimeCryptAlg, "des3");
+	break;
+    case 3:
+	mutt_str_replace (&SmimeCryptAlg, "rc2-40");
+	break;
+    case 4:
+	mutt_str_replace (&SmimeCryptAlg, "rc2-64");
+	break;
+    case 5:
+	mutt_str_replace (&SmimeCryptAlg, "rc2-128");
+	break;
+    case 6: /* forget it */
+	break;
     }
     break;
 
+  case 2: /* (s)ign */
+      
+    if(!SmimeDefaultKey)
+	mutt_message("Can\'t sign: No key specified. use sign(as).");
+    else
+	msg->security |= SIGN;
+    break;
+
+  case 4: /* sign (a)s */
+
+    if ((p = crypt_smime_ask_for_key (_("Sign as: "), NULL, 0))) {
+      p[mutt_strlen (p)-1] = '\0';
+      mutt_str_replace (&SmimeDefaultKey, p);
+	
+      msg->security |= SIGN;
+
+      /* probably need a different passphrase */
+      crypt_smime_void_passphrase ();
+    }
+    else
+      msg->security &= ~SIGN;
+
+    *redraw = REDRAW_FULL;
+    break;
+
+  case 5: /* (b)oth */
+    msg->security = ENCRYPT | SIGN;
+    break;
+
   case 6: /* (f)orget it */
-    bits = 0;
+    msg->security = 0;
     break;
   }
-  if (!*redraw)
-    redraw_pgp_lines (bits);
-  return (bits);
+
+  if (msg->security && msg->security != APPLICATION_SMIME)
+    msg->security |= APPLICATION_SMIME;
+  else
+    msg->security = 0;
+
+  if(*redraw)
+      redraw_crypt_lines (msg);
+  return (msg->security);
 }
-#endif /* _PGPPATH */
 
 
+#ifdef MIXMASTER
+
+static void redraw_mix_line (LIST *chain)
+{
+  int c;
+  char *t;
+
+  mvaddstr (HDR_MIX, 0,     "     Mix: ");
+
+  if (!chain)
+  {
+    addstr ("<no chain defined>");
+    clrtoeol ();
+    return;
+  }
+  
+  for (c = 12; chain; chain = chain->next)
+  {
+    t = chain->data;
+    if (t && t[0] == '0' && t[1] == '\0')
+      t = "<random>";
+    
+    if (c + mutt_strlen (t) + 2 >= COLS)
+      break;
+
+    addstr (NONULL(t));
+    if (chain->next)
+      addstr (", ");
+
+    c += mutt_strlen (t) + 2;
+  }
+}
+#endif /* MIXMASTER */
 
 static int
 check_attachments(ATTACHPTR **idx, short idxlen)
@@ -255,8 +374,9 @@ static void draw_envelope_addr (int line, ADDRESS *addr)
   char buf[STRING];
 
   buf[0] = 0;
-  rfc822_write_address (buf, sizeof (buf), addr);
-  mvprintw (line, 0, TITLE_FMT "%-*.*s", Prompts[line - 1], W, W, buf);
+  rfc822_write_address (buf, sizeof (buf), addr, 1);
+  mvprintw (line, 0, TITLE_FMT, Prompts[line - 1]);
+  mutt_paddstr (W, buf);
 }
 
 static void draw_envelope (HEADER *msg, char *fcc)
@@ -265,27 +385,35 @@ static void draw_envelope (HEADER *msg, char *fcc)
   draw_envelope_addr (HDR_TO, msg->env->to);
   draw_envelope_addr (HDR_CC, msg->env->cc);
   draw_envelope_addr (HDR_BCC, msg->env->bcc);
-  mvprintw (HDR_SUBJECT, 0, TITLE_FMT "%-*.*s", Prompts[HDR_SUBJECT - 1], W, W,
-	    NONULL(msg->env->subject));
+  mvprintw (HDR_SUBJECT, 0, TITLE_FMT, Prompts[HDR_SUBJECT - 1]);
+  mutt_paddstr (W, NONULL (msg->env->subject));
   draw_envelope_addr (HDR_REPLYTO, msg->env->reply_to);
-  mvprintw (HDR_FCC, 0, TITLE_FMT "%-*.*s", Prompts[HDR_FCC - 1], W, W, fcc);
+  mvprintw (HDR_FCC, 0, TITLE_FMT, Prompts[HDR_FCC - 1]);
+  mutt_paddstr (W, fcc);
 
+  if (WithCrypto)
+    redraw_crypt_lines (msg);
 
+#ifdef MIXMASTER
+  redraw_mix_line (msg->chain);
+#endif
 
-#ifdef _PGPPATH
-  redraw_pgp_lines (msg->pgp);
-#endif /* _PGPPATH */
+  SETCOLOR (MT_COLOR_STATUS);
+  mvaddstr (HDR_ATTACH - 1, 0, _("-- Attachments"));
+  BKGDSET (MT_COLOR_STATUS);
+  clrtoeol ();
 
-
-
-  mvaddstr (HDR_ATTACH - 1, 0, _("===== Attachments ====="));
+  BKGDSET (MT_COLOR_NORMAL);
+  SETCOLOR (MT_COLOR_NORMAL);
 }
 
 static int edit_address_list (int line, ADDRESS **addr)
 {
   char buf[HUGE_STRING] = ""; /* needs to be large for alias expansion */
-
-  rfc822_write_address (buf, sizeof (buf), *addr);
+  char *err = NULL;
+  
+  mutt_addrlist_to_local (*addr);
+  rfc822_write_address (buf, sizeof (buf), *addr, 0);
   if (mutt_get_field (Prompts[line - 1], buf, sizeof (buf), M_ALIAS) == 0)
   {
     rfc822_free_address (addr);
@@ -299,11 +427,19 @@ static int edit_address_list (int line, ADDRESS **addr)
     return (REDRAW_FULL);
   }
 
+  if (mutt_addrlist_to_idna (*addr, &err) != 0)
+  {
+    mutt_error (_("Warning: '%s' is a bad IDN."), err);
+    mutt_refresh();
+    FREE (&err);
+  }
+
   /* redraw the expanded list so the user can see the result */
   buf[0] = 0;
-  rfc822_write_address (buf, sizeof (buf), *addr);
-  mvprintw (line, HDR_XOFFSET, "%-*.*s", W, W, buf);
-
+  rfc822_write_address (buf, sizeof (buf), *addr, 1);
+  move (line, HDR_XOFFSET);
+  mutt_paddstr (W, buf);
+  
   return 0;
 }
 
@@ -333,8 +469,8 @@ static int delete_attachment (MUTTMENU *menu, short *idxlen, int x)
   idx[x]->content->next = NULL;
   idx[x]->content->parts = NULL;
   mutt_free_body (&(idx[x]->content));
-  safe_free ((void **) &idx[x]->tree);
-  safe_free ((void **) &idx[x]);
+  FREE (&idx[x]->tree);
+  FREE (&idx[x]);
   for (; x < *idxlen - 1; x++)
     idx[x] = idx[x+1];
   menu->max = --(*idxlen);
@@ -347,6 +483,7 @@ static void update_idx (MUTTMENU *menu, ATTACHPTR **idx, short idxlen)
   idx[idxlen]->level = (idxlen > 0) ? idx[idxlen-1]->level : 0;
   if (idxlen)
     idx[idxlen - 1]->content->next = idx[idxlen]->content;
+  idx[idxlen]->content->aptr = idx[idxlen];
   menu->current = idxlen++;
   mutt_update_tree (idx, idxlen);
   menu->max = idxlen;
@@ -354,6 +491,125 @@ static void update_idx (MUTTMENU *menu, ATTACHPTR **idx, short idxlen)
 }
 
 
+/* 
+ * cum_attachs_size: Cumulative Attachments Size
+ *
+ * Returns the total number of bytes used by the attachments in the
+ * attachment list _after_ content-transfer-encodings have been
+ * applied.
+ * 
+ */
+
+static unsigned long cum_attachs_size (MUTTMENU *menu)
+{
+  size_t s;
+  unsigned short i;
+  ATTACHPTR **idx = menu->data;
+  CONTENT *info;
+  BODY *b;
+  
+  for (i = 0, s = 0; i < menu->max; i++)
+  {
+    b = idx[i]->content;
+
+    if (!b->content)
+      b->content = mutt_get_content_info (b->filename, b);
+
+    if ((info = b->content))
+    {
+      switch (b->encoding)
+      {
+	case ENCQUOTEDPRINTABLE:
+	  s += 3 * (info->lobin + info->hibin) + info->ascii + info->crlf;
+	  break;
+	case ENCBASE64:
+	  s += (4 * (info->lobin + info->hibin + info->ascii + info->crlf)) / 3;
+	  break;
+	default:
+	  s += info->lobin + info->hibin + info->ascii + info->crlf;
+	  break;
+      }
+    }
+  }
+
+  return s;
+}
+
+/* prototype for use below */
+void compose_status_line (char *buf, size_t buflen, MUTTMENU *menu, 
+      const char *p);
+
+/*
+ * compose_format_str()
+ *
+ * %a = total number of attachments 
+ * %h = hostname  [option]
+ * %l = approx. length of current message (in bytes) 
+ * %v = Mutt version 
+ *
+ * This function is similar to status_format_str().  Look at that function for
+ * help when modifying this function.
+ */
+
+static const char *
+compose_format_str (char *buf, size_t buflen, char op, const char *src,
+		   const char *prefix, const char *ifstring,
+		   const char *elsestring,
+		   unsigned long data, format_flag flags)
+{
+  char fmt[SHORT_STRING], tmp[SHORT_STRING];
+  int optional = (flags & M_FORMAT_OPTIONAL);
+  MUTTMENU *menu = (MUTTMENU *) data;
+
+  *buf = 0;
+  switch (op)
+  {
+    case 'a': /* total number of attachments */
+	snprintf (fmt, sizeof (fmt), "%%%sd", prefix);
+	snprintf (buf, buflen, fmt, menu->max);
+      break;
+
+    case 'h':  /* hostname */
+      snprintf (fmt, sizeof (fmt), "%%%ss", prefix);
+      snprintf (buf, buflen, fmt, NONULL(Hostname));
+      break;
+
+    case 'l': /* approx length of current message in bytes */
+	snprintf (fmt, sizeof (fmt), "%%%ss", prefix);
+	mutt_pretty_size (tmp, sizeof (tmp), menu ? cum_attachs_size(menu) : 0);
+	snprintf (buf, buflen, fmt, tmp);
+      break;
+
+    case 'v':
+      snprintf (fmt, sizeof (fmt), "Mutt %%s");
+      snprintf (buf, buflen, fmt, MUTT_VERSION);
+      break;
+
+    case 0:
+      *buf = 0;
+      return (src);
+
+    default:
+      snprintf (buf, buflen, "%%%s%c", prefix, op);
+      break;
+  }
+
+  if (optional)
+    compose_status_line (buf, buflen, menu, ifstring);
+  else if (flags & M_FORMAT_OPTIONAL)
+    compose_status_line (buf, buflen, menu, elsestring);
+
+  return (src);
+}
+
+void compose_status_line (char *buf, size_t buflen, MUTTMENU *menu, 
+      const char *p)
+{
+  mutt_FormatString (buf, buflen, p, compose_format_str, 
+        (unsigned long) menu, 0);
+}
+
+
 /* return values:
  *
  * 1	message should be postponed
@@ -379,9 +635,10 @@ int mutt_compose_menu (HEADER *msg,   /* structure for new message */
   int fccSet = 0;	/* has the user edited the Fcc: field ? */
   CONTEXT *ctx = NULL, *this = NULL;
   /* Sort, SortAux could be changed in mutt_index_menu() */
-  int oldSort = Sort, oldSortAux = SortAux;
+  int oldSort, oldSortAux;
   struct stat st;
 
+  mutt_attach_init (msg->content);
   idx = mutt_gen_attach_list (msg->content, -1, idx, &idxlen, &idxmax, 0, 1);
 
   menu = mutt_new_menu ();
@@ -390,7 +647,6 @@ int mutt_compose_menu (HEADER *msg,   /* structure for new message */
   menu->max = idxlen;
   menu->make_entry = snd_entry;
   menu->tag = mutt_tag_attach;
-  menu->title = _("Compose");
   menu->data = idx;
   menu->help = mutt_compile_help (helpstr, sizeof (helpstr), MENU_COMPOSE, ComposeHelp);
   
@@ -399,7 +655,6 @@ int mutt_compose_menu (HEADER *msg,   /* structure for new message */
     switch (op = mutt_menuLoop (menu))
     {
       case OP_REDRAW:
-	menu_redraw_status (menu);
 	draw_envelope (msg, fcc);
 	menu->offset = HDR_ATTACH;
 	menu->pagelen = LINES - HDR_ATTACH - 2;
@@ -423,12 +678,11 @@ int mutt_compose_menu (HEADER *msg,   /* structure for new message */
 	  buf[0] = 0;
 	if (mutt_get_field ("Subject: ", buf, sizeof (buf), 0) == 0)
 	{
-	  safe_free ((void **) &msg->env->subject);
-	  msg->env->subject = safe_strdup (buf);
+	  mutt_str_replace (&msg->env->subject, buf);
 	  move (HDR_SUBJECT, HDR_XOFFSET);
 	  clrtoeol ();
 	  if (msg->env->subject)
-	    printw ("%-*.*s", W, W, msg->env->subject);
+	    mutt_paddstr (W, msg->env->subject);
 	}
 	break;
       case OP_COMPOSE_EDIT_REPLY_TO:
@@ -440,7 +694,8 @@ int mutt_compose_menu (HEADER *msg,   /* structure for new message */
 	{
 	  strfcpy (fcc, buf, _POSIX_PATH_MAX);
 	  mutt_pretty_mailbox (fcc);
-	  mvprintw (HDR_FCC, HDR_XOFFSET, "%-*.*s", W, W, fcc);
+	  move (HDR_FCC, HDR_XOFFSET);
+	  mutt_paddstr (W, fcc);
 	  fccSet = 1;
 	}
 	MAYBE_REDRAW (menu->redraw);
@@ -450,7 +705,7 @@ int mutt_compose_menu (HEADER *msg,   /* structure for new message */
 	{
 	  mutt_edit_file (Editor, msg->content->filename);
 	  mutt_update_encoding (msg->content);
-	  menu->redraw = REDRAW_CURRENT;
+	  menu->redraw = REDRAW_CURRENT | REDRAW_STATUS;
 	  break;
 	}
 	/* fall through */
@@ -458,8 +713,15 @@ int mutt_compose_menu (HEADER *msg,   /* structure for new message */
 	if (op == OP_COMPOSE_EDIT_HEADERS ||
 	    (op == OP_COMPOSE_EDIT_MESSAGE && option (OPTEDITHDRS)))
 	{
-	  mutt_edit_headers ((!Editor || mutt_strcmp ("builtin", Editor) == 0) ? NONULL(Visual) : NONULL(Editor),
-			     msg->content->filename, msg, fcc, fcclen);
+	  char *tag = NULL, *err = NULL;
+	  mutt_env_to_local (msg->env);
+	  mutt_edit_headers (NONULL (Editor), msg->content->filename, msg,
+			     fcc, fcclen);
+	  if (mutt_env_to_idna (msg->env, &tag, &err))
+	  {
+	    mutt_error (_("Bad IDN in \"%s\": '%s'"), tag, err);
+	    FREE (&err);
+	  }
 	}
 	else
 	{
@@ -475,7 +737,7 @@ int mutt_compose_menu (HEADER *msg,   /* structure for new message */
 	if (idxlen && idx[idxlen - 1]->content->next)
 	{
 	  for (i = 0; i < idxlen; i++)
-	    safe_free ((void **) &idx[i]);
+	    FREE (&idx[i]);
 	  idxlen = 0;
 	  idx = mutt_gen_attach_list (msg->content, -1, idx, &idxlen, &idxmax, 0, 1);
 	  menu->data = idx;
@@ -487,85 +749,111 @@ int mutt_compose_menu (HEADER *msg,   /* structure for new message */
 
 
 
-#ifdef _PGPPATH
       case OP_COMPOSE_ATTACH_KEY:
-
+        if (!(WithCrypto & APPLICATION_PGP))
+          break;       
 	if (idxlen == idxmax)
         {
-	  safe_realloc ((void **) &idx, sizeof (ATTACHPTR *) * (idxmax += 5));
+	  safe_realloc (&idx, sizeof (ATTACHPTR *) * (idxmax += 5));
 	  menu->data = idx;
 	}
 	
 	idx[idxlen] = (ATTACHPTR *) safe_calloc (1, sizeof (ATTACHPTR));
-	if ((idx[idxlen]->content = pgp_make_key_attachment(NULL)) != NULL)
+	if ((idx[idxlen]->content = crypt_pgp_make_key_attachment(NULL)) != NULL)
 	{
-	  idx[idxlen]->level = (idxlen > 0) ? idx[idxlen-1]->level : 0;
-
-	  if(idxlen)
-	    idx[idxlen - 1]->content->next = idx[idxlen]->content;
-	  
-	  menu->current = idxlen++;
-	  mutt_update_tree (idx, idxlen);
-	  menu->max = idxlen;
+	  update_idx (menu, idx, idxlen++);
 	  menu->redraw |= REDRAW_INDEX;
 	}
 	else
-	  safe_free ((void **) &idx[idxlen]);
+	  FREE (&idx[idxlen]);
 
 	menu->redraw |= REDRAW_STATUS;
 
-	if(option(OPTNEEDREDRAW))
+	if (option(OPTNEEDREDRAW))
 	{
 	  menu->redraw = REDRAW_FULL;
 	  unset_option(OPTNEEDREDRAW);
 	}
 	
 	break;
-#endif
-
 
 
       case OP_COMPOSE_ATTACH_FILE:
-      case OP_COMPOSE_ATTACH_MESSAGE:
-
-	fname[0] = 0;
 	{
-	  char* prompt;
-	  int flag;
+	  char *prompt, **files;
+	  int error, numfiles;
 
-	  if (op == OP_COMPOSE_ATTACH_FILE)
-	  {
-	    prompt = _("Attach file");
-	    flag = 0;
-	  }
-	  else
-	  {
-	    prompt = _("Open mailbox to attach message from");
-	    if (Context)
-	    {
-	      strfcpy (fname, NONULL (Context->path), sizeof (fname));
-	      mutt_pretty_mailbox (fname);
-	    }
-	    flag = 1;
-	  }
+	  fname[0] = 0;
+	  prompt = _("Attach file");
+	  numfiles = 0;
+	  files = NULL;
 
-	  if (mutt_enter_fname (prompt, fname, sizeof (fname), &menu->redraw, flag) == -1)
+	  if (_mutt_enter_fname (prompt, fname, sizeof (fname), &menu->redraw, 0, 1, &files, &numfiles) == -1 ||
+	      *fname == '\0')
 	    break;
+
+	  if (idxlen + numfiles >= idxmax)
+	  {
+	    safe_realloc (&idx, sizeof (ATTACHPTR *) * (idxmax += 5 + numfiles));
+	    menu->data = idx;
+	  }
+
+	  error = 0;
+	  if (numfiles > 1)
+	    mutt_message _("Attaching selected files...");
+	  for (i = 0; i < numfiles; i++)
+	  {
+	    char *att = files[i];
+	    idx[idxlen] = (ATTACHPTR *) safe_calloc (1, sizeof (ATTACHPTR));
+	    idx[idxlen]->content = mutt_make_file_attach (att);
+	    if (idx[idxlen]->content != NULL)
+	      update_idx (menu, idx, idxlen++);
+	    else
+	    {
+	      error = 1;
+	      mutt_error (_("Unable to attach %s!"), att);
+	      FREE (&idx[idxlen]);
+	    }
+	  }
+	  
+	  FREE (&files);
+	  if (!error) mutt_clear_error ();
+
+	  menu->redraw |= REDRAW_INDEX | REDRAW_STATUS;
 	}
+	break;
 
-	if (!fname[0])
-	  continue;
-	mutt_expand_path (fname, sizeof (fname));
-
-	/* check to make sure the file exists and is readable */
-	if (access (fname, R_OK) == -1)
+      case OP_COMPOSE_ATTACH_MESSAGE:
 	{
-	  mutt_perror (fname);
-	  break;
-	}
+	  char *prompt;
+	  HEADER *h;
+
+	  fname[0] = 0;
+	  prompt = _("Open mailbox to attach message from");
+
+	  if (Context)
+	  {
+	    strfcpy (fname, NONULL (Context->path), sizeof (fname));
+	    mutt_pretty_mailbox (fname);
+	  }
+
+	  if (mutt_enter_fname (prompt, fname, sizeof (fname), &menu->redraw, 1) == -1 || !fname[0])
+	    break;
+
+	  mutt_expand_path (fname, sizeof (fname));
+#ifdef USE_IMAP
+          if (!mx_is_imap (fname))
+#endif
+#ifdef USE_POP
+          if (!mx_is_pop (fname))
+#endif
+	  /* check to make sure the file exists and is readable */
+	  if (access (fname, R_OK) == -1)
+	  {
+	    mutt_perror (fname);
+	    break;
+	  }
 
-	if (op == OP_COMPOSE_ATTACH_MESSAGE)
-	{
 	  menu->redraw = REDRAW_FULL;
 
 	  ctx = mx_open_mailbox (fname, M_READONLY, NULL);
@@ -577,13 +865,15 @@ int mutt_compose_menu (HEADER *msg,   /* structure for new message */
 
 	  if (!ctx->msgcount)
 	  {
-	    mx_close_mailbox (ctx);
-	    safe_free ((void **) &ctx);
+	    mx_close_mailbox (ctx, NULL);
+	    FREE (&ctx);
 	    mutt_error _("No messages in that folder.");
 	    break;
 	  }
+
+	  this = Context; /* remember current folder and sort methods*/
+	  oldSort = Sort; oldSortAux = SortAux;
 	  
-	  this = Context; /* remember current folder */
 	  Context = ctx;
 	  set_option(OPTATTACHMSG);
 	  mutt_message _("Tag the messages you want to attach!");
@@ -600,36 +890,13 @@ int mutt_compose_menu (HEADER *msg,   /* structure for new message */
 	    menu->redraw |= REDRAW_INDEX | REDRAW_STATUS;
 	    break;
 	  }
-	}
-        {
-	  int numtag = 0;
 
-	  if (op == OP_COMPOSE_ATTACH_MESSAGE)
-	    numtag = Context->tagged;
-	  if (idxlen + numtag >= idxmax)
+	  if (idxlen + Context->tagged >= idxmax)
 	  {
-	    safe_realloc ((void **) &idx, sizeof (ATTACHPTR *) * (idxmax += 5 + numtag));
+	    safe_realloc (&idx, sizeof (ATTACHPTR *) * (idxmax += 5 + Context->tagged));
 	    menu->data = idx;
 	  }
-	}
 
-	if (op == OP_COMPOSE_ATTACH_FILE)
-	{
-	  idx[idxlen] = (ATTACHPTR *) safe_calloc (1, sizeof (ATTACHPTR));
-	  idx[idxlen]->content = mutt_make_file_attach (fname);
-	  if (idx[idxlen]->content != NULL)
-	    update_idx (menu, idx, idxlen++);
-	  else
-	  {
-	    mutt_error _("Unable to attach!");
-	    safe_free ((void **) &idx[idxlen]);
-	  }
-	  menu->redraw |= REDRAW_INDEX | REDRAW_STATUS;
-	  break;
-        }
-	else
-	{
-	  HEADER *h;
 	  for (i = 0; i < Context->msgcount; i++)
 	  {
 	    h = Context->hdrs[i];
@@ -642,24 +909,24 @@ int mutt_compose_menu (HEADER *msg,   /* structure for new message */
 	      else
 	      {
 		mutt_error _("Unable to attach!");
-		safe_free ((void **) &idx[idxlen]);
+		FREE (&idx[idxlen]);
 	      }
 	    }
 	  }
 	  menu->redraw |= REDRAW_FULL;
-	}
 
-	if (close == OP_QUIT) 
-	  mx_close_mailbox (Context);
-	else
-	  mx_fastclose_mailbox (Context);
-	safe_free ((void **) &Context);
-	
-	/* go back to the folder we started from */
-	Context = this;
-	/* Restore old $sort and $sort_aux */
-	Sort = oldSort;
-	SortAux = oldSortAux;
+	  if (close == OP_QUIT) 
+	    mx_close_mailbox (Context, NULL);
+	  else
+	    mx_fastclose_mailbox (Context);
+	  FREE (&Context);
+
+	  /* go back to the folder we started from */
+	  Context = this;
+	  /* Restore old $sort and $sort_aux */
+	  Sort = oldSort;
+	  SortAux = oldSortAux;
+	}
 	break;
 
       case OP_DELETE:
@@ -677,8 +944,30 @@ int mutt_compose_menu (HEADER *msg,   /* structure for new message */
 
 	if (menu->current == 0)
 	  msg->content = idx[0]->content;
+
+        menu->redraw |= REDRAW_STATUS;
 	break;
 
+#define CURRENT idx[menu->current]->content
+      
+      case OP_COMPOSE_TOGGLE_RECODE:
+      {      
+        CHECK_COUNT;
+        if (!mutt_is_text_part (CURRENT))
+        {
+	  mutt_error (_("Recoding only affects text attachments."));
+	  break;
+	}
+        CURRENT->noconv = !CURRENT->noconv;
+        if (CURRENT->noconv)
+	  mutt_message (_("The current attachment won't be converted."));
+        else
+	  mutt_message (_("The current attachment will be converted."));
+	menu->redraw = REDRAW_CURRENT;
+        break;
+      }
+#undef CURRENT
+
       case OP_COMPOSE_EDIT_DESCRIPTION:
 	CHECK_COUNT;
 	strfcpy (buf,
@@ -688,77 +977,44 @@ int mutt_compose_menu (HEADER *msg,   /* structure for new message */
 	/* header names should not be translated */
 	if (mutt_get_field ("Description: ", buf, sizeof (buf), 0) == 0)
 	{
-	  safe_free ((void **) &idx[menu->current]->content->description);
-	  idx[menu->current]->content->description = safe_strdup (buf);
+	  mutt_str_replace (&idx[menu->current]->content->description, buf);
 	  menu->redraw = REDRAW_CURRENT;
 	}
 	break;
 
       case OP_COMPOSE_UPDATE_ENCODING:
         CHECK_COUNT;
-        if(menu->tagprefix)
+        if (menu->tagprefix)
         {
 	  BODY *top;
-	  for(top = msg->content; top; top = top->next)
+	  for (top = msg->content; top; top = top->next)
 	  {
-	    if(top->tagged)
-	      mutt_update_encoding(top);
+	    if (top->tagged)
+	      mutt_update_encoding (top);
 	  }
 	  menu->redraw = REDRAW_FULL;
 	}
         else
         {
           mutt_update_encoding(idx[menu->current]->content);
-	  menu->redraw = REDRAW_CURRENT;
+	  menu->redraw = REDRAW_CURRENT | REDRAW_STATUS;
 	}
         break;
       
-      case OP_COMPOSE_EDIT_TYPE:
+      case OP_COMPOSE_TOGGLE_DISPOSITION:
+	/* toggle the content-disposition between inline/attachment */
+	idx[menu->current]->content->disposition = (idx[menu->current]->content->disposition == DISPINLINE) ? DISPATTACH : DISPINLINE;
+	menu->redraw = REDRAW_CURRENT;
+	break;
+
+      case OP_EDIT_TYPE:
 	CHECK_COUNT;
-	snprintf (buf, sizeof (buf), "%s/%s",
-		  TYPE (idx[menu->current]->content),
-		  idx[menu->current]->content->subtype);
-	if (mutt_get_field ("Content-Type: ", buf, sizeof (buf), 0) == 0 && buf[0])
-	{
-	  char *s;
-	  PARAMETER *par;
-	  BODY *b;
-	  
-	  b = idx[menu->current]->content;
-	  
-	  s  = b->filename;  par = b->parameter;
-	  b->filename  = NULL;  b->parameter = NULL;
-	  
-	  mutt_parse_content_type (buf, b);
+        {
+	  mutt_edit_content_type (NULL, idx[menu->current]->content, NULL);
 
-	  safe_free ((void **) &b->filename);
-	  b->filename = s;
-	  
-	  if ((s = mutt_get_parameter("charset", b->parameter)))
-	    mutt_set_parameter("charset", s, &par);
-	  
-	  /* These are needed for "traditional" PGP.
-	   * Should we switch to a "negative list" instead?
-	   */
+	  /* this may have been a change to text/something */
+	  mutt_update_encoding (idx[menu->current]->content);
 
-	  if ((s = mutt_get_parameter("x-action", b->parameter)))
-	    mutt_set_parameter("x-action", s, &par);
-	  if ((s = mutt_get_parameter("format", b->parameter)))
-	    mutt_set_parameter("format", s, &par);
-
-	  /* ignore the other parameters for now */
-	  mutt_free_parameter(&b->parameter);
-	  b->parameter = par;
-
-	  /* this may have been a "structured" message */
-	  if  (b->parts)
-	    mutt_free_body (&b->parts);
-	  if (b->hdr)
-	  {
-	    b->hdr->content = NULL;
-	    mutt_free_header (&b->hdr);
-	  }
-	  
 	  menu->redraw = REDRAW_CURRENT;
 	}
 	break;
@@ -773,7 +1029,7 @@ int mutt_compose_menu (HEADER *msg,   /* structure for new message */
 	  if ((i = mutt_check_encoding (buf)) != ENCOTHER && i != ENCUUENCODED)
 	  {
 	    idx[menu->current]->content->encoding = i;
-	    menu->redraw = REDRAW_CURRENT;
+	    menu->redraw = REDRAW_CURRENT | REDRAW_STATUS;
 	    mutt_clear_error();
 	  }
 	  else
@@ -788,6 +1044,12 @@ int mutt_compose_menu (HEADER *msg,   /* structure for new message */
 	  menu->redraw = REDRAW_FULL;
 	  break;
 	}
+
+      
+#ifdef MIXMASTER
+        if (msg->chain && mix_check_message (msg) != 0)
+	  break;
+#endif
       
 	if (!fccSet && *fcc)
 	{
@@ -804,17 +1066,23 @@ int mutt_compose_menu (HEADER *msg,   /* structure for new message */
 
       case OP_COMPOSE_EDIT_FILE:
 	CHECK_COUNT;
-	mutt_edit_file ((!Editor || mutt_strcmp ("builtin", Editor) == 0) ? NONULL(Visual) : NONULL(Editor),
-			idx[menu->current]->content->filename);
+	mutt_edit_file (NONULL(Editor), idx[menu->current]->content->filename);
 	mutt_update_encoding (idx[menu->current]->content);
-	menu->redraw = REDRAW_CURRENT;
+	menu->redraw = REDRAW_CURRENT | REDRAW_STATUS;
 	break;
 
       case OP_COMPOSE_TOGGLE_UNLINK:
 	CHECK_COUNT;
 	idx[menu->current]->content->unlink = !idx[menu->current]->content->unlink;
-	if (option (OPTRESOLVE) && menu->current + 1 < menu->max)
+
+#if 0
+        /* OPTRESOLVE is otherwise ignored on this menu.
+	 * Where's the bug?
+	 */
+
+        if (option (OPTRESOLVE) && menu->current + 1 < menu->max)
 	  menu->current++;
+# endif
 	menu->redraw = REDRAW_INDEX;
 	break;
 
@@ -842,9 +1110,9 @@ int mutt_compose_menu (HEADER *msg,   /* structure for new message */
 	if (mutt_get_field (_("Rename to: "), fname, sizeof (fname), M_FILE)
 							== 0 && fname[0])
 	{
-	  if(stat(idx[menu->current]->content->filename, &st) == -1)
+	  if (stat(idx[menu->current]->content->filename, &st) == -1)
 	  {
-	    mutt_error (_("Can't stat: %s"), fname);
+	    mutt_error (_("Can't stat %s: %s"), fname, strerror (errno));
 	    break;
 	  }
 
@@ -852,8 +1120,7 @@ int mutt_compose_menu (HEADER *msg,   /* structure for new message */
 	  if(mutt_rename_file (idx[menu->current]->content->filename, fname))
 	    break;
 	  
-	  safe_free ((void **) &idx[menu->current]->content->filename);
-	  idx[menu->current]->content->filename = safe_strdup (fname);
+	  mutt_str_replace (&idx[menu->current]->content->filename, fname);
 	  menu->redraw = REDRAW_CURRENT;
 
 	  if(idx[menu->current]->content->stamp >= st.st_mtime)
@@ -895,7 +1162,7 @@ int mutt_compose_menu (HEADER *msg,   /* structure for new message */
 	  }
 	  if (idxlen == idxmax)
 	  {
-	    safe_realloc ((void **) &idx, sizeof (ATTACHPTR *) * (idxmax += 5));
+	    safe_realloc (&idx, sizeof (ATTACHPTR *) * (idxmax += 5));
 	    menu->data = idx;
 	  }
 
@@ -904,7 +1171,7 @@ int mutt_compose_menu (HEADER *msg,   /* structure for new message */
 	  if (!(fp = safe_fopen (fname, "w")))
 	  {
 	    mutt_error (_("Can't create file %s"), fname);
-	    safe_free ((void **) &idx[idxlen]);
+	    FREE (&idx[idxlen]);
 	    continue;
 	  }
 	  fclose (fp);
@@ -914,18 +1181,10 @@ int mutt_compose_menu (HEADER *msg,   /* structure for new message */
 	    mutt_error _("What we have here is a failure to make an attachment");
 	    continue;
 	  }
-	  
-	  idx[idxlen]->level = (idxlen > 0) ? idx[idxlen-1]->level : 0;
-	  if (idxlen)
-	    idx[idxlen - 1]->content->next = idx[idxlen]->content;
-	  
-	  menu->current = idxlen++;
-	  mutt_update_tree (idx, idxlen);
-	  menu->max = idxlen;
+	  update_idx (menu, idx, idxlen++);
 
 	  idx[menu->current]->content->type = itype;
-	  safe_free ((void **) &idx[menu->current]->content->subtype);
-	  idx[menu->current]->content->subtype = safe_strdup (p);
+	  mutt_str_replace (&idx[menu->current]->content->subtype, p);
 	  idx[menu->current]->content->unlink = 1;
 	  menu->redraw |= REDRAW_INDEX | REDRAW_STATUS;
 
@@ -949,14 +1208,13 @@ int mutt_compose_menu (HEADER *msg,   /* structure for new message */
       case OP_VIEW_ATTACH:
       case OP_DISPLAY_HEADERS:
 	CHECK_COUNT;
-	mutt_attach_display_loop (menu, op, NULL, idx);
+	mutt_attach_display_loop (menu, op, NULL, NULL, NULL, &idx, &idxlen, NULL, 0);
 	menu->redraw = REDRAW_FULL;
-	mutt_clear_error ();
 	break;
 
       case OP_SAVE:
 	CHECK_COUNT;
-	mutt_save_attachment_list (NULL, menu->tagprefix, menu->tagprefix ?  msg->content : idx[menu->current]->content, NULL);
+	mutt_save_attachment_list (NULL, menu->tagprefix, menu->tagprefix ?  msg->content : idx[menu->current]->content, NULL, menu);
 	MAYBE_REDRAW (menu->redraw);
 	break;
 
@@ -971,6 +1229,7 @@ int mutt_compose_menu (HEADER *msg,   /* structure for new message */
 	mutt_pipe_attachment_list (NULL, menu->tagprefix, menu->tagprefix ? msg->content : idx[menu->current]->content, op == OP_FILTER);
 	if (op == OP_FILTER) /* cte might have changed */
 	  menu->redraw = menu->tagprefix ? REDRAW_FULL : REDRAW_CURRENT; 
+        menu->redraw |= REDRAW_STATUS;
 	break;
 
       case OP_EXIT:
@@ -982,10 +1241,10 @@ int mutt_compose_menu (HEADER *msg,   /* structure for new message */
 	    idx[idxlen]->content->next = NULL;
 	    idx[idxlen]->content->parts = NULL;
 	    mutt_free_body (&idx[idxlen]->content);
-	    safe_free ((void **) &idx[idxlen]->tree);
-	    safe_free ((void **) &idx[idxlen]);
+	    FREE (&idx[idxlen]->tree);
+	    FREE (&idx[idxlen]);
 	  }
-	  safe_free ((void **) &idx);
+	  FREE (&idx);
 	  idxlen = 0;
 	  idxmax = 0;
 	  r = -1;
@@ -998,7 +1257,7 @@ int mutt_compose_menu (HEADER *msg,   /* structure for new message */
 	/* fall through to postpone! */
 
       case OP_COMPOSE_POSTPONE_MESSAGE:
-      
+
         if(check_attachments(idx, idxlen) != 0)
         {
 	  menu->redraw = REDRAW_FULL;
@@ -1012,8 +1271,13 @@ int mutt_compose_menu (HEADER *msg,   /* structure for new message */
       case OP_COMPOSE_ISPELL:
 	endwin ();
 	snprintf (buf, sizeof (buf), "%s -x %s", NONULL(Ispell), msg->content->filename);
-	mutt_system (buf);
-        mutt_update_encoding(msg->content);
+	if (mutt_system (buf) == -1)
+	  mutt_error (_("Error running \"%s\"!"), buf);
+	else
+        {
+	  mutt_update_encoding (msg->content);
+	  menu->redraw |= REDRAW_STATUS;
+	}
 	break;
 
       case OP_COMPOSE_WRITE_MESSAGE:
@@ -1029,41 +1293,84 @@ int mutt_compose_menu (HEADER *msg,   /* structure for new message */
        if (mutt_enter_fname (_("Write message to mailbox"), fname, sizeof (fname),
                              &menu->redraw, 1) != -1 && fname[0])
        {
-	 int oldhdrdate;
          mutt_message (_("Writing message to %s ..."), fname);
          mutt_expand_path (fname, sizeof (fname));
 
          if (msg->content->next)
            msg->content = mutt_make_multipart (msg->content);
 
-	 oldhdrdate = option(OPTUSEHEADERDATE);
-	 set_option(OPTUSEHEADERDATE);
          if (mutt_write_fcc (NONULL (fname), msg, NULL, 1, NULL) < 0)
            msg->content = mutt_remove_multipart (msg->content);
          else
            mutt_message _("Message written.");
-	 if(!oldhdrdate) unset_option(OPTUSEHEADERDATE);
        }
        break;
 
 
 
-#ifdef _PGPPATH
       case OP_COMPOSE_PGP_MENU:
-
-	msg->pgp = pgp_send_menu (msg->pgp, &menu->redraw);
+        if (!(WithCrypto & APPLICATION_PGP))
+          break;
+	if ((WithCrypto & APPLICATION_SMIME)
+            && msg->security & APPLICATION_SMIME)
+	{
+	  if (mutt_yesorno (_("S/MIME already selected. Clear & continue ? "),
+			     M_YES) != M_YES)
+	  {
+	    mutt_clear_error ();
+	    break;
+	  }
+	  msg->security = 0;
+	}
+	msg->security = pgp_send_menu (msg, &menu->redraw);
+	redraw_crypt_lines (msg);
 	break;
 
+
       case OP_FORGET_PASSPHRASE:
-
-	mutt_forget_passphrase ();
+	crypt_forget_passphrase ();
 	break;
 
-#endif /* _PGPPATH */
+
+      case OP_COMPOSE_SMIME_MENU:
+        if (!(WithCrypto & APPLICATION_SMIME))
+          break;
+
+	if ((WithCrypto & APPLICATION_PGP)
+            && msg->security & APPLICATION_PGP)
+	{
+	  if (mutt_yesorno (_("PGP already selected. Clear & continue ? "),
+			      M_YES) != M_YES)
+	  {
+	     mutt_clear_error ();
+	     break;
+	  }
+	  msg->security = 0;
+	}
+	msg->security = smime_send_menu(msg, &menu->redraw);
+	redraw_crypt_lines (msg);
+	break;
 
 
+#ifdef MIXMASTER
+      case OP_COMPOSE_MIX:
+      
+      	mix_make_chain (&msg->chain, &menu->redraw);
+        break;
+#endif
 
     }
+
+    /* Draw formated compose status line */
+    if (menu->redraw & REDRAW_STATUS) 
+    {
+       	compose_status_line (buf, sizeof (buf), menu, NONULL(ComposeFormat));
+	CLEARLINE (option (OPTSTATUSONTOP) ? 0 : LINES-2);
+	SETCOLOR (MT_COLOR_STATUS);
+	printw ("%-*.*s", COLS, COLS, buf);
+	SETCOLOR (MT_COLOR_NORMAL);
+	menu->redraw &= ~REDRAW_STATUS;
+    }
   }
 
   mutt_menuDestroy (&menu);
@@ -1072,12 +1379,16 @@ int mutt_compose_menu (HEADER *msg,   /* structure for new message */
   {
     msg->content = idx[0]->content;
     for (i = 0; i < idxlen; i++)
-      safe_free ((void **) &idx[i]);
+    {
+      idx[i]->content->aptr = NULL;
+      FREE (&idx[i]);
+    }
   }
   else
     msg->content = NULL;
 
-  safe_free ((void **) &idx);
+  FREE (&idx);
 
   return (r);
 }
+
