@@ -1,6 +1,6 @@
 /*
  * Copyright (C) 1998-2000 Werner Koch <werner.koch@guug.de>
- * Copyright (C) 1999-2000 Thomas Roessler <roessler@guug.de>
+ * Copyright (C) 1999-2000 Thomas Roessler <roessler@does-not-exist.org>
  *
  *     This program is free software; you can redistribute it
  *     and/or modify it under the terms of the GNU General Public
@@ -67,15 +67,16 @@
 
 /* decode the backslash-escaped user ids. */
 
-static CHARSET *_chs;
+static char *_chs = 0;
 
 static void fix_uid (char *uid)
 {
   char *s, *d;
-  
+  iconv_t cd;
+
   for (s = d = uid; *s;)
   {
-    if (*s == '\\' && *(s+1) == 'x' && isxdigit (*(s+2)) && isxdigit (*(s+3)))
+    if (*s == '\\' && *(s+1) == 'x' && isxdigit ((unsigned char) *(s+2)) && isxdigit ((unsigned char) *(s+3)))
     {
       *d++ = hexval (*(s+2)) << 4 | hexval (*(s+3));
       s += 4;
@@ -84,16 +85,40 @@ static void fix_uid (char *uid)
       *d++ = *s++;
   }
   *d = '\0';
-  
-  mutt_decode_utf8_string (uid, _chs);
+
+  if (_chs && (cd = mutt_iconv_open (_chs, "utf-8", 0)) != (iconv_t)-1)
+  {
+    int n = s - uid + 1; /* chars available in original buffer */
+    char *buf;
+    ICONV_CONST char *ib;
+    char *ob;
+    size_t ibl, obl;
+
+    buf = safe_malloc (n+1);
+    ib = uid, ibl = d - uid + 1, ob = buf, obl = n;
+    iconv (cd, &ib, &ibl, &ob, &obl);
+    if (!ibl)
+    {
+      if (ob-buf < n)
+      {
+	memcpy (uid, buf, ob-buf);
+	uid[ob-buf] = '\0';
+      }
+      else if (ob-buf == n && (buf[n] = 0, strlen (buf) < n))
+	memcpy (uid, buf, n);
+    }
+    FREE (&buf);
+    iconv_close (cd);
+  }
 }
 
-static pgp_key_t *parse_pub_line (char *buf, int *is_subkey, pgp_key_t *k)
+static pgp_key_t parse_pub_line (char *buf, int *is_subkey, pgp_key_t k)
 {
   pgp_uid_t *uid = NULL;
   int field = 0, is_uid = 0;
   char *pend, *p;
   int trust = 0;
+  int flags = 0;
 
   *is_subkey = 0;
   if (!*buf)
@@ -128,26 +153,25 @@ static pgp_key_t *parse_pub_line (char *buf, int *is_subkey, pgp_key_t *k)
 	else
 	  return NULL;
 	
-	if (!is_uid)
-	  k = safe_calloc (sizeof (pgp_key_t), 1);
+	if (!(is_uid || (*is_subkey && option (OPTPGPIGNORESUB))))
+	  k = safe_calloc (sizeof *k, 1);
 
 	break;
       }
       case 2:			/* trust info */
       {
-
 	dprint (2, (debugfile, "trust info: %s\n", p));
 	
 	switch (*p)
 	{				/* look only at the first letter */
 	  case 'e':
-	    k->flags |= KEYFLAG_EXPIRED;
+	    flags |= KEYFLAG_EXPIRED;
 	    break;
 	  case 'r':
-	    k->flags |= KEYFLAG_REVOKED;
+	    flags |= KEYFLAG_REVOKED;
 	    break;
 	  case 'd':
-	    k->flags |= KEYFLAG_DISABLED;
+	    flags |= KEYFLAG_DISABLED;
 	    break;
 	  case 'n':
 	    trust = 1;
@@ -162,6 +186,10 @@ static pgp_key_t *parse_pub_line (char *buf, int *is_subkey, pgp_key_t *k)
 	    trust = 3;
 	    break;
 	}
+
+        if (!is_uid && !(*is_subkey && option (OPTPGPIGNORESUB)))
+	  k->flags |= flags;
+
 	break;
       }
       case 3:			/* key length  */
@@ -169,7 +197,8 @@ static pgp_key_t *parse_pub_line (char *buf, int *is_subkey, pgp_key_t *k)
 	
 	dprint (2, (debugfile, "key len: %s\n", p));
 	
-	k->keylen = atoi (p);	/* fixme: add validation checks */
+	if (!(*is_subkey && option (OPTPGPIGNORESUB)))
+	  k->keylen = atoi (p);	/* fixme: add validation checks */
 	break;
       }
       case 4:			/* pubkey algo */
@@ -177,17 +206,19 @@ static pgp_key_t *parse_pub_line (char *buf, int *is_subkey, pgp_key_t *k)
 	
 	dprint (2, (debugfile, "pubkey algorithm: %s\n", p));
 	
-	k->numalg = atoi (p);
-	k->algorithm = pgp_pkalgbytype (atoi (p));
-	k->flags |= pgp_get_abilities (atoi (p));
+	if (!(*is_subkey && option (OPTPGPIGNORESUB)))
+	{
+	  k->numalg = atoi (p);
+	  k->algorithm = pgp_pkalgbytype (atoi (p));
+	}
 	break;
       }
       case 5:			/* 16 hex digits with the long keyid. */
       {
 	dprint (2, (debugfile, "key id: %s\n", p));
 	
-	/* We really should do a check here */
-	k->keyid = safe_strdup (p);
+	if (!(*is_subkey && option (OPTPGPIGNORESUB)))
+	  mutt_str_replace (&k->keyid, p);
 	break;
 
       }
@@ -222,6 +253,10 @@ static pgp_key_t *parse_pub_line (char *buf, int *is_subkey, pgp_key_t *k)
       {
 	if (!pend || !*p)
 	  break;			/* empty field or no trailing colon */
+
+	/* ignore user IDs on subkeys */
+	if (!is_uid && (*is_subkey && option (OPTPGPIGNORESUB)))
+	  break;
 	
 	dprint (2, (debugfile, "user ID: %s\n", p));
 	
@@ -229,6 +264,7 @@ static pgp_key_t *parse_pub_line (char *buf, int *is_subkey, pgp_key_t *k)
 	fix_uid (p);
 	uid->addr = safe_strdup (p);
 	uid->trust = trust;
+	uid->flags |= flags;
 	uid->parent = k;
 	uid->next = k->address;
 	k->address = uid;
@@ -242,6 +278,36 @@ static pgp_key_t *parse_pub_line (char *buf, int *is_subkey, pgp_key_t *k)
       }
       case 11:			/* signature class  */
         break;
+      case 12:			/* key capabilities */
+	dprint (2, (debugfile, "capabilities info: %s\n", p));
+	
+	while(*p)
+	  {
+	    switch(*p++)
+	      {
+	      case 'D':
+		flags |= KEYFLAG_DISABLED;
+		break;
+
+	      case 'e':
+		flags |= KEYFLAG_CANENCRYPT;
+		break;
+
+	      case 's':
+		flags |= KEYFLAG_CANSIGN;
+		break;
+	      }
+	  }
+
+        if (!is_uid && 
+	    (!*is_subkey || !option (OPTPGPIGNORESUB)
+	     || !((flags & KEYFLAG_DISABLED)
+		  || (flags & KEYFLAG_REVOKED)
+		  || (flags & KEYFLAG_EXPIRED))))
+	  k->flags |= flags;
+
+	break;
+      
       default:
         break;
     }
@@ -249,19 +315,19 @@ static pgp_key_t *parse_pub_line (char *buf, int *is_subkey, pgp_key_t *k)
   return k;
 }
 
-pgp_key_t *pgp_get_candidates (pgp_ring_t keyring, LIST * hints)
+pgp_key_t pgp_get_candidates (pgp_ring_t keyring, LIST * hints)
 {
   FILE *fp;
   pid_t thepid;
   char buf[LONG_STRING];
-  pgp_key_t *db = NULL, **kend, *k = NULL, *kk, *mainkey = NULL;
+  pgp_key_t db = NULL, *kend, k = NULL, kk, mainkey = NULL;
   int is_sub;
   int devnull;
 
   if ((devnull = open ("/dev/null", O_RDWR)) == -1)
     return NULL;
 
-  _chs = mutt_get_charset (Charset);
+  mutt_str_replace (&_chs, Charset);
   
   thepid = pgp_invoke_list_keys (NULL, &fp, NULL, -1, -1, devnull,
 				 keyring, hints);
@@ -288,7 +354,7 @@ pgp_key_t *pgp_get_candidates (pgp_ring_t keyring, LIST * hints)
       if (is_sub)
       {
 	pgp_uid_t **l;
-
+	
 	k->flags  |= KEYFLAG_SUBKEY;
 	k->parent  = mainkey;
 	for (l = &k->address; *l; l = &(*l)->next)
@@ -307,7 +373,7 @@ pgp_key_t *pgp_get_candidates (pgp_ring_t keyring, LIST * hints)
   mutt_wait_filter (thepid);
 
   close (devnull);
-
+  
   return db;
 }
 
