@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 1996-2000 Michael R. Elkins <me@cs.hmc.edu>
+ * Copyright (C) 1996-2002 Michael R. Elkins <me@mutt.org>
  * 
  *     This program is free software; you can redistribute it and/or modify
  *     it under the terms of the GNU General Public License as published by
@@ -22,17 +22,13 @@
 #include "mutt_regex.h"
 #include "history.h"
 #include "keymap.h"
+#include "mbyte.h"
+#include "charset.h"
+#include "mutt_crypt.h"
+#include "mutt_idna.h"
 
-
-#ifdef HAVE_PGP
-#include "pgp.h"
-#endif
-
-
-#ifdef USE_SSL
-#include "imap.h"
-#include "imap_socket.h"
-#include "imap_ssl.h"
+#if defined(USE_SSL) || defined(USE_NSS)
+#include "mutt_ssl.h"
 #endif
 
 
@@ -41,7 +37,6 @@
 #include "init.h"
 #include "mailbox.h"
 
-#include <pwd.h>
 #include <ctype.h>
 #include <stdlib.h>
 #include <unistd.h>
@@ -106,36 +101,6 @@ int mutt_option_index (char *s)
   return (-1);
 }
 
-static void add_char (BUFFER *buf, char ch)
-{
-  size_t offset;
-
-  if (buf->dptr >= buf->data + buf->dsize)
-  {
-    offset = buf->dptr - buf->data;
-    buf->dsize += 4;
-    safe_realloc ((void **) &buf->data, buf->dsize);
-    buf->dptr = buf->data + offset;
-  }
-  *buf->dptr++ = ch;
-}
-
-static void add_str (BUFFER *buf, const char *s)
-{
-  size_t slen = mutt_strlen (s);
-  size_t offset;
-
-  if (buf->dptr + slen > buf->data + buf->dsize)
-  {
-    offset = buf->dptr - buf->data;
-    buf->dsize += slen;
-    safe_realloc ((void **) &buf->data, buf->dsize);
-    buf->dptr = buf->data + offset;
-  }
-  memcpy (buf->dptr, s, slen);
-  buf->dptr += slen;
-}
-
 int mutt_extract_token (BUFFER *dest, BUFFER *tok, int flags)
 {
   char		ch;
@@ -166,27 +131,32 @@ int mutt_extract_token (BUFFER *dest, BUFFER *tok, int flags)
       qc = ch;
     else if (ch == '\\' && qc != '\'')
     {
+	if (!*tok->dptr)
+	    return -1; /* premature end of token */
       switch (ch = *tok->dptr++)
       {
 	case 'c':
 	case 'C':
-	  add_char (dest, (toupper (*tok->dptr) - '@') & 0x7f);
+	    if (!*tok->dptr)
+		return -1; /* premature end of token */
+	  mutt_buffer_addch (dest, (toupper ((unsigned char) *tok->dptr)
+                                    - '@') & 0x7f);
 	  tok->dptr++;
 	  break;
 	case 'r':
-	  add_char (dest, '\r');
+	  mutt_buffer_addch (dest, '\r');
 	  break;
 	case 'n':
-	  add_char (dest, '\n');
+	  mutt_buffer_addch (dest, '\n');
 	  break;
 	case 't':
-	  add_char (dest, '\t');
+	  mutt_buffer_addch (dest, '\t');
 	  break;
 	case 'f':
-	  add_char (dest, '\f');
+	  mutt_buffer_addch (dest, '\f');
 	  break;
 	case 'e':
-	  add_char (dest, '\033');
+	  mutt_buffer_addch (dest, '\033');
 	  break;
 	default:
 	  if (isdigit ((unsigned char) ch) &&
@@ -194,26 +164,28 @@ int mutt_extract_token (BUFFER *dest, BUFFER *tok, int flags)
 	      isdigit ((unsigned char) *(tok->dptr + 1)))
 	  {
 
-	    add_char (dest, (ch << 6) + (*tok->dptr << 3) + *(tok->dptr + 1) - 3504);
+	    mutt_buffer_addch (dest, (ch << 6) + (*tok->dptr << 3) + *(tok->dptr + 1) - 3504);
 	    tok->dptr += 2;
 	  }
 	  else
-	    add_char (dest, ch);
+	    mutt_buffer_addch (dest, ch);
       }
     }
     else if (ch == '^' && (flags & M_TOKEN_CONDENSE))
     {
+	if (!*tok->dptr)
+	    return -1; /* premature end of token */
       ch = *tok->dptr++;
       if (ch == '^')
-	add_char (dest, ch);
+	mutt_buffer_addch (dest, ch);
       else if (ch == '[')
-	add_char (dest, '\033');
+	mutt_buffer_addch (dest, '\033');
       else if (isalpha ((unsigned char) ch))
-	add_char (dest, toupper (ch) - '@');
+	mutt_buffer_addch (dest, toupper ((unsigned char) ch) - '@');
       else
       {
-	add_char (dest, '^');
-	add_char (dest, ch);
+	mutt_buffer_addch (dest, '^');
+	mutt_buffer_addch (dest, ch);
       }
     }
     else if (ch == '`' && (!qc || qc == '"'))
@@ -243,6 +215,7 @@ int mutt_extract_token (BUFFER *dest, BUFFER *tok, int flags)
       if ((pid = mutt_create_filter (cmd, NULL, &fp, NULL)) < 0)
       {
 	dprint (1, (debugfile, "mutt_get_token: unable to fork command: %s", cmd));
+	FREE (&cmd);
 	return (-1);
       }
       FREE (&cmd);
@@ -257,13 +230,20 @@ int mutt_extract_token (BUFFER *dest, BUFFER *tok, int flags)
 
       /* if we got output, make a new string consiting of the shell ouptput
 	 plus whatever else was left on the original line */
-      if (expn.data)
+      /* BUT: If this is inside a quoted string, directly add output to 
+       * the token */
+      if (expn.data && qc)
+      {
+	mutt_buffer_addstr (dest, expn.data);
+	FREE (&expn.data);
+      }
+      else if (expn.data)
       {
 	expnlen = mutt_strlen (expn.data);
 	tok->dsize = expnlen + mutt_strlen (tok->dptr) + 1;
 	ptr = safe_malloc (tok->dsize);
 	memcpy (ptr, expn.data, expnlen);
-	strcpy (ptr + expnlen, tok->dptr);
+	strcpy (ptr + expnlen, tok->dptr);	/* __STRCPY_CHECKED__ */
 	if (tok->destroy)
 	  FREE (&tok->data);
 	tok->data = ptr;
@@ -275,7 +255,7 @@ int mutt_extract_token (BUFFER *dest, BUFFER *tok, int flags)
     }
     else if (ch == '$' && (!qc || qc == '"') && (*tok->dptr == '{' || isalpha ((unsigned char) *tok->dptr)))
     {
-      char *env, *var;
+      char *env = NULL, *var = NULL;
 
       if (*tok->dptr == '{')
       {
@@ -293,14 +273,14 @@ int mutt_extract_token (BUFFER *dest, BUFFER *tok, int flags)
 	var = mutt_substrdup (tok->dptr, pc);
 	tok->dptr = pc;
       }
-      if ((env = getenv (var)))
-	add_str (dest, env);
+      if (var && (env = getenv (var)))
+	mutt_buffer_addstr (dest, env);
       FREE (&var);
     }
     else
-      add_char (dest, ch);
+      mutt_buffer_addch (dest, ch);
   }
-  add_char (dest, 0); /* terminate the string */
+  mutt_buffer_addch (dest, 0); /* terminate the string */
   SKIPWS (tok->dptr);
   return 0;
 }
@@ -309,10 +289,14 @@ static void add_to_list (LIST **list, const char *str)
 {
   LIST *t, *last = NULL;
 
+  /* don't add a NULL or empty string to the list */
+  if (!str || *str == '\0')
+    return;
+
   /* check to make sure the item is not already on this list */
   for (last = *list; last; last = last->next)
   {
-    if (mutt_strcasecmp (str, last->data) == 0)
+    if (ascii_strcasecmp (str, last->data) == 0)
     {
       /* already on the list, so just ignore it */
       last = NULL;
@@ -336,6 +320,52 @@ static void add_to_list (LIST **list, const char *str)
   }
 }
 
+static int add_to_rx_list (RX_LIST **list, const char *s, int flags, BUFFER *err)
+{
+  RX_LIST *t, *last = NULL;
+  REGEXP *rx;
+
+  if (!s || !*s)
+    return 0;
+
+  if (!(rx = mutt_compile_regexp (s, flags)))
+  {
+    snprintf (err->data, err->dsize, "Bad regexp: %s\n", s);
+    return -1;
+  }
+
+  /* check to make sure the item is not already on this list */
+  for (last = *list; last; last = last->next)
+  {
+    if (ascii_strcasecmp (rx->pattern, last->rx->pattern) == 0)
+    {
+      /* already on the list, so just ignore it */
+      last = NULL;
+      break;
+    }
+    if (!last->next)
+      break;
+  }
+
+  if (!*list || last)
+  {
+    t = mutt_new_rx_list();
+    t->rx = rx;
+    if (last)
+    {
+      last->next = t;
+      last = last->next;
+    }
+    else
+      *list = last = t;
+  }
+  else /* duplicate */
+    mutt_free_regexp (&rx);
+
+  return 0;
+}
+
+
 static void remove_from_list (LIST **l, const char *str)
 {
   LIST *p, *last = NULL;
@@ -348,14 +378,44 @@ static void remove_from_list (LIST **l, const char *str)
     last = NULL;
     while (p)
     {
-      if (mutt_strcasecmp (str, p->data) == 0)
+      if (ascii_strcasecmp (str, p->data) == 0)
       {
-	safe_free ((void **) &p->data);
+	FREE (&p->data);
 	if (last)
 	  last->next = p->next;
 	else
 	  (*l) = p->next;
-	safe_free ((void **) &p);
+	FREE (&p);
+      }
+      else
+      {
+	last = p;
+	p = p->next;
+      }
+    }
+  }
+}
+
+static void remove_from_rx_list (RX_LIST **l, const char *str)
+{
+  RX_LIST *p, *last = NULL;
+
+  if (mutt_strcmp ("*", str) == 0)
+    mutt_free_rx_list (l);    /* ``unCMD *'' means delete all current entries */
+  else
+  {
+    p = *l;
+    last = NULL;
+    while (p)
+    {
+      if (ascii_strcasecmp (str, p->rx->pattern) == 0)
+      {
+	mutt_free_regexp (&p->rx);
+	if (last)
+	  last->next = p->next;
+	else
+	  (*l) = p->next;
+	FREE (&p);
       }
       else
       {
@@ -408,11 +468,55 @@ static int parse_list (BUFFER *buf, BUFFER *s, unsigned long data, BUFFER *err)
   return 0;
 }
 
+static int _parse_rx_list (BUFFER *buf, BUFFER *s, unsigned long data, BUFFER *err, int flags)
+{
+  do 
+  {
+    mutt_extract_token (buf, s, 0);
+    if (add_to_rx_list ((RX_LIST **) data, buf->data, flags, err) != 0)
+      return -1;
+	
+  }
+  while (MoreArgs (s));
+  
+  return 0;
+}
+
+static int parse_rx_list (BUFFER *buf, BUFFER *s, unsigned long data, BUFFER *err)
+{
+  return _parse_rx_list (buf, s, data, err, REG_ICASE);
+}
+
+static int parse_rx_unlist (BUFFER *buf, BUFFER *s, unsigned long data, BUFFER *err)
+{
+  do
+  {
+    mutt_extract_token (buf, s, 0);
+    if (mutt_strcmp (buf->data, "*") == 0)
+    {
+      mutt_free_rx_list ((RX_LIST **) data);
+      break;
+    }
+    remove_from_rx_list ((RX_LIST **) data, buf->data);
+  }
+  while (MoreArgs (s));
+  
+  return 0;
+}
+
 static int parse_unlist (BUFFER *buf, BUFFER *s, unsigned long data, BUFFER *err)
 {
   do
   {
     mutt_extract_token (buf, s, 0);
+    /*
+     * Check for deletion of entire list
+     */
+    if (mutt_strcmp (buf->data, "*") == 0)
+    {
+      mutt_free_list ((LIST **) data);
+      break;
+    }
     remove_from_list ((LIST **) data, buf->data);
   }
   while (MoreArgs (s));
@@ -426,8 +530,8 @@ static int parse_unlists (BUFFER *buf, BUFFER *s, unsigned long data, BUFFER *er
   do
   {
     mutt_extract_token (buf, s, 0);
-    remove_from_list (&MailLists, buf->data);
-    remove_from_list (&SubscribedLists, buf->data);
+    remove_from_rx_list (&MailLists, buf->data);
+    remove_from_rx_list (&SubscribedLists, buf->data);
   }
   while (MoreArgs (s));
 
@@ -439,8 +543,10 @@ static int parse_subscribe (BUFFER *buf, BUFFER *s, unsigned long data, BUFFER *
   do
   {
     mutt_extract_token (buf, s, 0);
-    add_to_list (&MailLists, buf->data);
-    add_to_list (&SubscribedLists, buf->data);
+    if (add_to_rx_list (&MailLists, buf->data, REG_ICASE, err) != 0)
+      return -1;
+    if (add_to_rx_list (&SubscribedLists, buf->data, REG_ICASE, err) != 0)
+      return -1;
   }
   while (MoreArgs (s));
 
@@ -455,21 +561,40 @@ static int parse_unalias (BUFFER *buf, BUFFER *s, unsigned long data, BUFFER *er
   {
     mutt_extract_token (buf, s, 0);
 
-    tmp = Aliases;
-    for (tmp = Aliases; tmp; tmp = tmp->next)
+    if (mutt_strcmp ("*", buf->data) == 0)
     {
-      if (mutt_strcasecmp (buf->data, tmp->name) == 0)
+      if (CurrentMenu == MENU_ALIAS)
       {
-	if (last)
-	  last->next = tmp->next;
-	else
-	  Aliases = tmp->next;
-	tmp->next = NULL;
-	mutt_free_alias (&tmp);
-	break;
+	for (tmp = Aliases; tmp ; tmp = tmp->next)
+	  tmp->del = 1;
+	set_option (OPTFORCEREDRAWINDEX);
       }
-      last = tmp;
+      else
+	mutt_free_alias (&Aliases);
+      break;
     }
+    else
+      for (tmp = Aliases; tmp; tmp = tmp->next)
+      {
+	if (mutt_strcasecmp (buf->data, tmp->name) == 0)
+	{
+	  if (CurrentMenu == MENU_ALIAS)
+	  {
+	    tmp->del = 1;
+	    set_option (OPTFORCEREDRAWINDEX);
+	    break;
+	  }
+
+	  if (last)
+	    last->next = tmp->next;
+	  else
+	    Aliases = tmp->next;
+	  tmp->next = NULL;
+	  mutt_free_alias (&tmp);
+	  break;
+	}
+	last = tmp;
+      }
   }
   while (MoreArgs (s));
   return 0;
@@ -479,21 +604,20 @@ static int parse_alias (BUFFER *buf, BUFFER *s, unsigned long data, BUFFER *err)
 {
   ALIAS *tmp = Aliases;
   ALIAS *last = NULL;
-  char *p;
-  size_t len;
-
-  if ((p = strpbrk (s->dptr, " \t")) == NULL)
+  char *estr = NULL;
+  
+  if (!MoreArgs (s))
   {
     strfcpy (err->data, _("alias: no address"), err->dsize);
     return (-1);
   }
 
-  len = p - s->dptr;
+  mutt_extract_token (buf, s, 0);
 
   /* check to see if an alias with this name already exists */
   for (; tmp; tmp = tmp->next)
   {
-    if (!mutt_strncasecmp (tmp->name, s->dptr, len) && *(tmp->name + len) == 0)
+    if (!mutt_strcasecmp (tmp->name, buf->data))
       break;
     last = tmp;
   }
@@ -502,16 +626,19 @@ static int parse_alias (BUFFER *buf, BUFFER *s, unsigned long data, BUFFER *err)
   {
     /* create a new alias */
     tmp = (ALIAS *) safe_calloc (1, sizeof (ALIAS));
-    tmp->name = safe_malloc (len + 1);
-    memcpy (tmp->name, s->dptr, len);
-    tmp->name[len] = 0;
+    tmp->self = tmp;
+    tmp->name = safe_strdup (buf->data);
+    /* give the main addressbook code a chance */
+    if (CurrentMenu == MENU_ALIAS)
+      set_option (OPTMENUCALLER);
   }
   else
   {
     /* override the previous value */
     rfc822_free_address (&tmp->addr);
+    if (CurrentMenu == MENU_ALIAS)
+      set_option (OPTFORCEREDRAWINDEX);
   }
-  s->dptr = p;
 
   mutt_extract_token (buf, s, M_TOKEN_QUOTE | M_TOKEN_SPACE | M_TOKEN_SEMICOLON);
   tmp->addr = mutt_parse_adrlist (tmp->addr, buf->data);
@@ -519,6 +646,12 @@ static int parse_alias (BUFFER *buf, BUFFER *s, unsigned long data, BUFFER *err)
     last->next = tmp;
   else
     Aliases = tmp;
+  if (mutt_addrlist_to_idna (tmp->addr, &estr))
+  {
+    snprintf (err->data, err->dsize, _("Warning: Bad IDN '%s' in alias '%s'.\n"),
+	      estr, tmp->name);
+    return -1;
+  }
   return 0;
 }
 
@@ -546,7 +679,7 @@ parse_unmy_hdr (BUFFER *buf, BUFFER *s, unsigned long data, BUFFER *err)
 
       while (tmp)
       {
-	if (mutt_strncasecmp (buf->data, tmp->data, l) == 0 && tmp->data[l] == ':')
+	if (ascii_strncasecmp (buf->data, tmp->data, l) == 0 && tmp->data[l] == ':')
 	{
 	  ptr = tmp;
 	  if (last)
@@ -588,10 +721,10 @@ static int parse_my_hdr (BUFFER *buf, BUFFER *s, unsigned long data, BUFFER *err
     for (tmp = UserHeader; ; tmp = tmp->next)
     {
       /* see if there is already a field by this name */
-      if (mutt_strncasecmp (buf->data, tmp->data, keylen) == 0)
+      if (ascii_strncasecmp (buf->data, tmp->data, keylen) == 0)
       {
 	/* replace the old value */
-	safe_free ((void **) &tmp->data);
+	FREE (&tmp->data);
 	tmp->data = buf->data;
 	memset (buf, 0, sizeof (BUFFER));
 	return 0;
@@ -640,6 +773,41 @@ parse_sort (short *val, const char *s, const struct mapping_t *map, BUFFER *err)
   return 0;
 }
 
+static void mutt_set_default (struct option_t *p)
+{
+  switch (p->type & DT_MASK)
+  {
+    case DT_STR:
+      if (!p->init && *((char **) p->data))
+        p->init = (unsigned long) safe_strdup (* ((char **) p->data));
+      break;
+    case DT_PATH:
+      if (!p->init && *((char **) p->data))
+      {
+	char *cp = safe_strdup (*((char **) p->data));
+	/* mutt_pretty_mailbox (cp); */
+        p->init = (unsigned long) cp;
+      }
+      break;
+    case DT_ADDR:
+      if (!p->init && *((ADDRESS **) p->data))
+      {
+	char tmp[HUGE_STRING];
+	*tmp = '\0';
+	rfc822_write_address (tmp, sizeof (tmp), *((ADDRESS **) p->data), 0);
+	p->init = (unsigned long) safe_strdup (tmp);
+      }
+      break;
+    case DT_RX:
+    {
+      REGEXP *pp = (REGEXP *) p->data;
+      if (!p->init && pp->pattern)
+	p->init = (unsigned long) safe_strdup (pp->pattern);
+      break;
+    }
+  }
+}
+
 static void mutt_restore_default (struct option_t *p)
 {
   switch (p->type & DT_MASK)
@@ -690,6 +858,7 @@ static void mutt_restore_default (struct option_t *p)
 	  regfree (pp->rx);
 	  FREE (&pp->rx);
 	}
+
 	if (p->init)
 	{
 	  char *s = (char *) p->init;
@@ -707,8 +876,8 @@ static void mutt_restore_default (struct option_t *p)
 	  }
 	  if (REGCOMP (pp->rx, s, flags) != 0)
 	  {
-	    fprintf (stderr, _("mutt_restore_default: error in regexp: %s\n"),
-		     pp->pattern);
+	    fprintf (stderr, _("mutt_restore_default(%s): error in regexp: %s\n"),
+		     p->option, pp->pattern);
 	    FREE (&pp->pattern);
 	    regfree (pp->rx);
 	    FREE (&pp->rx);
@@ -717,6 +886,7 @@ static void mutt_restore_default (struct option_t *p)
       }
       break;
   }
+
   if (p->flags & R_INDEX)
     set_option (OPTFORCEREDRAWINDEX);
   if (p->flags & R_PAGER)
@@ -725,6 +895,10 @@ static void mutt_restore_default (struct option_t *p)
     set_option (OPTSORTSUBTHREADS);
   if (p->flags & R_RESORT)
     set_option (OPTNEEDRESORT);
+  if (p->flags & R_RESORT_INIT)
+    set_option (OPTRESORTINIT);
+  if (p->flags & R_TREE)
+    set_option (OPTREDRAWTREE);
 }
 
 static int parse_set (BUFFER *tmp, BUFFER *s, unsigned long data, BUFFER *err)
@@ -807,9 +981,9 @@ static int parse_set (BUFFER *tmp, BUFFER *s, unsigned long data, BUFFER *err)
 
 	s->dptr++;
 	mutt_extract_token (tmp, s, 0);
-	if (mutt_strcasecmp ("yes", tmp->data) == 0)
+	if (ascii_strcasecmp ("yes", tmp->data) == 0)
 	  unset = inv = 0;
-	else if (mutt_strcasecmp ("no", tmp->data) == 0)
+	else if (ascii_strcasecmp ("no", tmp->data) == 0)
 	  unset = 1;
 	else
 	{
@@ -841,7 +1015,7 @@ static int parse_set (BUFFER *tmp, BUFFER *s, unsigned long data, BUFFER *err)
 	if (DTYPE (MuttVars[idx].type) == DT_ADDR)
 	  rfc822_free_address ((ADDRESS **) MuttVars[idx].data);
 	else
-	  safe_free ((void **) MuttVars[idx].data);
+	  FREE (MuttVars[idx].data);
       }
       else if (query || *s->dptr != '=')
       {
@@ -851,7 +1025,7 @@ static int parse_set (BUFFER *tmp, BUFFER *s, unsigned long data, BUFFER *err)
 	if (DTYPE (MuttVars[idx].type) == DT_ADDR)
 	{
 	  _tmp[0] = '\0';
-	  rfc822_write_address (_tmp, sizeof (_tmp), *((ADDRESS **) MuttVars[idx].data));
+	  rfc822_write_address (_tmp, sizeof (_tmp), *((ADDRESS **) MuttVars[idx].data), 0);
 	  val = _tmp;
 	}
 	else
@@ -882,6 +1056,8 @@ static int parse_set (BUFFER *tmp, BUFFER *s, unsigned long data, BUFFER *err)
         else if (DTYPE (MuttVars[idx].type) == DT_STR)
         {
 	  *((char **) MuttVars[idx].data) = safe_strdup (tmp->data);
+	  if (mutt_strcmp (MuttVars[idx].option, "charset") == 0)
+	    mutt_set_charset (Charset);
         }
         else
         {
@@ -1002,9 +1178,6 @@ static int parse_set (BUFFER *tmp, BUFFER *s, unsigned long data, BUFFER *err)
 	  case M_MMDF:
 	    p = "MMDF";
 	    break;
-	  case M_KENDRA:
-	    p = "KENDRA";
-	    break;
 	  case M_MH:
 	    p = "MH";
 	    break;
@@ -1085,13 +1258,13 @@ static int parse_set (BUFFER *tmp, BUFFER *s, unsigned long data, BUFFER *err)
       {
 	s->dptr++;
 	mutt_extract_token (tmp, s, 0);
-	if (mutt_strcasecmp ("yes", tmp->data) == 0)
+	if (ascii_strcasecmp ("yes", tmp->data) == 0)
 	  set_quadoption (MuttVars[idx].data, M_YES);
-	else if (mutt_strcasecmp ("no", tmp->data) == 0)
+	else if (ascii_strcasecmp ("no", tmp->data) == 0)
 	  set_quadoption (MuttVars[idx].data, M_NO);
-	else if (mutt_strcasecmp ("ask-yes", tmp->data) == 0)
+	else if (ascii_strcasecmp ("ask-yes", tmp->data) == 0)
 	  set_quadoption (MuttVars[idx].data, M_ASKYES);
-	else if (mutt_strcasecmp ("ask-no", tmp->data) == 0)
+	else if (ascii_strcasecmp ("ask-no", tmp->data) == 0)
 	  set_quadoption (MuttVars[idx].data, M_ASKNO);
 	else
 	{
@@ -1112,7 +1285,7 @@ static int parse_set (BUFFER *tmp, BUFFER *s, unsigned long data, BUFFER *err)
     }
     else if (DTYPE (MuttVars[idx].type) == DT_SORT)
     {
-      const struct mapping_t *map;
+      const struct mapping_t *map = NULL;
 
       switch (MuttVars[idx].type & DT_SUBTYPE_MASK)
       {
@@ -1122,11 +1295,10 @@ static int parse_set (BUFFER *tmp, BUFFER *s, unsigned long data, BUFFER *err)
 	case DT_SORT_BROWSER:
 	  map = SortBrowserMethods;
 	  break;
-#ifdef HAVE_PGP
 	case DT_SORT_KEYS:
-	  map = SortKeyMethods;
+          if ((WithCrypto & APPLICATION_PGP))
+            map = SortKeyMethods;
 	  break;
-#endif
 	case DT_SORT_AUX:
 	  map = SortAuxMethods;
 	  break;
@@ -1135,6 +1307,13 @@ static int parse_set (BUFFER *tmp, BUFFER *s, unsigned long data, BUFFER *err)
 	  break;
       }
 
+      if (!map)
+      {
+	snprintf (err->data, err->dsize, _("%s: Unknown type."), MuttVars[idx].option);
+	r = -1;
+	break;
+      }
+      
       if (query || *s->dptr != '=')
       {
 	p = mutt_getnamebyvalue (*((short *) MuttVars[idx].data) & SORT_MASK, map);
@@ -1149,7 +1328,10 @@ static int parse_set (BUFFER *tmp, BUFFER *s, unsigned long data, BUFFER *err)
       mutt_extract_token (tmp, s , 0);
 
       if (parse_sort ((short *) MuttVars[idx].data, tmp->data, map, err) == -1)
+      {
+	r = -1;
 	break;
+      }
     }
     else
     {
@@ -1166,6 +1348,10 @@ static int parse_set (BUFFER *tmp, BUFFER *s, unsigned long data, BUFFER *err)
       set_option (OPTSORTSUBTHREADS);
     if (MuttVars[idx].flags & R_RESORT)
       set_option (OPTNEEDRESORT);
+    if (MuttVars[idx].flags & R_RESORT_INIT)
+      set_option (OPTRESORTINIT);
+    if (MuttVars[idx].flags & R_TREE)
+      set_option (OPTREDRAWTREE);
   }
   return (r);
 }
@@ -1182,6 +1368,18 @@ static int source_rc (const char *rcfile, BUFFER *err)
   char *linebuf = NULL;
   size_t buflen;
   pid_t pid;
+  struct stat s;
+
+  if (stat (rcfile, &s) < 0)
+  {
+    snprintf (err->data, err->dsize, _("%s: stat: %s"), rcfile, strerror (errno));
+    return (-1);
+  }
+  if (!S_ISREG (s.st_mode))
+  {
+    snprintf (err->data, err->dsize, _("%s: not a regular file"), rcfile);
+    return (-1);
+  }
 
   if ((f = mutt_open_read (rcfile, &pid)) == NULL)
   {
@@ -1203,7 +1401,7 @@ static int source_rc (const char *rcfile, BUFFER *err)
         rc = -1;
   }
   FREE (&token.data);
-  safe_free ((void **) &linebuf);
+  FREE (&linebuf);
   fclose (f);
   if (pid != -1)
     mutt_wait_filter (pid);
@@ -1293,7 +1491,6 @@ finish:
 }
 
 
-#define MAX(a,b) (((a) >= (b)) ? (a) : (b))
 #define NUMVARS (sizeof (MuttVars)/sizeof (MuttVars[0]))
 #define NUMCOMMANDS (sizeof (Commands)/sizeof (Commands[0]))
 /* initial string that starts completion. No telling how much crap 
@@ -1524,7 +1721,7 @@ int mutt_var_value_complete (char *buffer, size_t len, int pos)
       }
       else if (DTYPE (MuttVars[idx].type) == DT_ADDR)
       {
-	rfc822_write_address (tmp, sizeof (tmp), *((ADDRESS **) MuttVars[idx].data));
+	rfc822_write_address (tmp, sizeof (tmp), *((ADDRESS **) MuttVars[idx].data), 0);
       }
       else if (DTYPE (MuttVars[idx].type) == DT_QUAD)
 	strfcpy (tmp, vals[quadoption (MuttVars[idx].data)], sizeof (tmp));
@@ -1543,11 +1740,12 @@ int mutt_var_value_complete (char *buffer, size_t len, int pos)
 	  case DT_SORT_BROWSER:
 	    map = SortBrowserMethods;
 	    break;
-#ifdef HAVE_PGP
 	  case DT_SORT_KEYS:
-	    map = SortKeyMethods;
+            if ((WithCrypto & APPLICATION_PGP))
+              map = SortKeyMethods;
+            else
+              map = SortMethods;
 	    break;
-#endif
 	  default:
 	    map = SortMethods;
 	    break;
@@ -1580,6 +1778,38 @@ int mutt_var_value_complete (char *buffer, size_t len, int pos)
   return 0;
 }
 
+/* Implement the -Q command line flag */
+int mutt_query_variables (LIST *queries)
+{
+  LIST *p;
+  
+  char errbuff[STRING];
+  char command[STRING];
+  
+  BUFFER err, token;
+  
+  memset (&err, 0, sizeof (err));
+  memset (&token, 0, sizeof (token));
+  
+  err.data = errbuff;
+  err.dsize = sizeof (errbuff);
+  
+  for (p = queries; p; p = p->next)
+  {
+    snprintf (command, sizeof (command), "set ?%s\n", p->data);
+    if (mutt_parse_rc_line (command, &token, &err) == -1)
+    {
+      fprintf (stderr, "%s\n", err.data);
+      FREE (&token.data);
+      return 1;
+    }
+    printf ("%s\n", err.data);
+  }
+  
+  FREE (&token.data);
+  return 0;
+}
+
 char *mutt_getnamebyvalue (int val, const struct mapping_t *map)
 {
   int i;
@@ -1595,7 +1825,7 @@ int mutt_getvaluebyname (const char *name, const struct mapping_t *map)
   int i;
 
   for (i = 0; map[i].name; i++)
-    if (mutt_strcasecmp (map[i].name, name) == 0)
+    if (ascii_strcasecmp (map[i].name, name) == 0)
       return (map[i].value);
   return (-1);
 }
@@ -1659,6 +1889,12 @@ void mutt_init (int skip_sys_rc, LIST *commands)
   err.data = error;
   err.dsize = sizeof (error);
 
+  /* 
+   * XXX - use something even more difficult to predict?
+   */
+  snprintf (AttachmentMarker, sizeof (AttachmentMarker),
+	    "\033]9;%ld\a", (long) time (NULL));
+  
   /* on one of the systems I use, getcwd() does not return the same prefix
      as is listed in the passwd file */
   if ((p = getenv ("HOME")))
@@ -1667,13 +1903,13 @@ void mutt_init (int skip_sys_rc, LIST *commands)
   /* Get some information about the user */
   if ((pw = getpwuid (getuid ())))
   {
+    char rnbuf[STRING];
+
     Username = safe_strdup (pw->pw_name);
     if (!Homedir)
       Homedir = safe_strdup (pw->pw_dir);
-    if ((p = strchr (pw->pw_gecos, ',')))
-      Realname = mutt_substrdup (pw->pw_gecos, p);
-    else
-      Realname = safe_strdup (pw->pw_gecos);
+
+    Realname = safe_strdup (mutt_gecos_name (rnbuf, sizeof (rnbuf), pw));
     Shell = safe_strdup (pw->pw_shell);
   }
   else 
@@ -1722,19 +1958,21 @@ void mutt_init (int skip_sys_rc, LIST *commands)
     if (*DOMAIN != '@')
   {
     Fqdn = safe_malloc (mutt_strlen (DOMAIN) + mutt_strlen (Hostname) + 2);
-    sprintf (Fqdn, "%s.%s", NONULL(Hostname), DOMAIN);
+    sprintf (Fqdn, "%s.%s", NONULL(Hostname), DOMAIN);	/* __SPRINTF_CHECKED__ */
   }
   else
     Fqdn = safe_strdup(NONULL(Hostname));
 
   if ((p = getenv ("MAIL")))
     Spoolfile = safe_strdup (p);
+  else if ((p = getenv ("MAILDIR")))
+    Spoolfile = safe_strdup (p);
   else
   {
 #ifdef HOMESPOOL
-    snprintf (buffer, sizeof (buffer), "%s/%s", NONULL(Homedir), MAILPATH);
+    mutt_concat_path (buffer, NONULL (Homedir), MAILPATH, sizeof (buffer));
 #else
-    snprintf (buffer, sizeof (buffer), "%s/%s", MAILPATH, NONULL(Username));
+    mutt_concat_path (buffer, MAILPATH, NONULL(Username), sizeof (buffer));
 #endif
     Spoolfile = safe_strdup (buffer);
   }
@@ -1744,17 +1982,20 @@ void mutt_init (int skip_sys_rc, LIST *commands)
   else
   {
     /* Default search path from RFC1524 */
-    MailcapPath = safe_strdup ("~/.mailcap:" SHAREDIR "/mailcap:" SYSCONFDIR "/mailcap:/etc/mailcap:/usr/etc/mailcap:/usr/local/etc/mailcap");
+    MailcapPath = safe_strdup ("~/.mailcap:" PKGDATADIR "/mailcap:" SYSCONFDIR "/mailcap:/etc/mailcap:/usr/etc/mailcap:/usr/local/etc/mailcap");
   }
 
   Tempdir = safe_strdup ((p = getenv ("TMPDIR")) ? p : "/tmp");
 
-#ifdef USE_POP
-  PopUser = safe_strdup (Username);
-#endif
-
-  Editor = safe_strdup ((p = getenv ("EDITOR")) ? p : "vi");
-  Visual = safe_strdup ((p = getenv ("VISUAL")) ? p : Editor);
+  p = getenv ("VISUAL");
+  if (!p)
+  {
+    p = getenv ("EDITOR");
+    if (!p)
+      p = "vi";
+  }
+  Editor = safe_strdup (p);
+  Visual = safe_strdup (p);
 
   if ((p = getenv ("REPLYTO")) != NULL)
   {
@@ -1771,12 +2012,23 @@ void mutt_init (int skip_sys_rc, LIST *commands)
     FREE (&token.data);
   }
 
+  if ((p = getenv ("EMAIL")) != NULL)
+    From = rfc822_parse_adrlist (NULL, p);
+  
+  mutt_set_langinfo_charset ();
+  mutt_set_charset (Charset);
+  
+  
   /* Set standard defaults */
   for (i = 0; MuttVars[i].option; i++)
+  {
+    mutt_set_default (&MuttVars[i]);
     mutt_restore_default (&MuttVars[i]);
+  }
 
   CurrentMenu = MENU_MAIN;
 
+
 #ifndef LOCALES_HACK
   /* Do we have a locale definition? */
   if (((p = getenv ("LC_ALL")) != NULL && p[0]) ||
@@ -1785,6 +2037,12 @@ void mutt_init (int skip_sys_rc, LIST *commands)
     set_option (OPTLOCALES);
 #endif
 
+#ifdef HAVE_GETSID
+  /* Unset suspend by default if we're the session leader */
+  if (getsid(0) == getpid())
+    unset_option (OPTSUSPEND);
+#endif
+
   mutt_init_history ();
 
   
@@ -1834,9 +2092,9 @@ void mutt_init (int skip_sys_rc, LIST *commands)
     if (access (buffer, F_OK) == -1)
       snprintf (buffer, sizeof(buffer), "%s/Muttrc", SYSCONFDIR);
     if (access (buffer, F_OK) == -1)
-      snprintf (buffer, sizeof (buffer), "%s/Muttrc-%s", SHAREDIR, MUTT_VERSION);
+      snprintf (buffer, sizeof (buffer), "%s/Muttrc-%s", PKGDATADIR, MUTT_VERSION);
     if (access (buffer, F_OK) == -1)
-      snprintf (buffer, sizeof (buffer), "%s/Muttrc", SHAREDIR);
+      snprintf (buffer, sizeof (buffer), "%s/Muttrc", PKGDATADIR);
     if (access (buffer, F_OK) != -1)
     {
       if (source_rc (buffer, &err) != 0)
@@ -1887,7 +2145,7 @@ int mutt_get_hook_type (const char *name)
   struct command_t *c;
 
   for (c = Commands ; c->name ; c++)
-    if (c->func == mutt_parse_hook && mutt_strcasecmp (c->name, name) == 0)
+    if (c->func == mutt_parse_hook && ascii_strcasecmp (c->name, name) == 0)
       return c->data;
   return 0;
 }
