@@ -1,6 +1,6 @@
 /*
- * Copyright (C) 1996-2000 Michael R. Elkins <me@cs.hmc.edu>
- * Copyright (C) 1999-2000 Thomas Roessler <roessler@guug.de>
+ * Copyright (C) 1996-2000 Michael R. Elkins <me@mutt.org>
+ * Copyright (C) 1999-2000 Thomas Roessler <roessler@does-not-exist.org>
  * 
  *     This program is free software; you can redistribute it
  *     and/or modify it under the terms of the GNU General Public
@@ -54,6 +54,14 @@ void *safe_calloc (size_t nmemb, size_t size)
 
   if (!nmemb || !size)
     return NULL;
+
+  if (((size_t) -1) / nmemb <= size)
+  {
+    mutt_error _("Integer overflow -- can't allocate memory!");
+    sleep (1);
+    mutt_exit (1);
+  }
+  
   if (!(p = calloc (nmemb, size)))
   {
     mutt_error _("Out of memory!");
@@ -78,9 +86,10 @@ void *safe_malloc (size_t siz)
   return (p);
 }
 
-void safe_realloc (void **p, size_t siz)
+void safe_realloc (void *ptr, size_t siz)
 {
   void *r;
+  void **p = (void **)ptr;
 
   if (siz == 0)
   {
@@ -110,8 +119,9 @@ void safe_realloc (void **p, size_t siz)
   *p = r;
 }
 
-void safe_free (void **p)
+void safe_free (void *ptr)
 {
+  void **p = (void **)ptr;
   if (*p)
   {
     free (*p);				/* __MEM_CHECKED__ */
@@ -145,14 +155,14 @@ char *safe_strdup (const char *s)
 
 void mutt_str_replace (char **p, const char *s)
 {
-  safe_free ((void **) p);
+  FREE (p);
   *p = safe_strdup (s);
 }
 
 void mutt_str_adjust (char **p)
 {
   if (!p || !*p) return;
-  safe_realloc ((void **) p, strlen (*p) + 1);
+  safe_realloc (p, strlen (*p) + 1);
 }
 
 /* convert all characters in the string to lowercase */
@@ -162,7 +172,7 @@ char *mutt_strlower (char *s)
 
   while (*p)
   {
-    *p = tolower (*p);
+    *p = tolower ((unsigned char) *p);
     p++;
   }
 
@@ -183,8 +193,8 @@ void mutt_unlink (const char *s)
       memset (buf, 0, sizeof (buf));
       while (sb.st_size > 0)
       {
-	fwrite (buf, 1, sizeof (buf), f);
-	sb.st_size -= sizeof (buf);
+	fwrite (buf, 1, MIN (sizeof (buf), sb.st_size), f);
+	sb.st_size -= MIN (sizeof (buf), sb.st_size);
       }
       fclose (f);
     }
@@ -451,7 +461,7 @@ char *mutt_read_line (char *s, size_t *size, FILE *fp, int *line)
   {
     if (fgets (s + offset, *size - offset, fp) == NULL)
     {
-      safe_free ((void **) &s);
+      FREE (&s);
       return NULL;
     }
     if ((ch = strchr (s + offset, '\n')) != NULL)
@@ -484,7 +494,7 @@ char *mutt_read_line (char *s, size_t *size, FILE *fp, int *line)
         /* There wasn't room for the line -- increase ``s'' */
         offset = *size - 1; /* overwrite the terminating 0 */
         *size += STRING;
-        safe_realloc ((void **) &s, *size);
+        safe_realloc (&s, *size);
       }
     }
   }
@@ -584,6 +594,11 @@ size_t mutt_strlen(const char *a)
   return a ? strlen (a) : 0;
 }
 
+int mutt_strcoll(const char *a, const char *b)
+{
+  return strcoll(NONULL(a), NONULL(b));
+}
+
 const char *mutt_stristr (const char *haystack, const char *needle)
 {
   const char *p, *q;
@@ -595,7 +610,10 @@ const char *mutt_stristr (const char *haystack, const char *needle)
 
   while (*(p = haystack))
   {
-    for (q = needle; *p && *q && tolower (*p) == tolower (*q); p++, q++)
+    for (q = needle;
+         *p && *q &&
+           tolower ((unsigned char) *p) == tolower ((unsigned char) *q);
+         p++, q++)
       ;
     if (!*q)
       return (haystack);
@@ -617,3 +635,23 @@ void mutt_remove_trailing_ws (char *s)
   for (p = s + mutt_strlen (s) - 1 ; p >= s && ISSPACE (*p) ; p--)
     *p = 0;
 }
+
+char *mutt_concat_path (char *d, const char *dir, const char *fname, size_t l)
+{
+  const char *fmt = "%s/%s";
+  
+  if (!*fname || (*dir && dir[strlen(dir)-1] == '/'))
+    fmt = "%s%s";
+  
+  snprintf (d, l, fmt, dir, fname);
+  return d;
+}
+
+const char *mutt_basename (const char *f)
+{
+  const char *p = strrchr (f, '/');
+  if (p)
+    return p + 1;
+  else
+    return f;
+}
