@@ -1,6 +1,6 @@
 /*
- * Copyright (C) 1996-2000 Michael R. Elkins <me@cs.hmc.edu>
- * Copyright (C) 1999-2000 Thomas Roessler <roessler@guug.de>
+ * Copyright (C) 1996-2000 Michael R. Elkins <me@mutt.org>
+ * Copyright (C) 1999-2000 Thomas Roessler <roessler@does-not-exist.org>
  * 
  *     This program is free software; you can redistribute it
  *     and/or modify it under the terms of the GNU General Public
@@ -45,7 +45,7 @@
 
 # ifdef ENABLE_NLS
 #  include <libintl.h>
-#  define _(a) (gettext (a))
+# define _(a) (gettext (a))
 #  ifdef gettext_noop
 #   define N_(a) gettext_noop (a)
 #  else
@@ -64,11 +64,29 @@
 # define STRING          256
 # define SHORT_STRING    128
 
-# define FREE(x) safe_free((void **)x)
+/*
+ * Create a format string to be used with scanf.
+ * To use it, write, for instance, MUTT_FORMAT(HUGE_STRING).
+ * 
+ * See K&R 2nd ed, p. 231 for an explanation.
+ */
+# define _MUTT_FORMAT_2(a,b)	"%" a  b
+# define _MUTT_FORMAT_1(a, b)	_MUTT_FORMAT_2(#a, b)
+# define MUTT_FORMAT(a)		_MUTT_FORMAT_1(a, "s")
+# define MUTT_FORMAT2(a,b)	_MUTT_FORMAT_1(a, b)
+
+
+# define FREE(x) safe_free(x)
 # define NONULL(x) x?x:""
 # define ISSPACE(c) isspace((unsigned char)c)
 # define strfcpy(A,B,C) strncpy(A,B,C), *(A+(C)-1)=0
 
+# undef MAX
+# undef MIN
+# define MAX(a,b) ((a) < (b) ? (b) : (a))
+# define MIN(a,b) ((a) < (b) ? (a) : (b))
+
+
 #define FOREVER while (1)
 
 /* this macro must check for *c == 0 since isspace(0) has unreliable behavior
@@ -91,6 +109,7 @@ void mutt_exit (int);
 
 FILE *safe_fopen (const char *, const char *);
 
+char *mutt_concat_path (char *, const char *, const char *, size_t);
 char *mutt_read_line (char *, size_t *, FILE *, int *);
 char *mutt_skip_whitespace (char *);
 char *mutt_strlower (char *);
@@ -99,6 +118,7 @@ char *mutt_substrdup (const char *, const char *);
 char *safe_strdup (const char *);
 
 const char *mutt_stristr (const char *, const char *);
+const char *mutt_basename (const char *);
 
 int mutt_copy_stream (FILE *, FILE *);
 int mutt_copy_bytes (FILE *, FILE *, size_t);
@@ -107,6 +127,7 @@ int mutt_strcasecmp (const char *, const char *);
 int mutt_strcmp (const char *, const char *);
 int mutt_strncasecmp (const char *, const char *, size_t);
 int mutt_strncmp (const char *, const char *, size_t);
+int mutt_strcoll (const char *, const char *);
 int safe_open (const char *, int);
 int safe_symlink (const char *, const char *);
 int safe_rename (const char *, const char *);
@@ -120,8 +141,10 @@ void *safe_malloc (size_t);
 void mutt_nocurses_error (const char *, ...);
 void mutt_remove_trailing_ws (char *);
 void mutt_sanitize_filename (char *, short);
+void mutt_str_replace (char **p, const char *s);
+void mutt_str_adjust (char **p);
 void mutt_unlink (const char *);
-void safe_free (void **);
-void safe_realloc (void **, size_t);
+void safe_free (void *);
+void safe_realloc (void *, size_t);
 
 #endif
