@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 1996-2000 Michael R. Elkins <me@cs.hmc.edu>
+ * Copyright (C) 1996-2002 Michael R. Elkins <me@mutt.org>
  * 
  *     This program is free software; you can redistribute it and/or modify
  *     it under the terms of the GNU General Public License as published by
@@ -16,6 +16,9 @@
  *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
  */ 
 
+#ifndef MUTT_H
+#define MUTT_H 
+
 #include "config.h"
 
 #include <stdio.h>
@@ -50,6 +53,13 @@
 #include "hash.h"
 #include "charset.h"
 
+#ifndef HAVE_WC_FUNCS
+# ifdef MB_LEN_MAX
+#  undef MB_LEN_MAX
+# endif
+# define MB_LEN_MAX 16
+#endif
+
 #ifdef SUBVERSION
 # define MUTT_VERSION (VERSION SUBVERSION)
 #else  
@@ -65,6 +75,8 @@
 #define INITVAL(x) 
 #endif
 
+#include "mutt_regex.h"
+
 /* flags for mutt_copy_header() */
 #define CH_UPDATE	1      /* update the status and x-status fields? */
 #define CH_WEED		(1<<1) /* weed the headers? */
@@ -81,6 +93,7 @@
 #define CH_NOLEN	(1<<12) /* don't write Content-Length: and Lines: */
 #define CH_WEED_DELIVERED (1<<13) /* weed eventual Delivered-To headers */
 #define CH_FORCE_FROM	(1<<14)	/* give CH_FROM precedence over CH_WEED? */
+#define CH_NOQFROM	(1<<15)	/* give CH_FROM precedence over CH_WEED? */
 
 /* flags for mutt_enter_string() */
 #define  M_ALIAS   1      /* do alias "completion" by calling up the alias-menu */
@@ -101,6 +114,9 @@
 #define M_TOKEN_COMMENT		(1<<5)	/* don't reap comments */
 #define M_TOKEN_SEMICOLON	(1<<6)	/* don't treat ; as special */
 
+/* flags for km_dokey() */
+#define M_KM_UNBUFFERED 1 /* don't read from the key buffer */
+
 typedef struct
 {
   char *data;	/* pointer to data */
@@ -139,10 +155,9 @@ typedef enum
 #define M_CHARSETHOOK	(1<<5)
 #define M_ICONVHOOK 	(1<<6)
 #define M_MESSAGEHOOK	(1<<7)
-#ifdef HAVE_PGP
-#define M_PGPHOOK	(1<<8)
-#endif
+#define M_CRYPTHOOK	(1<<8)
 #define M_ACCOUNTHOOK	(1<<9)
+#define M_REPLYHOOK	(1<<10)
 
 /* tree characters for linearize_tree and print_enriched_string */
 #define M_TREE_LLCORNER		1
@@ -216,11 +231,10 @@ enum
   M_PERSONAL_RECIP,
   M_PERSONAL_FROM,
   M_ADDRESS,
-#ifdef HAVE_PGP
-  M_PGP_SIGN,
-  M_PGP_ENCRYPT,
+  M_CRYPT_SIGN,
+  M_CRYPT_VERIFIED,
+  M_CRYPT_ENCRYPT,
   M_PGP_KEY,
-#endif
   M_XLABEL,
   
   /* Options for Mailcap lookup */
@@ -252,34 +266,33 @@ enum
 /* quad-option vars */
 enum
 {
-
-#ifdef HAVE_PGP
-  OPT_VERIFYSIG,      /* verify PGP signatures */
-  OPT_PGPTRADITIONAL, /* create old-style PGP messages */
-#endif
-
-#ifdef USE_SSL
-  OPT_SSLSTARTTLS,
-#endif
-
-  OPT_PRINT,
-  OPT_INCLUDE,
+  OPT_ABORT,
+  OPT_BOUNCE,
+  OPT_COPY,
   OPT_DELETE,
+  OPT_FORWEDIT,
+  OPT_INCLUDE,
   OPT_MFUPTO,
   OPT_MIMEFWD,
+  OPT_MIMEFWDREST,
   OPT_MOVE,
-  OPT_COPY,
+  OPT_PGPTRADITIONAL, /* create old-style PGP messages */
 #ifdef USE_POP
   OPT_POPDELETE,
   OPT_POPRECONNECT,
 #endif
   OPT_POSTPONE,
+  OPT_PRINT,
   OPT_QUIT,
   OPT_REPLYTO,
-  OPT_ABORT,
   OPT_RECALL,
+#ifdef USE_SSL
+  OPT_SSLSTARTTLS,
+#endif
   OPT_SUBJECT,
-  OPT_MIMEFWDREST,
+  OPT_VERIFYSIG,      /* verify PGP signatures */
+    
+  /* THIS MUST BE THE LAST VALUE. */
   OPT_MAX
 };
 
@@ -367,6 +380,7 @@ enum
   OPTMETOO,
   OPTMHPURGE,
   OPTMIMEFORWDECODE,
+  OPTNARROWTREE,
   OPTPAGERSTOP,
   OPTPIPEDECODE,
   OPTPIPESPLIT,
@@ -402,6 +416,7 @@ enum
   OPTUSE8BITMIME,
   OPTUSEDOMAIN,
   OPTUSEFROM,
+  OPTUSEGPGAGENT,
 #ifdef HAVE_GETADDRINFO
   OPTUSEIPV6,
 #endif
@@ -414,14 +429,21 @@ enum
 
   /* PGP options */
   
-#ifdef HAVE_PGP
-  OPTPGPAUTOSIGN,
-  OPTPGPAUTOENCRYPT,
+  OPTCRYPTAUTOSIGN,
+  OPTCRYPTAUTOENCRYPT,
+  OPTCRYPTAUTOPGP,
+  OPTCRYPTAUTOSMIME,
+  OPTCRYPTREPLYENCRYPT,
+  OPTCRYPTREPLYSIGN,
+  OPTCRYPTREPLYSIGNENCRYPTED,
+  OPTCRYPTTIMESTAMP,
+  OPTSMIMEISDEFAULT,
+  OPTASKCERTLABEL,
+  OPTSDEFAULTDECRYPTKEY,
   OPTPGPIGNORESUB,
+  OPTPGPCHECKEXIT,
   OPTPGPLONGIDS,
-  OPTPGPREPLYENCRYPT,
-  OPTPGPREPLYSIGN,
-  OPTPGPREPLYSIGNENCRYPTED,
+  OPTPGPAUTOTRAD,
 #if 0
   OPTPGPENCRYPTSELF,
 #endif
@@ -429,7 +451,6 @@ enum
   OPTPGPSTRICTENC,
   OPTFORWDECRYPT,
   OPTPGPSHOWUNUSABLE,
-#endif
 
   /* pseudo options */
 
@@ -456,13 +477,10 @@ enum
 			 * 	    external program.
 			 */
   OPTMENUCALLER,	/* (pseudo) tell menu to give caller a take */
-#ifdef HAVE_PGP
+  OPTREDRAWTREE,	/* (pseudo) redraw the thread tree */
   OPTPGPCHECKTRUST,	/* (pseudo) used by pgp_select_key () */
   OPTDONTHANDLEPGPKEYS,	/* (pseudo) used to extract PGP keys */
-#endif
-
-
-
+  OPTUNBUFFEREDINPUT,   /* (pseudo) don't use key buffer */
 
   OPTMAX
 };
@@ -488,8 +506,16 @@ typedef struct list_t
   struct list_t *next;
 } LIST;
 
+typedef struct rx_list_t
+{
+  REGEXP *rx;
+  struct rx_list_t *next;
+} RX_LIST;
+
 #define mutt_new_list() safe_calloc (1, sizeof (LIST))
+#define mutt_new_rx_list() safe_calloc (1, sizeof (RX_LIST))
 void mutt_free_list (LIST **);
+void mutt_free_rx_list (RX_LIST **);
 int mutt_matches_ignore (const char *, LIST *);
 
 /* add an element to a list */
@@ -582,6 +608,8 @@ typedef struct body
   struct body *parts;           /* parts of a multipart or message/rfc822 */
   struct header *hdr;		/* header information for message/rfc822 */
 
+  struct attachptr *aptr;	/* Menu information, used in recvattach.c */
+  
   time_t stamp;			/* time stamp of last
 				 * encoding update.
 				 */
@@ -603,9 +631,8 @@ typedef struct body
 				 * set when in send-mode.
 				 */
 
-#ifdef HAVE_PGP
-  unsigned int goodsig : 1;	/* good PGP signature */
-#endif
+  unsigned int goodsig : 1;	/* good cryptographic signature */
+  unsigned int badsig : 1;	/* bad cryptographic signature (needed to check encrypted s/mime-signatures) */
 
   unsigned int collapsed : 1;	/* used by recvattach */
 
@@ -613,9 +640,8 @@ typedef struct body
 
 typedef struct header
 {
-#ifdef HAVE_PGP
-  unsigned int pgp : 4;
-#endif
+  unsigned int security : 9;  /* bit 0-6: flags, bit 7,8: application.
+				 see: crypt.h pgplib.h, smime.h */
 
   unsigned int mime : 1;    		/* has a Mime-Version header? */
   unsigned int flagged : 1; 		/* marked important? */
@@ -682,6 +708,8 @@ typedef struct header
 #if defined USE_POP || defined USE_IMAP
   void *data;            	/* driver-specific data */
 #endif
+  
+  char *maildir_flags;		/* unknown maildir flags */
 } HEADER;
 
 typedef struct thread
@@ -702,7 +730,6 @@ typedef struct thread
   HEADER *sort_key;
 } THREAD;
 
-#include "mutt_regex.h"
 
 /* flag to mutt_pattern_comp() */
 #define M_FULL_MSG	1	/* enable body and header matching */
@@ -795,15 +822,7 @@ typedef struct
 
 /* flags for the STATE struct */
 #define M_DISPLAY	(1<<0) /* output is displayed to the user */
-
-
-
-#ifdef HAVE_PGP
 #define M_VERIFY	(1<<1) /* perform signature verification */
-#endif
-
-
-
 #define M_PENDINGPREFIX (1<<2) /* prefix to write, but character must follow */
 #define M_WEED          (1<<3) /* weed headers even when not in display mode */
 #define M_CHARCONV	(1<<4) /* Do character set conversions */
@@ -823,3 +842,5 @@ int  state_printf(STATE *, const char *, ...);
 #include "protos.h"
 #include "lib.h"
 #include "globals.h"
+
+#endif /*MUTT_H*/
