@@ -1,6 +1,6 @@
 /*
- * Copyright (C) 1996-2000 Michael R. Elkins <me@cs.hmc.edu>
- * Copyright (C) 1999-2000 Thomas Roessler <roessler@guug.de>
+ * Copyright (C) 1996-2000 Michael R. Elkins <me@mutt.org>
+ * Copyright (C) 1999-2000 Thomas Roessler <roessler@does-not-exist.org>
  * 
  *     This program is free software; you can redistribute it and/or modify
  *     it under the terms of the GNU General Public License as published by
@@ -22,20 +22,27 @@
 #include "mime.h"
 #include "mailbox.h"
 #include "mx.h"
+#include "url.h"
 
-#ifdef HAVE_PGP
-#include "pgp.h"
+#include "reldate.h"
+
+#ifdef USE_IMAP
+#include "imap.h"
 #endif
 
+#include "mutt_crypt.h"
+
 #include <string.h>
 #include <ctype.h>
 #include <unistd.h>
 #include <stdlib.h>
 #include <sys/wait.h>
 #include <errno.h>
-#include <pwd.h>
 #include <sys/stat.h>
 #include <fcntl.h>
+#include <time.h>
+#include <sys/types.h>
+#include <utime.h>
 
 BODY *mutt_new_body (void)
 {
@@ -46,14 +53,6 @@ BODY *mutt_new_body (void)
   return (p);
 }
 
-BODY *mutt_dup_body (BODY *b)
-{
-  BODY *bn;
-
-  bn = mutt_new_body();
-  memcpy(bn, b, sizeof (BODY));
-  return bn;
-}
 
 /* Modified by blong to accept a "suggestion" for file name.  If
  * that file exists, then construct one with unique name but 
@@ -128,13 +127,12 @@ int mutt_copy_body (FILE *fp, BODY **tgt, BODY *src)
   memcpy (b, src, sizeof (BODY));
   b->parts = NULL;
   b->next  = NULL;
-  b->hdr = NULL;
 
   b->filename = safe_strdup (tmp);
   b->use_disp = use_disp;
   b->unlink = 1;
 
-  if (mutt_is_text_type (b->type, b->subtype))
+  if (mutt_is_text_part (b))
     b->noconv = 1;
 
   b->xtype = safe_strdup (b->xtype);
@@ -144,6 +142,12 @@ int mutt_copy_body (FILE *fp, BODY **tgt, BODY *src)
   b->d_filename = safe_strdup (b->d_filename);
   b->description = safe_strdup (b->description);
 
+  /* 
+   * we don't seem to need the HEADER structure currently.
+   * XXX - this may change in the future
+   */
+
+  if (b->hdr) b->hdr = NULL;
   
   /* copy parameters */
   for (par = b->parameter, ppar = &b->parameter; par; ppar = &(*ppar)->next, par = par->next)
@@ -172,13 +176,19 @@ void mutt_free_body (BODY **p)
     if (b->parameter)
       mutt_free_parameter (&b->parameter);
     if (b->unlink && b->filename)
+    {
+      dprint (1, (debugfile, "mutt_free_body: Unlinking %s.\n", b->filename));
       unlink (b->filename);
-    safe_free ((void **) &b->filename);
-    safe_free ((void **) &b->content);
-    safe_free ((void **) &b->xtype);
-    safe_free ((void **) &b->subtype);
-    safe_free ((void **) &b->description);
-    safe_free ((void **) &b->form_name);
+    }
+    else if (b->filename) 
+      dprint (1, (debugfile, "mutt_free_body: Not unlinking %s.\n", b->filename));
+
+    FREE (&b->filename);
+    FREE (&b->content);
+    FREE (&b->xtype);
+    FREE (&b->subtype);
+    FREE (&b->description);
+    FREE (&b->form_name);
 
     if (b->hdr)
     {
@@ -190,7 +200,7 @@ void mutt_free_body (BODY **p)
     if (b->parts)
       mutt_free_body (&b->parts);
 
-    safe_free ((void **) &b);
+    FREE (&b);
   }
 
   *p = 0;
@@ -203,11 +213,11 @@ void mutt_free_parameter (PARAMETER **p)
 
   while (t)
   {
-    safe_free ((void **) &t->attribute);
-    safe_free ((void **) &t->value);
+    FREE (&t->attribute);
+    FREE (&t->value);
     o = t;
     t = t->next;
-    safe_free ((void **) &o);
+    FREE (&o);
   }
   *p = 0;
 }
@@ -240,8 +250,8 @@ void mutt_free_list (LIST **list)
   {
     p = *list;
     *list = (*list)->next;
-    safe_free ((void **) &p->data);
-    safe_free ((void **) &p);
+    FREE (&p->data);
+    FREE (&p);
   }
 }
 
@@ -259,12 +269,16 @@ void mutt_free_header (HEADER **h)
   if(!h || !*h) return;
   mutt_free_envelope (&(*h)->env);
   mutt_free_body (&(*h)->content);
-  safe_free ((void **) &(*h)->tree);
-  safe_free ((void **) &(*h)->path);
+  FREE (&(*h)->maildir_flags);
+  FREE (&(*h)->tree);
+  FREE (&(*h)->path);
 #ifdef MIXMASTER
   mutt_free_list (&(*h)->chain);
 #endif
-  safe_free ((void **) h);
+#if defined USE_POP || defined USE_IMAP
+  FREE (&(*h)->data);
+#endif
+  FREE (h);
 }
 
 /* returns true if the header contained in "s" is in list "t" */
@@ -272,7 +286,7 @@ int mutt_matches_ignore (const char *s, LIST *t)
 {
   for (; t; t = t->next)
   {
-    if (!mutt_strncasecmp (s, t->data, mutt_strlen (t->data)) || *t->data == '*')
+    if (!ascii_strncasecmp (s, t->data, mutt_strlen (t->data)) || *t->data == '*')
       return 1;
   }
   return 0;
@@ -365,8 +379,10 @@ char *_mutt_expand_path (char *s, size_t slen, int rx)
       case '+':    
       {
 #ifdef USE_IMAP
-	/* special case: folder = {host}: don't append slash */
-	if (mx_is_imap (NONULL (Maildir)) && Maildir[strlen (Maildir) - 1] == '}')
+	/* if folder = {host} or imap[s]://host/: don't append slash */
+	if (mx_is_imap (NONULL (Maildir)) &&
+	    (Maildir[strlen (Maildir) - 1] == '}' ||
+	     Maildir[strlen (Maildir) - 1] == '/'))
 	  strfcpy (p, NONULL (Maildir), sizeof (p));
 	else
 #endif
@@ -455,14 +471,68 @@ char *_mutt_expand_path (char *s, size_t slen, int rx)
   }
   while (recurse);
 
+#ifdef USE_IMAP
+  /* Rewrite IMAP path in canonical form - aids in string comparisons of
+   * folders. May possibly fail, in which case s should be the same. */
+  if (mx_is_imap (s))
+    imap_expand_path (s, slen);
+#endif
+
   return (s);
 }
 
+/* Extract the real name from /etc/passwd's GECOS field.
+ * When set, honor the regular expression in GecosMask,
+ * otherwise assume that the GECOS field is a 
+ * comma-separated list.
+ * Replace "&" by a capitalized version of the user's login
+ * name.
+ */
+
+char *mutt_gecos_name (char *dest, size_t destlen, struct passwd *pw)
+{
+  regmatch_t pat_match[1];
+  size_t pwnl;
+  int idx;
+  char *p;
+  
+  if (!pw || !pw->pw_gecos) 
+    return NULL;
+
+  memset (dest, 0, destlen);
+  
+  if (GecosMask.rx)
+  {
+    if (regexec (GecosMask.rx, pw->pw_gecos, 1, pat_match, 0) == 0)
+      strfcpy (dest, pw->pw_gecos + pat_match[0].rm_so, 
+	       MIN (pat_match[0].rm_eo - pat_match[0].rm_so + 1, destlen));
+  }
+  else if ((p = strchr (pw->pw_gecos, ',')))
+    strfcpy (dest, pw->pw_gecos, MIN (destlen, p - pw->pw_gecos + 1));
+  else
+    strfcpy (dest, pw->pw_gecos, destlen);
+
+  pwnl = strlen (pw->pw_name);
+
+  for (idx = 0; dest[idx]; idx++)
+  {
+    if (dest[idx] == '&')
+    {
+      memmove (&dest[idx + pwnl], &dest[idx + 1],
+	       MAX(destlen - idx - pwnl - 1, 0));
+      memcpy (&dest[idx], pw->pw_name, MIN(destlen - idx - 1, pwnl));
+      dest[idx] = toupper ((unsigned char) dest[idx]);
+    }
+  }
+      
+  return dest;
+}
+  
 
 char *mutt_get_parameter (const char *s, PARAMETER *p)
 {
   for (; p; p = p->next)
-    if (mutt_strcasecmp (s, p->attribute) == 0)
+    if (ascii_strcasecmp (s, p->attribute) == 0)
       return (p->value);
 
   return NULL;
@@ -471,9 +541,16 @@ char *mutt_get_parameter (const char *s, PARAMETER *p)
 void mutt_set_parameter (const char *attribute, const char *value, PARAMETER **p)
 {
   PARAMETER *q;
+
+  if (!value)
+  {
+    mutt_delete_parameter (attribute, p);
+    return;
+  }
+  
   for(q = *p; q; q = q->next)
   {
-    if (mutt_strcasecmp (attribute, q->attribute) == 0)
+    if (ascii_strcasecmp (attribute, q->attribute) == 0)
     {
       mutt_str_replace (&q->value, value);
       return;
@@ -487,8 +564,21 @@ void mutt_set_parameter (const char *attribute, const char *value, PARAMETER **p
   *p = q;
 }
 
-
-
+void mutt_delete_parameter (const char *attribute, PARAMETER **p)
+{
+  PARAMETER *q;
+  
+  for (q = *p; q; p = &q->next, q = q->next)
+  {
+    if (ascii_strcasecmp (attribute, q->attribute) == 0)
+    {
+      *p = q->next;
+      q->next = NULL;
+      mutt_free_parameter (&q);
+      return;
+    }
+  }
+}
 
 /* returns 1 if Mutt can't display this type of data, 0 otherwise */
 int mutt_needs_mailcap (BODY *m)
@@ -497,53 +587,49 @@ int mutt_needs_mailcap (BODY *m)
   {
     case TYPETEXT:
 
-      if (!mutt_strcasecmp ("plain", m->subtype) ||
-	  !mutt_strcasecmp ("rfc822-headers", m->subtype) ||
-	  !mutt_strcasecmp ("enriched", m->subtype))
+      if (!ascii_strcasecmp ("plain", m->subtype) ||
+	  !ascii_strcasecmp ("rfc822-headers", m->subtype) ||
+	  !ascii_strcasecmp ("enriched", m->subtype))
 	return 0;
       break;
 
-
-
-#ifdef HAVE_PGP
     case TYPEAPPLICATION:
-      if(mutt_is_application_pgp(m))
+      if((WithCrypto & APPLICATION_PGP) && mutt_is_application_pgp(m))
+	return 0;
+      if((WithCrypto & APPLICATION_SMIME) && mutt_is_application_smime(m))
 	return 0;
       break;
-#endif /* HAVE_PGP */
-
 
     case TYPEMULTIPART:
     case TYPEMESSAGE:
-
       return 0;
   }
 
   return 1;
 }
 
-int mutt_is_text_type (int t, char *s)
+int mutt_is_text_part (BODY *b)
 {
+  int t = b->type;
+  char *s = b->subtype;
+  
+  if ((WithCrypto & APPLICATION_PGP) && mutt_is_application_pgp (b))
+    return 0;
+
   if (t == TYPETEXT)
     return 1;
 
   if (t == TYPEMESSAGE)
   {
-    if (!mutt_strcasecmp ("delivery-status", s))
+    if (!ascii_strcasecmp ("delivery-status", s))
       return 1;
   }
 
-
-
-#ifdef HAVE_PGP
-  if (t == TYPEAPPLICATION)
+  if ((WithCrypto & APPLICATION_PGP) && t == TYPEAPPLICATION)
   {
-    if (!mutt_strcasecmp ("pgp-keys", s))
+    if (!ascii_strcasecmp ("pgp-keys", s))
       return 1;
   }
-#endif /* HAVE_PGP */
-
-
 
   return 0;
 }
@@ -559,18 +645,20 @@ void mutt_free_envelope (ENVELOPE **p)
   rfc822_free_address (&(*p)->from);
   rfc822_free_address (&(*p)->reply_to);
   rfc822_free_address (&(*p)->mail_followup_to);
-  safe_free ((void **) &(*p)->subject);
-  safe_free ((void **) &(*p)->message_id);
-  safe_free ((void **) &(*p)->supersedes);
-  safe_free ((void **) &(*p)->date);
+  FREE (&(*p)->subject);
+  FREE (&(*p)->message_id);
+  FREE (&(*p)->supersedes);
+  FREE (&(*p)->date);
   mutt_free_list (&(*p)->references);
+  mutt_free_list (&(*p)->in_reply_to);
   mutt_free_list (&(*p)->userhdrs);
-  safe_free ((void **) p);
+  FREE (p);
 }
 
-void mutt_mktemp (char *s)
+void _mutt_mktemp (char *s, const char *src, int line)
 {
   snprintf (s, _POSIX_PATH_MAX, "%s/mutt-%s-%d-%d", NONULL (Tempdir), NONULL(Hostname), (int) getpid (), Counter++);
+  dprint (1, (debugfile, "%s:%d: mutt_mktemp returns \"%s\".\n", src, line, s));
   unlink (s);
 }
 
@@ -582,9 +670,9 @@ void mutt_free_alias (ALIAS **p)
   {
     t = *p;
     *p = (*p)->next;
-    safe_free ((void **) &t->name);
+    FREE (&t->name);
     rfc822_free_address (&t->addr);
-    safe_free ((void **) &t);
+    FREE (&t);
   }
 }
 
@@ -593,7 +681,29 @@ void mutt_pretty_mailbox (char *s)
 {
   char *p = s, *q = s;
   size_t len;
+  url_scheme_t scheme;
 
+  scheme = url_check_scheme (s);
+
+#ifdef USE_IMAP
+  if (scheme == U_IMAP || scheme == U_IMAPS)
+  {
+    imap_pretty_mailbox (s);
+    return;
+  }
+#endif
+
+  /* if s is an url, only collapse path component */
+  if (scheme != U_UNKNOWN)
+  {
+    p = strchr(s, ':')+1;
+    if (!strncmp (p, "//", 2))
+      q = strchr (p+2, '/');
+    if (!q)
+      q = strchr (p, '\0');
+    p = q;
+  }
+  
   /* first attempt to collapse the pathname */
   while (*p)
   {
@@ -612,16 +722,17 @@ void mutt_pretty_mailbox (char *s)
   }
   *q = 0;
 
-  if (mutt_strncmp (s, Maildir, (len = mutt_strlen (Maildir))) == 0 && s[len] == '/')
+  if (mutt_strncmp (s, Maildir, (len = mutt_strlen (Maildir))) == 0 &&
+      s[len] == '/')
   {
     *s++ = '=';
-    strcpy (s, s + len);
+    memmove (s, s + len, mutt_strlen (s + len) + 1);
   }
   else if (mutt_strncmp (s, Homedir, (len = mutt_strlen (Homedir))) == 0 &&
 	   s[len] == '/')
   {
     *s++ = '~';
-    strcpy (s, s + len - 1);
+    memmove (s, s + len - 1, mutt_strlen (s + len - 1) + 1);
   }
 }
 
@@ -655,66 +766,59 @@ void mutt_expand_file_fmt (char *dest, size_t destlen, const char *fmt, const ch
 
 void mutt_expand_fmt (char *dest, size_t destlen, const char *fmt, const char *src)
 {
-  const char *p = fmt;
-  const char *last = p;
-  size_t len;
+  const char *p;
+  char *d;
   size_t slen;
   int found = 0;
 
   slen = mutt_strlen (src);
+  destlen--;
   
-  while ((p = strchr (p, '%')) != NULL)
+  for (p = fmt, d = dest; destlen && *p; p++)
   {
-    if (p[1] == 's')
+    if (*p == '%') 
     {
-      found++;
-
-      len = (size_t) (p - last);
-      if (len)
+      switch (p[1])
       {
-	if (len > destlen - 1)
-	  len = destlen - 1;
-
-	memcpy (dest, last, len);
-	dest += len;
-	destlen -= len;
-
-	if (destlen <= 0)
-	{
-	  *dest = 0;
-	  break; /* no more space */
-	}
+	case '%':
+	  *d++ = *p++;
+	  destlen--;
+	  break;
+	case 's':
+	  found = 1;
+	  strfcpy (d, src, destlen + 1);
+	  d       += destlen > slen ? slen : destlen;
+	  destlen -= destlen > slen ? slen : destlen;
+	  p++;
+	  break;
+	default:
+	  *d++ = *p; 
+	  destlen--;
+	  break;
       }
-
-      strfcpy (dest, src, destlen);
-      if (slen > destlen)
-      {
-	/* no more room */
-	break;
-      }
-      dest += slen;
-      destlen -= slen;
-
-      p += 2;
-      last = p;
     }
-    else if (p[1] == '%')
-      p++;
-
-    p++;
+    else
+    {
+      *d++ = *p;
+      destlen--;
+    }
+  }
+  
+  *d = '\0';
+  
+  if (!found && destlen > 0)
+  {
+    strncat (dest, " ", destlen);
+    strncat (dest, src, destlen-1);
   }
-
-  if (found)
-    strfcpy (dest, last, destlen);
-  else
-    snprintf (dest, destlen, "%s %s", fmt, src);
   
 }
 
-/* return 0 on success, -1 on error */
+/* return 0 on success, -1 on abort, 1 on error */
 int mutt_check_overwrite (const char *attname, const char *path,
-				char *fname, size_t flen, int *append) 
+				char *fname, size_t flen, int *append, char **directory) 
 {
+  int rc = 0;
   char tmp[_POSIX_PATH_MAX];
   struct stat st;
 
@@ -725,18 +829,38 @@ int mutt_check_overwrite (const char *attname, const char *path,
     return -1;
   if (S_ISDIR (st.st_mode))
   {
-    if (mutt_yesorno (_("File is a directory, save under it?"), 1) != M_YES) 
-      return (-1);
+    if (directory)
+    {
+      switch (mutt_multi_choice
+	      (_("File is a directory, save under it? [(y)es, (n)o, (a)ll]"), _("yna")))
+      {
+	case 3:		/* all */
+	  mutt_str_replace (directory, fname);
+	  break;
+	case 1:		/* yes */
+	  FREE (directory);
+	  break;
+	case -1:	/* abort */
+	  FREE (directory); 
+	  return -1;
+	case  2:	/* no */
+	  FREE (directory);
+	  return 1;
+      }
+    }
+    else if ((rc = mutt_yesorno (_("File is a directory, save under it?"), M_YES)) != M_YES)
+      return (rc == M_NO) ? 1 : -1;
+
     if (!attname || !attname[0])
     {
       tmp[0] = 0;
       if (mutt_get_field (_("File under directory: "), tmp, sizeof (tmp),
 				      M_FILE | M_CLEAR) != 0 || !tmp[0])
 	return (-1);
-      snprintf (fname, flen, "%s/%s", path, tmp);
+      mutt_concat_path (fname, path, tmp, flen);
     }
     else
-      snprintf (fname, flen, "%s/%s", path, attname);
+      mutt_concat_path (fname, path, mutt_basename (attname), flen);
   }
   
   if (*append == 0 && access (fname, F_OK) == 0)
@@ -745,8 +869,9 @@ int mutt_check_overwrite (const char *attname, const char *path,
 	    (_("File exists, (o)verwrite, (a)ppend, or (c)ancel?"), _("oac")))
     {
       case -1: /* abort */
+        return -1;
       case 3:  /* cancel */
-	return -1;
+	return 1;
 
       case 2: /* append */
         *append = M_SAVE_APPEND;
@@ -799,6 +924,7 @@ void mutt_FormatString (char *dest,		/* output buffer */
   char ifstring[SHORT_STRING], elsestring[SHORT_STRING];
   size_t wlen, count, len;
 
+  prefix[0] = '\0';
   destlen--; /* save room for the terminal \0 */
   wlen = (flags & M_FORMAT_ARROWCURSOR && option (OPTARROWCURSOR)) ? 3 : 0;
     
@@ -919,11 +1045,16 @@ void mutt_FormatString (char *dest,		/* output buffer */
       else
       {
 	short tolower =  0;
+	short nodots  = 0;
 	
-	if (ch == '_')
+	while (ch == '_' || ch == ':') 
 	{
+	  if (ch == '_')
+	    tolower = 1;
+	  else if (ch == ':') 
+	    nodots = 1;
+	  
 	  ch = *src++;
-	  tolower = 1;
 	}
 	
 	/* use callback function to handle this case */
@@ -931,6 +1062,13 @@ void mutt_FormatString (char *dest,		/* output buffer */
 
 	if (tolower)
 	  mutt_strlower (buf);
+	if (nodots) 
+	{
+	  char *p = buf;
+	  for (; *p; p++)
+	    if (*p == '.')
+		*p = '_';
+	}
 	
 	if ((len = mutt_strlen (buf)) + wlen > destlen)
 	  len = (destlen - wlen > 0) ? (destlen - wlen) : 0;
@@ -977,6 +1115,7 @@ void mutt_FormatString (char *dest,		/* output buffer */
   }
   *wptr = 0;
 
+#if 0
   if (flags & M_FORMAT_MAKEPRINT)
   {
     /* Make sure that the string is printable by changing all non-printable
@@ -986,6 +1125,7 @@ void mutt_FormatString (char *dest,		/* output buffer */
 	  !((flags & M_FORMAT_TREE) && (*cp <= M_TREE_MAX)))
 	*cp = isspace ((unsigned char) *cp) ? ' ' : '.';
   }
+#endif
 }
 
 /* This function allows the user to specify a command to read stdout from in
@@ -1003,9 +1143,9 @@ FILE *mutt_open_read (const char *path, pid_t *thepid)
     char *s = safe_strdup (path);
 
     s[len - 1] = 0;
-    endwin ();
+    mutt_endwin (NULL);
     *thepid = mutt_create_filter (s, NULL, &f, NULL);
-    safe_free ((void **) &s);
+    FREE (&s);
   }
   else
   {
@@ -1015,28 +1155,39 @@ FILE *mutt_open_read (const char *path, pid_t *thepid)
   return (f);
 }
 
-/* returns 1 if OK to proceed, 0 to abort */
+/* returns 0 if OK to proceed, -1 to abort, 1 to retry */
 int mutt_save_confirm (const char *s, struct stat *st)
 {
   char tmp[_POSIX_PATH_MAX];
-  int ret = 1;
+  int ret = 0;
+  int rc;
   int magic = 0;
 
   magic = mx_get_magic (s);
 
+#ifdef USE_POP
+  if (magic == M_POP)
+  {
+    mutt_error _("Can't save message to POP mailbox.");
+    return 1;
+  }
+#endif
+
   if (stat (s, st) != -1)
   {
     if (magic == -1)
     {
       mutt_error (_("%s is not a mailbox!"), s);
-      return 0;
+      return 1;
     }
 
     if (option (OPTCONFIRMAPPEND))
     {
       snprintf (tmp, sizeof (tmp), _("Append messages to %s?"), s);
-      if (mutt_yesorno (tmp, 1) < 1)
-	ret = 0;
+      if ((rc = mutt_yesorno (tmp, M_YES)) == M_NO)
+	ret = 1;
+      else if (rc == -1)
+	ret = -1;
     }
   }
   else
@@ -1053,14 +1204,16 @@ int mutt_save_confirm (const char *s, struct stat *st)
 	if (option (OPTCONFIRMCREATE))
 	{
 	  snprintf (tmp, sizeof (tmp), _("Create %s?"), s);
-	  if (mutt_yesorno (tmp, 1) < 1)
-	    ret = 0;
+	  if ((rc = mutt_yesorno (tmp, M_YES)) == M_NO)
+	    ret = 1;
+	  else if (rc == -1)
+	    ret = -1;
 	}
       }
       else
       {
 	mutt_perror (s);
-	return 0;
+	return 1;
       }
     }
   }
@@ -1069,33 +1222,50 @@ int mutt_save_confirm (const char *s, struct stat *st)
   return (ret);
 }
 
-void state_prefix_putc(char c, STATE *s)
+void state_prefix_putc (char c, STATE *s)
 {
   if (s->flags & M_PENDINGPREFIX)
   {
-    state_reset_prefix(s);
+    state_reset_prefix (s);
     if (s->prefix)
-      state_puts(s->prefix, s);
+      state_puts (s->prefix, s);
   }
 
-  state_putc(c, s);
+  state_putc (c, s);
 
-  if(c == '\n')
-    state_set_prefix(s);
+  if (c == '\n')
+    state_set_prefix (s);
 }
 
-int state_printf(STATE *s, const char *fmt, ...)
+int state_printf (STATE *s, const char *fmt, ...)
 {
   int rv;
   va_list ap;
 
   va_start (ap, fmt);
-  rv = vfprintf(s->fpout, fmt, ap);
-  va_end(ap);
+  rv = vfprintf (s->fpout, fmt, ap);
+  va_end (ap);
   
   return rv;
 }
 
+void state_mark_attach (STATE *s)
+{
+  if ((s->flags & M_DISPLAY) && !mutt_strcmp (Pager, "builtin"))
+    state_puts (AttachmentMarker, s);
+}
+
+void state_attach_puts (const char *t, STATE *s)
+{
+  if (*t != '\n') state_mark_attach (s);
+  while (*t)
+  {
+    state_putc (*t, s);
+    if (*t++ == '\n' && *t)
+      if (*t != '\n') state_mark_attach (s);
+  }
+}
+
 void mutt_display_sanitize (char *s)
 {
   for (; *s; s++)
@@ -1105,3 +1275,122 @@ void mutt_display_sanitize (char *s)
   }
 }
       
+void mutt_sleep (short s)
+{
+  if (SleepTime > s)
+    sleep (SleepTime);
+  else if (s)
+    sleep (s);
+}
+
+void mutt_buffer_addstr (BUFFER* buf, const char* s)
+{
+  mutt_buffer_add (buf, s, mutt_strlen (s));
+}
+
+void mutt_buffer_addch (BUFFER* buf, char c)
+{
+  mutt_buffer_add (buf, &c, 1);
+}
+
+/* dynamically grows a BUFFER to accomodate s, in increments of 128 bytes.
+ * Always one byte bigger than necessary for the null terminator, and
+ * the buffer is always null-terminated */
+void mutt_buffer_add (BUFFER* buf, const char* s, size_t len)
+{
+  size_t offset;
+
+  if (buf->dptr + len + 1 > buf->data + buf->dsize)
+  {
+    offset = buf->dptr - buf->data;
+    buf->dsize += len < 128 ? 128 : len + 1;
+    safe_realloc ((void**) &buf->data, buf->dsize);
+    buf->dptr = buf->data + offset;
+  }
+  memcpy (buf->dptr, s, len);
+  buf->dptr += len;
+  *(buf->dptr) = '\0';
+}
+
+/* Decrease a file's modification time by 1 second */
+
+time_t mutt_decrease_mtime (const char *f, struct stat *st)
+{
+  struct utimbuf utim;
+  struct stat _st;
+  time_t mtime;
+  
+  if (!st)
+  {
+    if (stat (f, &_st) == -1)
+      return -1;
+    st = &_st;
+  }
+
+  if ((mtime = st->st_mtime) == time (NULL))
+  {
+    mtime -= 1;
+    utim.actime = mtime;
+    utim.modtime = mtime;
+    utime (f, &utim);
+  }
+  
+  return mtime;
+}
+
+const char *mutt_make_version (void)
+{
+  static char vstring[STRING];
+  snprintf (vstring, sizeof (vstring), "Mutt %s (%s)",
+	    MUTT_VERSION, ReleaseDate);
+  return vstring;
+}
+
+REGEXP *mutt_compile_regexp (const char *s, int flags)
+{
+  REGEXP *pp = safe_calloc (sizeof (REGEXP), 1);
+  pp->pattern = safe_strdup (s);
+  pp->rx = safe_calloc (sizeof (regex_t), 1);
+  if (REGCOMP (pp->rx, NONULL(s), flags) != 0)
+    mutt_free_regexp (&pp);
+
+  return pp;
+}
+
+void mutt_free_regexp (REGEXP **pp)
+{
+  FREE (&(*pp)->pattern);
+  regfree ((*pp)->rx);
+  FREE (&(*pp)->rx);
+  FREE (pp);
+}
+
+void mutt_free_rx_list (RX_LIST **list)
+{
+  RX_LIST *p;
+  
+  if (!list) return;
+  while (*list)
+  {
+    p = *list;
+    *list = (*list)->next;
+    mutt_free_regexp (&p->rx);
+    FREE (&p);
+  }
+}
+
+int mutt_match_rx_list (const char *s, RX_LIST *l)
+{
+  if (!s)  return 0;
+  
+  for (; l; l = l->next)
+  {
+    if (regexec (l->rx->rx, s, (size_t) 0, (regmatch_t *) 0, (int) 0) == 0)
+    {
+      dprint (5, (debugfile, "mutt_match_rx_list: %s matches %s\n", s, l->rx->pattern));
+      return 1;
+    }
+  }
+
+  return 0;
+}
