@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 1996-2000 Michael R. Elkins <me@cs.hmc.edu>
+ * Copyright (C) 1996-2002 Michael R. Elkins <me@mutt.org>
  * 
  *     This program is free software; you can redistribute it and/or modify
  *     it under the terms of the GNU General Public License as published by
@@ -25,29 +25,22 @@
 #include "sort.h"
 #include "pager.h"
 #include "attach.h"
+#include "mbyte.h"
+
+#include "mx.h"
 
 #ifdef USE_IMAP
-#include "imap.h"
+#include "imap_private.h"
 #endif
 
-
-#ifdef HAVE_PGP
-#include "pgp.h"
-#endif
-
-
-
-
-
-
-
-
+#include "mutt_crypt.h"
 
 #include <sys/stat.h>
 #include <ctype.h>
 #include <unistd.h>
 #include <stdlib.h>
 #include <string.h>
+#include <errno.h>
 
 #define ISHEADER(x) ((x) == MT_COLOR_HEADER || (x) == MT_COLOR_HDEFAULT)
 
@@ -82,6 +75,17 @@ static const char *Function_not_permitted_in_attach_message_mode = N_("Function
 			break; \
 		     }
 
+#ifdef USE_IMAP 
+/* the error message returned here could be better. */
+#define CHECK_IMAP_ACL(aclbit) if (Context->magic == M_IMAP) \
+		if (mutt_bit_isset (((IMAP_DATA *)Context->data)->capabilities, ACL) \
+		&& !mutt_bit_isset(((IMAP_DATA *)Context->data)->rights,aclbit)){ \
+			mutt_flushinp(); \
+			mutt_error ("Operation not permitted by the IMAP ACL for this mailbox"); \
+			break; \
+		}
+#endif
+
 struct q_class_t
 {
   int length;
@@ -256,14 +260,25 @@ resolve_color (struct line_t *lineInfo, int n, int cnt, int flags, int special,
   /* handle "special" bold & underlined characters */
   if (special || a->attr)
   {
-    if (special == A_BOLD || (a->attr & ANSI_BOLD))
+#ifdef HAVE_COLOR
+    if ((a->attr & ANSI_COLOR))
+    {
+      if (a->pair == -1)
+	a->pair = mutt_alloc_color (a->fg, a->bg);
+      color = a->pair;
+      if (a->attr & ANSI_BOLD)
+	  color |= A_BOLD;
+    }
+    else
+#endif
+      if ((special & A_BOLD) || (a->attr & ANSI_BOLD))
     {
       if (ColorDefs[MT_COLOR_BOLD] && !search)
 	color = ColorDefs[MT_COLOR_BOLD];
       else
 	color ^= A_BOLD;
     }
-    else if (special == A_UNDERLINE || (a->attr & ANSI_UNDERLINE))
+    if ((special & A_UNDERLINE) || (a->attr & ANSI_UNDERLINE))
     {
       if (ColorDefs[MT_COLOR_UNDERLINE] && !search)
 	color = ColorDefs[MT_COLOR_UNDERLINE];
@@ -278,14 +293,6 @@ resolve_color (struct line_t *lineInfo, int n, int cnt, int flags, int special,
     {
       color ^= A_BLINK;
     }
-#ifdef HAVE_COLOR
-    else if (a->attr & ANSI_COLOR)
-    {
-      if (a->pair == -1)
-	a->pair = mutt_alloc_color (a->fg,a->bg);
-      color = a->pair;
-    }
-#endif
     else if (a->attr & ANSI_OFF)
     {
       a->attr = 0;
@@ -309,12 +316,9 @@ append_line (struct line_t *lineInfo, int n, int cnt)
   lineInfo[n+1].continuation = 1;
 
   /* find the real start of the line */
-  m = n;
-  while (m >= 0)
-  {
+  for (m = n; m >= 0; m--)
     if (lineInfo[m].continuation == 0) break;
-    m--;
-  }
+
   (lineInfo[n+1].syntax)[0].first = m;
   (lineInfo[n+1].syntax)[0].last = (lineInfo[n].continuation) ? 
     cnt + (lineInfo[n].syntax)[0].last : cnt;
@@ -376,8 +380,8 @@ cleanup_quote (struct q_class_t **QuoteList)
       cleanup_quote (&((*QuoteList)->down));
     ptr = (*QuoteList)->next;
     if ((*QuoteList)->prefix)
-      safe_free ((void **) &(*QuoteList)->prefix);
-    safe_free ((void **) QuoteList);
+      FREE (&(*QuoteList)->prefix);
+    FREE (QuoteList);
     *QuoteList = ptr;
   }
 
@@ -695,8 +699,10 @@ classify_quote (struct q_class_t **QuoteList, const char *qptr,
   return class;
 }
 
+static int check_attachment_marker (char *);
+
 static void
-resolve_types (char *buf, struct line_t *lineInfo, int n, int last,
+resolve_types (char *buf, char *raw, struct line_t *lineInfo, int n, int last,
 		struct q_class_t **QuoteList, int *q_level, int *force_redraw,
 		int q_classify)
 {
@@ -729,8 +735,15 @@ resolve_types (char *buf, struct line_t *lineInfo, int n, int last,
       }
     }
   }
+  else if (mutt_strncmp ("\033[0m", raw, 4) == 0)	/* a little hack... */
+    lineInfo[n].type = MT_COLOR_NORMAL;
+#if 0
   else if (mutt_strncmp ("[-- ", buf, 4) == 0)
     lineInfo[n].type = MT_COLOR_ATTACHMENT;
+#else
+  else if (check_attachment_marker ((char *) raw) == 0)
+    lineInfo[n].type = MT_COLOR_ATTACHMENT;
+#endif
   else if (mutt_strcmp ("-- \n", buf) == 0 || mutt_strcmp ("-- \r\n", buf) == 0)
   {
     i = n + 1;
@@ -745,7 +758,7 @@ resolve_types (char *buf, struct line_t *lineInfo, int n, int last,
 	if (lineInfo[i].chunks)
 	{
 	  lineInfo[i].chunks = 0;
-	  safe_realloc ((void **) &(lineInfo[n].syntax), 
+	  safe_realloc (&(lineInfo[n].syntax), 
 			sizeof (struct syntax_t));
 	}
 	lineInfo[i++].type = MT_COLOR_SIGNATURE;
@@ -820,7 +833,7 @@ resolve_types (char *buf, struct line_t *lineInfo, int n, int last,
 	    if (!found)
 	    {
 	      if (++(lineInfo[n].chunks) > 1)
-		safe_realloc ((void **)&(lineInfo[n].syntax), 
+		safe_realloc (&(lineInfo[n].syntax), 
 			      (lineInfo[n].chunks) * sizeof (struct syntax_t));
 	    }
 	    i = lineInfo[n].chunks - 1;
@@ -852,6 +865,106 @@ resolve_types (char *buf, struct line_t *lineInfo, int n, int last,
   }
 }
 
+static int is_ansi (unsigned char *buf)
+{
+  while (*buf && (isdigit(*buf) || *buf == ';'))
+    buf++;
+  return (*buf == 'm');
+}
+
+static int check_attachment_marker (char *p)
+{
+  char *q = AttachmentMarker;
+  
+  for (;*p == *q && *q && *p && *q != '\a' && *p != '\a'; p++, q++)
+    ;
+  return (int) (*p - *q);
+}
+
+static int grok_ansi(unsigned char *buf, int pos, ansi_attr *a)
+{
+  int x = pos;
+
+  while (isdigit(buf[x]) || buf[x] == ';')
+    x++;
+
+  /* Character Attributes */
+  if (option (OPTALLOWANSI) && a != NULL && buf[x] == 'm')
+  {
+    if (pos == x)
+    {
+#ifdef HAVE_COLOR
+      if (a->pair != -1)
+	mutt_free_color (a->fg, a->bg);
+#endif
+      a->attr = ANSI_OFF;
+      a->pair = -1;
+    }
+    while (pos < x)
+    {
+      if (buf[pos] == '1' && (pos+1 == x || buf[pos+1] == ';'))
+      {
+	a->attr |= ANSI_BOLD;
+	pos += 2;
+      } 
+      else if (buf[pos] == '4' && (pos+1 == x || buf[pos+1] == ';'))
+      {
+	a->attr |= ANSI_UNDERLINE;
+	pos += 2;
+      }
+      else if (buf[pos] == '5' && (pos+1 == x || buf[pos+1] == ';'))
+      {
+	a->attr |= ANSI_BLINK;
+	pos += 2;
+      }
+      else if (buf[pos] == '7' && (pos+1 == x || buf[pos+1] == ';'))
+      {
+	a->attr |= ANSI_REVERSE;
+	pos += 2;
+      }
+      else if (buf[pos] == '0' && (pos+1 == x || buf[pos+1] == ';'))
+      {
+#ifdef HAVE_COLOR
+	if (a->pair != -1)
+	  mutt_free_color(a->fg,a->bg);
+#endif
+	a->attr = ANSI_OFF;
+	a->pair = -1;
+	pos += 2;
+      }
+      else if (buf[pos] == '3' && isdigit(buf[pos+1]))
+      {
+#ifdef HAVE_COLOR
+	if (a->pair != -1)
+	  mutt_free_color(a->fg,a->bg);
+#endif
+	a->pair = -1;
+	a->attr |= ANSI_COLOR;
+	a->fg = buf[pos+1] - '0';
+	pos += 3;
+      }
+      else if (buf[pos] == '4' && isdigit(buf[pos+1]))
+      {
+#ifdef HAVE_COLOR
+	if (a->pair != -1)
+	  mutt_free_color(a->fg,a->bg);
+#endif
+	a->pair = -1;
+	a->attr |= ANSI_COLOR;
+	a->bg = buf[pos+1] - '0';
+	pos += 3;
+      }
+      else 
+      {
+	while (pos < x && buf[pos] != ';') pos++;
+	pos++;
+      }
+    }
+  }
+  pos = x;
+  return pos;
+}
+
 static int
 fill_buffer (FILE *f, long *last_pos, long offset, unsigned char *buf, 
 	     unsigned char *fmt, size_t blen, int *buf_ready)
@@ -889,6 +1002,17 @@ fill_buffer (FILE *f, long *last_pos, long offset, unsigned char *buf,
 	else			/* ^H */
 	  *fmt++ = *p++;
       }
+      else if (*p == '\033' && *(p+1) == '[' && is_ansi (p + 2))
+      {
+	while (*p++ != 'm')	/* skip ANSI sequence */
+	  ;
+      }
+      else if (*p == '\033' && *(p+1) == ']' && check_attachment_marker ((char *) p) == 0)
+      {
+	dprint (2, (debugfile, "fill_buffer: Seen attachment marker.\n"));
+	while (*p++ != '\a')	/* skip pseudo-ANSI sequence */
+	  ;
+      }
       else
 	*fmt++ = *p++;
     }
@@ -897,76 +1021,159 @@ fill_buffer (FILE *f, long *last_pos, long offset, unsigned char *buf,
   return b_read;
 }
 
-static int is_ansi (unsigned char *buf)
+
+static int format_line (struct line_t **lineInfo, int n, unsigned char *buf,
+			int flags, ansi_attr *pa, int cnt,
+			int *pspace, int *pvch, int *pcol, int *pspecial)
 {
-  while (buf && (isdigit(*buf) || *buf == ';'))
-    buf++;
-  return (*buf == 'm');
-}
+  int space = -1; /* index of the last space or TAB */
+  int col = option (OPTMARKERS) ? (*lineInfo)[n].continuation : 0;
+  int ch, vch, k, last_special = -1, special = 0, t;
+  wchar_t wc;
+  mbstate_t mbstate;
 
-static int grok_ansi(unsigned char *buf, int pos, ansi_attr *a)
-{
-  int x = pos;
+  int wrap_cols = COLS - WrapMargin;
+  
+  if (wrap_cols <= 0)
+    wrap_cols = COLS;
+  
+  /* FIXME: this should come from lineInfo */
+  memset(&mbstate, 0, sizeof(mbstate));
 
-  while (isdigit(buf[x]) || buf[x] == ';')
-    x++;
-
-  /* Character Attributes */
-  if (a != NULL && buf[x] == 'm')
+  for (ch = 0, vch = 0; ch < cnt; ch += k, vch += k)
   {
-    while (pos < x)
+    /* Handle ANSI sequences */
+    while (cnt-ch >= 2 && buf[ch] == '\033' && buf[ch+1] == '[' &&
+	   is_ansi (buf+ch+2))
+      ch = grok_ansi (buf, ch+2, pa) + 1;
+
+    while (cnt-ch >= 2 && buf[ch] == '\033' && buf[ch+1] == ']' &&
+	   check_attachment_marker ((char *) buf+ch) == 0)
     {
-      if (buf[pos] == '1' && (pos+1 == x || buf[pos+1] == ';'))
+      while (buf[ch++] != '\a')
+	if (ch >= cnt)
+	  break;
+    }
+
+    /* is anything left to do? */
+    if (ch >= cnt)
+      break;
+    
+    k = mbrtowc (&wc, (char *)buf+ch, cnt-ch, &mbstate);
+    if (k == -2 || k == -1)
+    {
+      dprint (1, (debugfile, "%s:%d: mbrtowc returned %d; errno = %d.\n",
+		  __FILE__, __LINE__, k, errno));
+      if (col + 4 > wrap_cols)
+	break;
+      col += 4;
+      if (pa)
+	printw ("\\%03o", buf[ch]);
+      k = 1;
+      continue;
+    }
+    if (k == 0)
+      k = 1;
+
+    /* Handle backspace */
+    special = 0;
+    if (IsWPrint (wc))
+    {
+      wchar_t wc1;
+      mbstate_t mbstate1;
+      int k1, k2;
+
+      while ((wc1 = 0, mbstate1 = mbstate,
+	      k1 = k + mbrtowc (&wc1, (char *)buf+ch+k, cnt-ch-k, &mbstate1),
+	      k1 - k > 0 && wc1 == '\b') &&
+	     (wc1 = 0,
+	      k2 = mbrtowc (&wc1, (char *)buf+ch+k1, cnt-ch-k1, &mbstate1),
+	      k2 > 0 && IsWPrint (wc1)))
       {
-	a->attr |= ANSI_BOLD;
-	pos += 2;
-      } 
-      else if (buf[pos] == '4' && (pos+1 == x || buf[pos+1] == ';'))
-      {
-	a->attr |= ANSI_UNDERLINE;
-	pos += 2;
-      }
-      else if (buf[pos] == '5' && (pos+1 == x || buf[pos+1] == ';'))
-      {
-	a->attr |= ANSI_BLINK;
-	pos += 2;
-      }
-      else if (buf[pos] == '7' && (pos+1 == x || buf[pos+1] == ';'))
-      {
-	a->attr |= ANSI_REVERSE;
-	pos += 2;
-      }
-      else if (buf[pos] == '0' && (pos+1 == x || buf[pos+1] == ';'))
-      {
-#ifdef HAVE_COLOR
-	if (a->pair != -1)
-	  mutt_free_color(a->fg,a->bg);
-#endif
-	a->attr = ANSI_OFF;
-	a->pair = -1;
-	pos += 2;
-      }
-      else if (buf[pos] == '3' && isdigit(buf[pos+1]))
-      {
-	a->attr |= ANSI_COLOR;
-	a->fg = buf[pos+1] - '0';
-	pos += 3;
-      }
-      else if (buf[pos] == '4' && isdigit(buf[pos+1]))
-      {
-	a->attr |= ANSI_COLOR;
-	a->bg = buf[pos+1] - '0';
-	pos += 3;
-      }
-      else 
-      {
-	while (pos < x && buf[pos] != ';') pos++;
-	pos++;
+	if (wc == wc1)
+	{
+	  special |= (wc == '_' && special & A_UNDERLINE)
+	    ? A_UNDERLINE : A_BOLD;
+	}
+	else if (wc == '_' || wc1 == '_')
+	{
+	  special |= A_UNDERLINE;
+	  wc = (wc1 == '_') ? wc : wc1;
+	}
+	else
+	{
+	  /* special = 0; / * overstrike: nothing to do! */
+	  wc = wc1;
+	}
+	ch += k1;
+	k = k2;
+	mbstate = mbstate1;
       }
     }
+
+    if (pa &&
+	((flags & (M_SHOWCOLOR | M_SEARCH | M_PAGER_MARKER)) ||
+	 special || last_special || pa->attr))
+    {
+      resolve_color (*lineInfo, n, vch, flags, special, pa);
+      last_special = special;
+    }
+
+    if (IsWPrint (wc))
+    {
+      if (wc == ' ')
+	space = ch;
+      t = wcwidth (wc);
+      if (col + t > wrap_cols)
+	break;
+      col += t;
+      if (pa)
+	mutt_addwch (wc);
+    }
+    else if (wc == '\n')
+      break;
+    else if (wc == '\t')
+    {
+      space = ch;
+      t = (col & ~7) + 8;
+      if (t > wrap_cols)
+	break;
+      if (pa)
+	for (; col < t; col++)
+	  addch (' ');
+      else
+	col = t;
+    }
+    else if (wc < 0x20 || wc == 0x7f)
+    {
+      if (col + 2 > wrap_cols)
+	break;
+      col += 2;
+      if (pa)
+	printw ("^%c", ('@' + wc) & 0x7f);
+    }
+    else if (wc < 0x100)
+    {
+      if (col + 4 > wrap_cols)
+	break;
+      col += 4;
+      if (pa)
+	printw ("\\%03o", wc);
+    }
+    else
+    {
+      if (col + 1 > wrap_cols)
+	break;
+      ++col;
+      if (pa)
+	addch (replacement_char ());
+    }
   }
-  pos = x;
-  return pos;
+  *pspace = space;
+  *pcol = col;
+  *pvch = vch;
+  *pspecial = special;
+  return ch;
 }
 
 /*
@@ -992,10 +1199,10 @@ display_line (FILE *f, long *last_pos, struct line_t **lineInfo, int n,
 	      int *q_level, int *force_redraw, regex_t *SearchRE)
 {
   unsigned char buf[LONG_STRING], fmt[LONG_STRING];
-  unsigned char *buf_ptr = buf, c;
-  int ch, vch, t, col, cnt, b_read;
+  unsigned char *buf_ptr = buf;
+  int ch, vch, col, cnt, b_read;
   int buf_ready = 0, change_last = 0;
-  int special = 0, last_special = 0;
+  int special;
   int offset;
   int def_color;
   int m;
@@ -1010,7 +1217,7 @@ display_line (FILE *f, long *last_pos, struct line_t **lineInfo, int n,
 
   if (*last == *max)
   {
-    safe_realloc ((void **)lineInfo, sizeof (struct line_t) * (*max += LINES));
+    safe_realloc (lineInfo, sizeof (struct line_t) * (*max += LINES));
     for (ch = *last; ch < *max ; ch++)
     {
       memset (&((*lineInfo)[ch]), 0, sizeof (struct line_t));
@@ -1034,7 +1241,7 @@ display_line (FILE *f, long *last_pos, struct line_t **lineInfo, int n,
 	return (-1);
       }
 
-      resolve_types ((char *) fmt, *lineInfo, n, *last,
+      resolve_types ((char *) fmt, (char *) buf, *lineInfo, n, *last,
 		      QuoteList, q_level, force_redraw, flags & M_SHOWCOLOR);
 
       /* avoid race condition for continuation lines when scrolling up */
@@ -1083,7 +1290,7 @@ display_line (FILE *f, long *last_pos, struct line_t **lineInfo, int n,
     while (regexec (SearchRE, (char *) fmt + offset, 1, pmatch, (offset ? REG_NOTBOL : 0)) == 0)
     {
       if (++((*lineInfo)[n].search_cnt) > 1)
-	safe_realloc ((void **) &((*lineInfo)[n].search),
+	safe_realloc (&((*lineInfo)[n].search),
 		      ((*lineInfo)[n].search_cnt) * sizeof (struct syntax_t));
       else
 	(*lineInfo)[n].search = safe_malloc (sizeof (struct syntax_t));
@@ -1121,73 +1328,13 @@ display_line (FILE *f, long *last_pos, struct line_t **lineInfo, int n,
   }
 
   /* now chose a good place to break the line */
-
-  ch = -1; /* index of the last space or TAB */
-  cnt = 0;
-  col = option (OPTMARKERS) ? (*lineInfo)[n].continuation : 0;
-  while (col < COLS && cnt < b_read)
-  {
-    c = *buf_ptr++;
-    if (c == '\n')
-      break;
-
-    while (*buf_ptr == '\010' && cnt + 2 < b_read)
-    {
-      cnt += 2;
-      buf_ptr += 2;
-      c = buf[cnt];
-    }
-
-    if (*buf_ptr == '\033' && *(buf_ptr + 1) && *(buf_ptr + 1) == '[' && is_ansi (buf_ptr+2))
-    {
-      cnt = grok_ansi(buf, cnt+3, NULL);
-      cnt++;
-      buf_ptr = buf + cnt;
-      continue;
-    }
-
-    if (c == '\t')
-    {
-      ch = cnt;
-      /* expand TABs */
-      if ((t = (col & ~7) + 8) < COLS)
-      {
-	col = t;
-	cnt++;
-      }
-      else
-	break;
-    }
-    else if (IsPrint (c))
-    {
-      if (c == ' ')
-	ch = cnt;
-      col++;
-      cnt++;
-    }
-    else if (iscntrl (c) && c < '@')
-    {
-      if (c == '\r' && *buf_ptr == '\n')
-	cnt++;
-      else if (col < COLS - 1)
-      {
-	col += 2;
-	cnt++;
-      }
-      else
-	break;
-    }
-    else
-    {
-      col++;
-      cnt++;
-    }
-  }
+  cnt = format_line (lineInfo, n, buf, flags, 0, b_read, &ch, &vch, &col, &special);
+  buf_ptr = buf + cnt;
 
   /* move the break point only if smart_wrap is set */
   if (option (OPTWRAP))
   {
-    if (col == COLS)
+    if (cnt < b_read)
     {
       if (ch != -1 && buf[cnt] != ' ' && buf[cnt] != '\t' && buf[cnt] != '\n' && buf[cnt] != '\r')
       {
@@ -1219,125 +1366,8 @@ display_line (FILE *f, long *last_pos, struct line_t **lineInfo, int n,
   if (!(flags & M_SHOW))
     return 0;
 
-  if (flags & M_SHOWCOLOR)
-  {
-    m = ((*lineInfo)[n].continuation) ? ((*lineInfo)[n].syntax)[0].first : n;
-    if ((*lineInfo)[m].type == MT_COLOR_HEADER)
-      def_color = ((*lineInfo)[m].syntax)[0].color;
-    else
-      def_color = (*lineInfo)[m].type;
-
-    attrset (def_color);
-#ifdef HAVE_BKGDSET
-    bkgdset (def_color | ' ');
-#endif
-    clrtoeol ();
-    SETCOLOR (MT_COLOR_NORMAL);
-    BKGDSET (MT_COLOR_NORMAL);
-  }
-  else
-    clrtoeol ();
-  
   /* display the line */
-  col = option (OPTMARKERS) ? (*lineInfo)[n].continuation : 0;
-  for (ch = 0, vch = 0; ch < cnt; ch++, vch++)
-  {
-    special = 0;
-
-    c = buf[ch];
-    while (buf[ch+1] == '\010' && ch+2 < b_read)
-    {
-      if (buf[ch+2] == c)
-      {
-	special = (c == '_' && last_special == A_UNDERLINE)
-	  ? A_UNDERLINE : A_BOLD;
-	ch += 2;
-      }
-      else if (buf[ch] == '_' || buf[ch+2] == '_')
-      {
-	special = A_UNDERLINE;
-	ch += 2;
-	c = (buf[ch] == '_') ? buf[ch-2] : buf[ch];
-      }
-      else
-      {
-	special = 0; /* overstrike: nothing to do! */
-	ch += 2;
-	c = buf[ch];
-      }
-      last_special = special;
-    }
-
-    /* Handle ANSI sequences */
-    if (c == '\033' && buf[ch+1] == '[' && is_ansi (buf+ch+2))
-    {
-      ch = grok_ansi(buf, ch+2, &a);
-      c = buf[ch];
-      continue;
-    }
-
-    if (c == '\t')
-    {
-      if ((flags & (M_SHOWCOLOR | M_SEARCH | M_PAGER_MARKER)) || last_special
-	  || a.attr)
-      {
-	resolve_color (*lineInfo, n, vch, flags, special, &a);
-	if (!special)
-	  last_special = 0;
-      }
-
-      t = (col & ~7) + 8;
-      while (col < t)
-      {
-	addch (' ');
-	col++;
-      }
-    }
-    else if (IsPrint (c))
-    {
-      if ((flags & (M_SHOWCOLOR | M_SEARCH | M_PAGER_MARKER)) || special
-	  || last_special || a.attr)
-	resolve_color (*lineInfo, n, vch, flags, special, &a);
-      if (!special)
-	last_special = 0;
-
-      addch (c);
-      col++;
-    }
-    else if (iscntrl (c) && (c < '@' || c == 127))
-    {
-      if ((c != '\r' && c !='\n') || (buf[ch+1] != '\n' && buf[ch+1] != '\0'))
-      {
-	if ((flags & (M_SHOWCOLOR | M_SEARCH | M_PAGER_MARKER)) || last_special
-	    || a.attr)
-	{
-	  resolve_color (*lineInfo, n, vch, flags, special, &a);
-	  if (!special)
-	    last_special = 0;
-	}
-
-	addch ('^');
-	addch ((c + '@') & 127);
-	col += 2;
-      }
-    }
-    else
-    {
-      if ((flags & (M_SHOWCOLOR | M_SEARCH | M_PAGER_MARKER)) || last_special
-	  || a.attr)
-      {
-	resolve_color (*lineInfo, n, vch, flags, special, &a);
-	if (!special)
-	  last_special = 0;
-      }
-
-      if (ISSPACE (c))
-	addch (c);		/* unbreakable space */
-      else
-	addch ('?');
-      col++;
-    }
-  }
+  format_line (lineInfo, n, buf, flags, &a, cnt, &ch, &vch, &col, &special);
 
   /* avoid a bug in ncurses... */
 #ifndef USE_SLANG_CURSES
@@ -1349,9 +1379,28 @@ display_line (FILE *f, long *last_pos, struct line_t **lineInfo, int n,
 #endif
 
   /* end the last color pattern (needed by S-Lang) */
-  if (last_special || (col != COLS && (flags & (M_SHOWCOLOR | M_SEARCH))))
+  if (special || (col != COLS && (flags & (M_SHOWCOLOR | M_SEARCH))))
     resolve_color (*lineInfo, n, vch, flags, 0, &a);
           
+  /*
+   * Fill the blank space at the end of the line with the prevailing color.
+   * ncurses does an implicit clrtoeol() when you do addch('\n') so we have
+   * to make sure to reset the color *after* that
+   */
+  if (flags & M_SHOWCOLOR)
+  {
+    m = ((*lineInfo)[n].continuation) ? ((*lineInfo)[n].syntax)[0].first : n;
+    if ((*lineInfo)[m].type == MT_COLOR_HEADER)
+      def_color = ((*lineInfo)[m].syntax)[0].color;
+    else
+      def_color = ColorDefs[ (*lineInfo)[m].type ];
+
+    attrset (def_color);
+#ifdef HAVE_BKGDSET
+    bkgdset (def_color | ' ');
+#endif
+  }
+
   /* ncurses always wraps lines when you get to the right side of the
    * screen, but S-Lang seems to only wrap if the next character is *not*
    * a newline (grr!).
@@ -1361,6 +1410,17 @@ display_line (FILE *f, long *last_pos, struct line_t **lineInfo, int n,
 #endif
       addch ('\n');
 
+  /*
+   * reset the color back to normal.  This *must* come after the
+   * addch('\n'), otherwise the color for this line will not be
+   * filled to the right margin.
+   */
+  if (flags & M_SHOWCOLOR)
+  {
+    SETCOLOR(MT_COLOR_NORMAL);
+    BKGDSET(MT_COLOR_NORMAL);
+  }
+
   /* build a return code */
   if (!(flags & M_SHOW))
     flags = 0;
@@ -1407,7 +1467,8 @@ mutt_pager (const char *banner, const char *fname, int flags, pager_t *extra)
 {
   static char searchbuf[STRING];
   char buffer[LONG_STRING];
-  char helpstr[SHORT_STRING];
+  char helpstr[SHORT_STRING*2];
+  char tmphelp[SHORT_STRING*2];
   int maxLine, lastLine = 0;
   struct line_t *lineInfo;
   struct q_class_t *QuoteList = NULL;
@@ -1473,15 +1534,15 @@ mutt_pager (const char *banner, const char *fname, int flags, pager_t *extra)
   mutt_compile_help (helpstr, sizeof (helpstr), MENU_PAGER, PagerHelp);
   if (IsHeader (extra))
   {
+    strfcpy (tmphelp, helpstr, sizeof (tmphelp));
     mutt_compile_help (buffer, sizeof (buffer), MENU_PAGER, PagerHelpExtra);
-    strcat (helpstr, "  ");
-    strcat (helpstr, buffer);
+    snprintf (helpstr, sizeof (helpstr), "%s %s", tmphelp, buffer);
   }
   if (!InHelp)
   {
+    strfcpy (tmphelp, helpstr, sizeof (tmphelp));
     mutt_make_help (buffer, sizeof (buffer), _("Help"), MENU_PAGER, OP_HELP);
-    strcat (helpstr, "  ");
-    strcat (helpstr, buffer);
+    snprintf (helpstr, sizeof (helpstr), "%s %s", tmphelp, buffer);
   }
 
   while (ch != -1)
@@ -1530,7 +1591,8 @@ mutt_pager (const char *banner, const char *fname, int flags, pager_t *extra)
       if (option (OPTHELP))
       {
 	SETCOLOR (MT_COLOR_STATUS);
-	mvprintw (helpoffset, 0, "%-*.*s", COLS, COLS, helpstr);
+	move (helpoffset, 0);
+	mutt_paddstr (COLS, helpstr);
 	SETCOLOR (MT_COLOR_NORMAL);
       }
 
@@ -1547,7 +1609,7 @@ mutt_pager (const char *banner, const char *fname, int flags, pager_t *extra)
 	lines = Resize->line;
 	redraw |= REDRAW_SIGWINCH;
 
-	safe_free ((void **) &Resize);
+	FREE (&Resize);
       }
 #endif
 
@@ -1654,8 +1716,9 @@ mutt_pager (const char *banner, const char *fname, int flags, pager_t *extra)
 			   COLS - 9 < sizeof (buffer) ? COLS - 9: sizeof (buffer),
 			   NONULL (PagerFmt), Context, extra->bdy->hdr, M_FORMAT_MAKEPRINT);
       }
-      printw ("%-*.*s -- (", COLS-10, COLS-10, 
-	      IsHeader (extra) || IsMsgAttach (extra) ? buffer : banner);
+      mutt_paddstr (COLS-10, IsHeader (extra) || IsMsgAttach (extra) ?
+		    buffer : banner);
+      addstr (" -- (");
       if (last_pos < sb.st_size - 1)
 	printw ("%d%%)", (int) (100 * last_offset / sb.st_size));
       else
@@ -1675,7 +1738,7 @@ mutt_pager (const char *banner, const char *fname, int flags, pager_t *extra)
  
       move (indexoffset + (option (OPTSTATUSONTOP) ? 0 : (indexlen - 1)), 0);
       SETCOLOR (MT_COLOR_STATUS);
-      printw ("%-*.*s", COLS, COLS, buffer);
+      mutt_paddstr (COLS, buffer);
       SETCOLOR (MT_COLOR_NORMAL);
     }
 
@@ -1726,10 +1789,10 @@ mutt_pager (const char *banner, const char *fname, int flags, pager_t *extra)
 	  lineInfo[i].search_cnt = -1;
 	  lineInfo[i].quote = NULL;
 
-	  safe_realloc ((void **)&(lineInfo[i].syntax),
+	  safe_realloc (&(lineInfo[i].syntax),
 			sizeof (struct syntax_t));
 	  if (SearchCompiled && lineInfo[i].search)
-	      safe_free ((void **) &(lineInfo[i].search));
+	      FREE (&(lineInfo[i].search));
 	}
 
 	lastLine = 0;
@@ -1740,6 +1803,7 @@ mutt_pager (const char *banner, const char *fname, int flags, pager_t *extra)
       }
 
       SigWinch = 0;
+      clearok(stdscr,TRUE);/*force complete redraw*/
       continue;
     }
 #endif
@@ -1808,7 +1872,10 @@ mutt_pager (const char *banner, const char *fname, int flags, pager_t *extra)
 	break;
 
       case OP_PAGER_TOP:
-	topline = 0;
+        if (topline)
+	  topline = 0;
+      	else
+	  mutt_error _("Top of message is shown.");
 	break;
 
       case OP_HALF_UP:
@@ -1840,6 +1907,7 @@ mutt_pager (const char *banner, const char *fname, int flags, pager_t *extra)
       case OP_SEARCH_OPPOSITE:
 	if (SearchCompiled)
 	{
+search_next:
 	  if ((!SearchBack && ch==OP_SEARCH_NEXT) ||
 	      (SearchBack &&ch==OP_SEARCH_OPPOSITE))
 	  {
@@ -1882,24 +1950,44 @@ mutt_pager (const char *banner, const char *fname, int flags, pager_t *extra)
 
       case OP_SEARCH:
       case OP_SEARCH_REVERSE:
+        strfcpy (buffer, searchbuf, sizeof (buffer));
+	if (mutt_get_field ((SearchBack ? _("Reverse search: ") :
+			  _("Search: ")), buffer, sizeof (buffer),
+			  M_CLEAR) != 0)
+	  break;
+
+	if (!strcmp (buffer, searchbuf))
+	{
+	  if (SearchCompiled)
+	  {
+	    /* do an implicit search-next */
+	    if (ch == OP_SEARCH)
+	      ch = OP_SEARCH_NEXT;
+	    else
+	      ch = OP_SEARCH_OPPOSITE;
+
+	    goto search_next;
+	  }
+	}
+      
+        if (!buffer[0])
+	  break;
+      
+	strfcpy (searchbuf, buffer, sizeof (searchbuf));
+
 	/* leave SearchBack alone if ch == OP_SEARCH_NEXT */
 	if (ch == OP_SEARCH)
 	  SearchBack = 0;
 	else if (ch == OP_SEARCH_REVERSE)
 	  SearchBack = 1;
 
-	if (mutt_get_field ((SearchBack ? _("Reverse search: ") :
-			  _("Search: ")), searchbuf, sizeof (searchbuf),
-			  M_CLEAR) != 0 || !searchbuf[0])
-	  break;
-
 	if (SearchCompiled)
 	{
 	  regfree (&SearchRE);
 	  for (i = 0; i < lastLine; i++)
 	  {
 	    if (lineInfo[i].search)
-	      safe_free ((void **) &(lineInfo[i].search));
+	      FREE (&(lineInfo[i].search));
 	    lineInfo[i].search_cnt = -1;
 	  }
 	}
@@ -1913,7 +2001,7 @@ mutt_pager (const char *banner, const char *fname, int flags, pager_t *extra)
 	  {
 	    /* cleanup */
 	    if (lineInfo[i].search)
-	      safe_free ((void **) &(lineInfo[i].search));
+	      FREE (&(lineInfo[i].search));
 	    lineInfo[i].search_cnt = -1;
 	  }
 	  SearchFlag = 0;
@@ -2098,6 +2186,11 @@ mutt_pager (const char *banner, const char *fname, int flags, pager_t *extra)
       case OP_DELETE:
 	CHECK_MODE(IsHeader (extra));
 	CHECK_READONLY;
+
+#ifdef USE_IMAP
+CHECK_IMAP_ACL(IMAP_ACL_DELETE);
+#endif
+
 	mutt_set_flag (Context, extra->hdr, M_DELETE, 1);
         if (option (OPTDELETEUNTAG))
 	  mutt_set_flag (Context, extra->hdr, M_TAG, 0);
@@ -2114,6 +2207,10 @@ mutt_pager (const char *banner, const char *fname, int flags, pager_t *extra)
 	CHECK_MODE(IsHeader (extra));
 	CHECK_READONLY;
 
+#ifdef USE_IMAP
+CHECK_IMAP_ACL(IMAP_ACL_DELETE);
+#endif
+
 	r = mutt_thread_set_flag (extra->hdr, M_DELETE, 1,
 				  ch == OP_DELETE_THREAD ? 0 : 1);
 
@@ -2193,9 +2290,9 @@ mutt_pager (const char *banner, const char *fname, int flags, pager_t *extra)
 	    lineInfo[i].search_cnt = -1;
 	    lineInfo[i].quote = NULL;
 
-	    safe_realloc ((void **)&(lineInfo[i].syntax), sizeof (struct syntax_t));
+	    safe_realloc (&(lineInfo[i].syntax), sizeof (struct syntax_t));
 	    if (SearchCompiled && lineInfo[i].search)
-		safe_free ((void **) &(lineInfo[i].search));
+		FREE (&(lineInfo[i].search));
 	  }
 
 	  if (SearchCompiled)
@@ -2232,6 +2329,20 @@ mutt_pager (const char *banner, const char *fname, int flags, pager_t *extra)
       case OP_FLAG_MESSAGE:
 	CHECK_MODE(IsHeader (extra));
 	CHECK_READONLY;
+
+#ifdef USE_POP
+	if (Context->magic == M_POP)
+	{
+	  mutt_flushinp ();
+	  mutt_error _("Can't change 'important' flag on POP server.");
+	  break;
+	}
+#endif
+
+#ifdef USE_IMAP
+CHECK_IMAP_ACL(IMAP_ACL_WRITE);
+#endif
+
 	mutt_set_flag (Context, extra->hdr, M_FLAG, !extra->hdr->flagged);
 	redraw = REDRAW_STATUS | REDRAW_INDEX;
 	if (option (OPTRESOLVE))
@@ -2261,7 +2372,7 @@ mutt_pager (const char *banner, const char *fname, int flags, pager_t *extra)
       case OP_MAIL:
 	CHECK_MODE(IsHeader (extra) && !IsAttach (extra));
         CHECK_ATTACH;      
-	ci_send_message (0, NULL, NULL, NULL, NULL);
+	ci_send_message (0, NULL, NULL, extra->ctx, extra->hdr);
 	redraw = REDRAW_FULL;
 	break;
 
@@ -2307,7 +2418,7 @@ mutt_pager (const char *banner, const char *fname, int flags, pager_t *extra)
 	break;
 
       case OP_FORWARD_MESSAGE:
-	CHECK_MODE(IsHeader (extra) && !IsMsgAttach (extra));
+	CHECK_MODE(IsHeader (extra) || IsMsgAttach (extra));
         CHECK_ATTACH;
         if (IsMsgAttach (extra))
 	  mutt_attach_forward (extra->fp, extra->hdr, extra->idx,
@@ -2317,37 +2428,38 @@ mutt_pager (const char *banner, const char *fname, int flags, pager_t *extra)
 	redraw = REDRAW_FULL;
 	break;
 
-#ifdef HAVE_PGP      
       case OP_DECRYPT_SAVE:
-#endif
+        if (!WithCrypto)
+        {
+          ch = -1;
+          break;
+        }
+	/* fall through */
       case OP_SAVE:
 	if (IsAttach (extra))
 	{
-	  mutt_save_attachment_list (extra->fp, 0, extra->bdy, extra->hdr);
+	  mutt_save_attachment_list (extra->fp, 0, extra->bdy, extra->hdr, NULL);
 	  break;
 	}
 	/* fall through */
       case OP_COPY_MESSAGE:
       case OP_DECODE_SAVE:
       case OP_DECODE_COPY:
-#ifdef HAVE_PGP
       case OP_DECRYPT_COPY:
-#endif
+        if (!WithCrypto && ch == OP_DECRYPT_COPY)
+        {
+          ch = -1;
+          break;
+        }
 	CHECK_MODE(IsHeader (extra));
 	if (mutt_save_message (extra->hdr,
-#ifdef HAVE_PGP
 			       (ch == OP_DECRYPT_SAVE) ||
-#endif			       
 			       (ch == OP_SAVE) || (ch == OP_DECODE_SAVE),
 			       (ch == OP_DECODE_SAVE) || (ch == OP_DECODE_COPY),
-#ifdef HAVE_PGP
 			       (ch == OP_DECRYPT_SAVE) || (ch == OP_DECRYPT_COPY) ||
-#endif
 			       0,
 			       &redraw) == 0 && (ch == OP_SAVE || ch == OP_DECODE_SAVE
-#ifdef HAVE_PGP
 						 || ch == OP_DECRYPT_SAVE
-#endif
 						 ))
 	{
 	  if (option (OPTRESOLVE))
@@ -2380,6 +2492,11 @@ mutt_pager (const char *banner, const char *fname, int flags, pager_t *extra)
       case OP_TOGGLE_NEW:
 	CHECK_MODE(IsHeader (extra));
 	CHECK_READONLY;
+
+#ifdef USE_IMAP
+CHECK_IMAP_ACL(IMAP_ACL_SEEN);
+#endif
+
 	if (extra->hdr->read || extra->hdr->old)
 	  mutt_set_flag (Context, extra->hdr, M_NEW, 1);
 	else if (!first)
@@ -2397,6 +2514,11 @@ mutt_pager (const char *banner, const char *fname, int flags, pager_t *extra)
       case OP_UNDELETE:
 	CHECK_MODE(IsHeader (extra));
 	CHECK_READONLY;
+
+#ifdef USE_IMAP
+CHECK_IMAP_ACL(IMAP_ACL_DELETE);
+#endif
+
 	mutt_set_flag (Context, extra->hdr, M_DELETE, 0);
 	redraw = REDRAW_STATUS | REDRAW_INDEX;
 	if (option (OPTRESOLVE))
@@ -2411,6 +2533,10 @@ mutt_pager (const char *banner, const char *fname, int flags, pager_t *extra)
 	CHECK_MODE(IsHeader (extra));
 	CHECK_READONLY;
 
+#ifdef USE_IMAP
+CHECK_IMAP_ACL(IMAP_ACL_DELETE);
+#endif
+
 	r = mutt_thread_set_flag (extra->hdr, M_DELETE, 0,
 				  ch == OP_UNDELETE_THREAD ? 0 : 1);
 
@@ -2434,7 +2560,17 @@ mutt_pager (const char *banner, const char *fname, int flags, pager_t *extra)
 	mutt_version ();
 	break;
 
+      case OP_BUFFY_LIST:
+	mutt_buffy_list ();
+	break;
+
       case OP_VIEW_ATTACHMENTS:
+        if (flags & M_PAGER_ATTACHMENT)
+        {
+	  ch = -1;
+	  rc = OP_ATTACH_COLLAPSE;
+	  break;
+	}
 	CHECK_MODE(IsHeader (extra));
 	mutt_view_attachments (extra->hdr);
 	if (extra->hdr->attach_del)
@@ -2443,27 +2579,33 @@ mutt_pager (const char *banner, const char *fname, int flags, pager_t *extra)
 	break;
 
 
-
-#ifdef HAVE_PGP
-      case OP_FORGET_PASSPHRASE:
-	mutt_forget_passphrase ();
-	break;
-
       case OP_MAIL_KEY:
+        if (!(WithCrypto & APPLICATION_PGP))
+        {
+          ch = -1;
+          break;
+        }
 	CHECK_MODE(IsHeader(extra));
         CHECK_ATTACH;
 	ci_send_message (SENDKEY, NULL, NULL, extra->ctx, extra->hdr);
 	redraw = REDRAW_FULL;
 	break;
-      
+
+
+      case OP_FORGET_PASSPHRASE:
+	crypt_forget_passphrase ();
+	break;
+
       case OP_EXTRACT_KEYS:
+        if (!WithCrypto)
+        {
+          ch = -1;
+          break;
+        }
         CHECK_MODE(IsHeader(extra));
-        pgp_extract_keys_from_messages(extra->hdr);
+	crypt_extract_keys_from_messages(extra->hdr);
         redraw = REDRAW_FULL;
         break;
-#endif /* HAVE_PGP */
-
-
 
       default:
 	ch = -1;
@@ -2479,16 +2621,16 @@ mutt_pager (const char *banner, const char *fname, int flags, pager_t *extra)
   
   for (i = 0; i < maxLine ; i++)
   {
-    safe_free ((void **) &(lineInfo[i].syntax));
+    FREE (&(lineInfo[i].syntax));
     if (SearchCompiled && lineInfo[i].search)
-      safe_free ((void **) &(lineInfo[i].search));
+      FREE (&(lineInfo[i].search));
   }
   if (SearchCompiled)
   {
     regfree (&SearchRE);
     SearchCompiled = 0;
   }
-  safe_free ((void **) &lineInfo);
+  FREE (&lineInfo);
   if (index)
     mutt_menuDestroy(&index);
   return (rc != -1 ? rc : 0);
