@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 1996-2000 Michael R. Elkins <me@cs.hmc.edu>
+ * Copyright (C) 1996-2002 Michael R. Elkins <me@mutt.org>
  * 
  *     This program is free software; you can redistribute it and/or modify
  *     it under the terms of the GNU General Public License as published by
@@ -24,6 +24,8 @@
 #include "mailbox.h"
 #include "copy.h"
 #include "mx.h"
+#include "mutt_crypt.h"
+#include "mutt_idna.h"
 
 #include <ctype.h>
 #include <stdlib.h>
@@ -33,10 +35,9 @@
 #include <sys/stat.h>
 #include <sys/wait.h>
 #include <dirent.h>
-
-#ifdef HAVE_PGP
-#include "pgp.h"
-#endif
+#include <time.h>
+#include <sys/types.h>
+#include <utime.h>
 
 #ifdef MIXMASTER
 #include "remailer.h"
@@ -64,7 +65,7 @@ static int mutt_addrcmp (ADDRESS *a, ADDRESS *b)
 {
   if (!a->mailbox || !b->mailbox)
     return 0;
-  if (mutt_strcasecmp (a->mailbox, b->mailbox))
+  if (ascii_strcasecmp (a->mailbox, b->mailbox))
     return 0;
   return 1;
 }
@@ -165,7 +166,7 @@ static ADDRESS *find_mailing_lists (ADDRESS *t, ADDRESS *c)
   {
     for (; t; t = t->next)
     {
-      if (mutt_is_mail_list (t))
+      if (mutt_is_mail_list (t) && !t->group)
       {
 	if (top)
 	{
@@ -183,13 +184,27 @@ static ADDRESS *find_mailing_lists (ADDRESS *t, ADDRESS *c)
 static int edit_address (ADDRESS **a, /* const */ char *field)
 {
   char buf[HUGE_STRING];
-
-  buf[0] = 0;
-  rfc822_write_address (buf, sizeof (buf), *a);
-  if (mutt_get_field (field, buf, sizeof (buf), M_ALIAS) != 0)
-    return (-1);
-  rfc822_free_address (a);
-  *a = mutt_expand_aliases (mutt_parse_adrlist (NULL, buf));
+  char *err = NULL;
+  int idna_ok = 0;
+  
+  do
+  {
+    buf[0] = 0;
+    mutt_addrlist_to_local (*a);
+    rfc822_write_address (buf, sizeof (buf), *a, 0);
+    if (mutt_get_field (field, buf, sizeof (buf), M_ALIAS) != 0)
+      return (-1);
+    rfc822_free_address (a);
+    *a = mutt_expand_aliases (mutt_parse_adrlist (NULL, buf));
+    if ((idna_ok = mutt_addrlist_to_idna (*a, &err)) != 0)
+    {
+      mutt_error (_("Error: '%s' is a bad IDN."), err);
+      mutt_refresh ();
+      mutt_sleep (2);
+      FREE (&err);
+    }
+  } 
+  while (idna_ok != 0);
   return 0;
 }
 
@@ -219,7 +234,7 @@ static int edit_envelope (ENVELOPE *en)
     buf[0] = 0;
     for (; uh; uh = uh->next)
     {
-      if (mutt_strncasecmp ("subject:", uh->data, 8) == 0)
+      if (ascii_strncasecmp ("subject:", uh->data, 8) == 0)
       {
 	p = uh->data + 8;
 	SKIPWS (p);
@@ -229,7 +244,7 @@ static int edit_envelope (ENVELOPE *en)
   }
   
   if (mutt_get_field ("Subject: ", buf, sizeof (buf), 0) != 0 ||
-      (!buf[0] && query_quadoption (OPT_SUBJECT, _("No subject, abort?")) != 0))
+      (!buf[0] && query_quadoption (OPT_SUBJECT, _("No subject, abort?")) != M_NO))
   {
     mutt_message _("No subject, aborting.");
     return (-1);
@@ -245,11 +260,11 @@ static void process_user_recips (ENVELOPE *env)
 
   for (; uh; uh = uh->next)
   {
-    if (mutt_strncasecmp ("to:", uh->data, 3) == 0)
+    if (ascii_strncasecmp ("to:", uh->data, 3) == 0)
       env->to = rfc822_parse_adrlist (env->to, uh->data + 3);
-    else if (mutt_strncasecmp ("cc:", uh->data, 3) == 0)
+    else if (ascii_strncasecmp ("cc:", uh->data, 3) == 0)
       env->cc = rfc822_parse_adrlist (env->cc, uh->data + 3);
-    else if (mutt_strncasecmp ("bcc:", uh->data, 4) == 0)
+    else if (ascii_strncasecmp ("bcc:", uh->data, 4) == 0)
       env->bcc = rfc822_parse_adrlist (env->bcc, uh->data + 4);
   }
 }
@@ -265,21 +280,23 @@ static void process_user_header (ENVELOPE *env)
 
   for (; uh; uh = uh->next)
   {
-    if (mutt_strncasecmp ("from:", uh->data, 5) == 0)
+    if (ascii_strncasecmp ("from:", uh->data, 5) == 0)
     {
       /* User has specified a default From: address.  Remove default address */
       rfc822_free_address (&env->from);
       env->from = rfc822_parse_adrlist (env->from, uh->data + 5);
     }
-    else if (mutt_strncasecmp ("reply-to:", uh->data, 9) == 0)
+    else if (ascii_strncasecmp ("reply-to:", uh->data, 9) == 0)
     {
       rfc822_free_address (&env->reply_to);
       env->reply_to = rfc822_parse_adrlist (env->reply_to, uh->data + 9);
     }
-    else if (mutt_strncasecmp ("to:", uh->data, 3) != 0 &&
-	     mutt_strncasecmp ("cc:", uh->data, 3) != 0 &&
-	     mutt_strncasecmp ("bcc:", uh->data, 4) != 0 &&
-	     mutt_strncasecmp ("subject:", uh->data, 8) != 0)
+    else if (ascii_strncasecmp ("message-id:", uh->data, 11) == 0)
+      mutt_str_replace (&env->message_id, uh->data + 11);
+    else if (ascii_strncasecmp ("to:", uh->data, 3) != 0 &&
+	     ascii_strncasecmp ("cc:", uh->data, 3) != 0 &&
+	     ascii_strncasecmp ("bcc:", uh->data, 4) != 0 &&
+	     ascii_strncasecmp ("subject:", uh->data, 8) != 0)
     {
       if (last)
       {
@@ -319,7 +336,7 @@ void mutt_forward_intro (FILE *fp, HEADER *cur)
   
   fputs ("----- Forwarded message from ", fp);
   buffer[0] = 0;
-  rfc822_write_address (buffer, sizeof (buffer), cur->env->from);
+  rfc822_write_address (buffer, sizeof (buffer), cur->env->from, 1);
   fputs (buffer, fp);
   fputs (" -----\n\n", fp);
 }
@@ -334,14 +351,14 @@ static int include_forward (CONTEXT *ctx, HEADER *cur, FILE *out)
 {
   int chflags = CH_DECODE, cmflags = 0;
   
+  mutt_parse_mime_message (ctx, cur);
+  mutt_message_hook (ctx, cur, M_MESSAGEHOOK);
 
-#ifdef HAVE_PGP
-  if ((cur->pgp & PGPENCRYPT) && option (OPTFORWDECODE))
+  if (WithCrypto && (cur->security & ENCRYPT) && option (OPTFORWDECODE))
   {
     /* make sure we have the user's passphrase before proceeding... */
-    pgp_valid_passphrase ();
+    crypt_valid_passphrase (cur->security);
   }
-#endif /* HAVE_PGP */
 
   mutt_forward_intro (out, cur);
 
@@ -356,7 +373,7 @@ static int include_forward (CONTEXT *ctx, HEADER *cur, FILE *out)
   }
   if (option (OPTFORWQUOTE))
     cmflags |= M_CM_PREFIX;
-  mutt_parse_mime_message (ctx, cur);
+
   mutt_copy_message (out, ctx, cur, cmflags, chflags);
   mutt_forward_trailer (out);
   return 0;
@@ -389,51 +406,40 @@ static int include_reply (CONTEXT *ctx, HEADER *cur, FILE *out)
   int cmflags = M_CM_PREFIX | M_CM_DECODE | M_CM_CHARCONV;
   int chflags = CH_DECODE;
 
-#ifdef HAVE_PGP
-  if (cur->pgp)
+  if (WithCrypto && (cur->security & ENCRYPT))
   {
-    if (cur->pgp & PGPENCRYPT)
-    {
-      /* make sure we have the user's passphrase before proceeding... */
-      pgp_valid_passphrase ();
-    }
+    /* make sure we have the user's passphrase before proceeding... */
+    crypt_valid_passphrase (cur->security);
   }
-#endif /* HAVE_PGP */
 
+  mutt_parse_mime_message (ctx, cur);
+  mutt_message_hook (ctx, cur, M_MESSAGEHOOK);
+  
   mutt_make_attribution (ctx, cur, out);
   
   if (!option (OPTHEADER))
     cmflags |= M_CM_NOHEADER;
   if (option (OPTWEED))
   {
-    chflags |= CH_WEED;
+    chflags |= CH_WEED | CH_REORDER;
     cmflags |= M_CM_WEED;
   }
 
-  mutt_parse_mime_message (ctx, cur);
   mutt_copy_message (out, ctx, cur, cmflags, chflags);
-  
+
   mutt_make_post_indent (ctx, cur, out);
   
   return 0;
 }
 
-static int default_to (ADDRESS **to, ENVELOPE *env, int flags)
+static int default_to (ADDRESS **to, ENVELOPE *env, int flags, int hmfupto)
 {
   char prompt[STRING];
-  int i = 0;
 
-  if (flags && env->mail_followup_to)
+  if (flags && env->mail_followup_to && hmfupto == M_YES) 
   {
-    snprintf (prompt, sizeof (prompt), _("Follow-up to %s%s?"),
-	      env->mail_followup_to->mailbox,
-	      env->mail_followup_to->next ? "..." : "");
-
-    if (query_quadoption (OPT_MFUPTO, prompt) == M_YES)
-    {
-      rfc822_append (to, env->mail_followup_to);
-      return 0;
-    }
+    rfc822_append (to, env->mail_followup_to);
+    return 0;
   }
 
   /* Exit now if we're setting up the default Cc list for list-reply
@@ -477,12 +483,19 @@ static int default_to (ADDRESS **to, ENVELOPE *env, int flags)
       snprintf (prompt, sizeof (prompt), _("Reply to %s%s?"),
 		env->reply_to->mailbox, 
 		env->reply_to->next?",...":"");
-      if ((i = query_quadoption (OPT_REPLYTO, prompt)) == M_YES)
+      switch (query_quadoption (OPT_REPLYTO, prompt))
+      {
+      case M_YES:
 	rfc822_append (to, env->reply_to);
-      else if (i == M_NO)
+	break;
+
+      case M_NO:
 	rfc822_append (to, env->from);
-      else
+	break;
+
+      default:
 	return (-1); /* abort */
+      }
     }
     else
       rfc822_append (to, env->reply_to);
@@ -495,23 +508,36 @@ static int default_to (ADDRESS **to, ENVELOPE *env, int flags)
 
 int mutt_fetch_recips (ENVELOPE *out, ENVELOPE *in, int flags)
 {
+  char prompt[STRING];
   ADDRESS *tmp;
+  int hmfupto = -1;
+
+  if ((flags & (SENDLISTREPLY|SENDGROUPREPLY)) && in->mail_followup_to)
+  {
+    snprintf (prompt, sizeof (prompt), _("Follow-up to %s%s?"),
+	      in->mail_followup_to->mailbox,
+	      in->mail_followup_to->next ? ",..." : "");
+
+    if ((hmfupto = query_quadoption (OPT_MFUPTO, prompt)) == -1)
+      return -1;
+  }
+
   if (flags & SENDLISTREPLY)
   {
     tmp = find_mailing_lists (in->to, in->cc);
     rfc822_append (&out->to, tmp);
     rfc822_free_address (&tmp);
 
-    if (in->mail_followup_to &&
-        default_to (&out->cc, in, flags & SENDLISTREPLY) == -1)
+    if (in->mail_followup_to && hmfupto == M_YES &&
+        default_to (&out->cc, in, flags & SENDLISTREPLY, hmfupto) == -1)
       return (-1); /* abort */
   }
   else
   {
-    if (default_to (&out->to, in, flags & SENDGROUPREPLY) == -1)
+    if (default_to (&out->to, in, flags & SENDGROUPREPLY, hmfupto) == -1)
       return (-1); /* abort */
 
-    if ((flags & SENDGROUPREPLY) && !in->mail_followup_to)
+    if ((flags & SENDGROUPREPLY) && (!in->mail_followup_to || hmfupto != M_YES))
     {
       /* if(!mutt_addr_is_user(in->to)) */
       rfc822_append (&out->cc, in->to);
@@ -523,11 +549,14 @@ int mutt_fetch_recips (ENVELOPE *out, ENVELOPE *in, int flags)
 
 LIST *mutt_make_references(ENVELOPE *e)
 {
-  LIST *t, *l;
+  LIST *t = NULL, *l = NULL;
+
+  if (e->references)
+    l = mutt_copy_list (e->references);
+  else
+    l = mutt_copy_list (e->in_reply_to);
   
-  l = mutt_copy_list(e->references);
-  
-  if(e->message_id)
+  if (e->message_id)
   {
     t = mutt_new_list();
     t->data = safe_strdup(e->message_id);
@@ -540,6 +569,8 @@ LIST *mutt_make_references(ENVELOPE *e)
 
 void mutt_fix_reply_recipients (ENVELOPE *env)
 {
+  mutt_expand_aliases_env (env);
+
   if (! option (OPTMETOO))
   {
     /* the order is important here.  do the CC: first so that if the
@@ -567,34 +598,63 @@ void mutt_make_forward_subject (ENVELOPE *env, CONTEXT *ctx, HEADER *cur)
 void mutt_make_misc_reply_headers (ENVELOPE *env, CONTEXT *ctx,
 				    HEADER *cur, ENVELOPE *curenv)
 {
-  LIST *tmp;
-  char buffer[STRING];
-
   if (curenv->real_subj)
   {
     env->subject = safe_malloc (mutt_strlen (curenv->real_subj) + 5);
-    sprintf (env->subject, "Re: %s", curenv->real_subj);
+    sprintf (env->subject, "Re: %s", curenv->real_subj);	/* __SPRINTF_CHECKED__ */
   }
   else
     env->subject = safe_strdup ("Re: your mail");
   
-  /* add the In-Reply-To field */
-  if (InReplyTo)
+}
+
+void mutt_add_to_reference_headers (ENVELOPE *env, ENVELOPE *curenv, LIST ***pp, LIST ***qq)
+{
+  LIST **p = NULL, **q = NULL;
+
+  if (pp) p = *pp;
+  if (qq) q = *qq;
+  
+  if (!p) p = &env->references;
+  if (!q) q = &env->in_reply_to;
+  
+  while (*p) p = &(*p)->next;
+  while (*q) q = &(*q)->next;
+  
+  *p = mutt_make_references (curenv);
+  
+  if (curenv->message_id)
   {
-    strfcpy (buffer, "In-Reply-To: ", sizeof (buffer));
-    mutt_make_string (buffer + 13, sizeof (buffer) - 13, InReplyTo, ctx, cur);
-    tmp = env->userhdrs;
-    while (tmp && tmp->next)
-      tmp = tmp->next;
-    if (tmp)
-    {
-      tmp->next = mutt_new_list ();
-      tmp = tmp->next;
-    }
-    else
-      tmp = env->userhdrs = mutt_new_list ();
-    tmp->data = safe_strdup (buffer);
+    *q = mutt_new_list();
+    (*q)->data = safe_strdup (curenv->message_id);
   }
+  
+  if (pp) *pp = p;
+  if (qq) *qq = q;
+  
+}
+
+static void 
+mutt_make_reference_headers (ENVELOPE *curenv, ENVELOPE *env, CONTEXT *ctx)
+{
+  env->references = NULL;
+  env->in_reply_to = NULL;
+  
+  if (!curenv)
+  {
+    HEADER *h;
+    LIST **p = NULL, **q = NULL;
+    int i;
+    
+    for(i = 0; i < ctx->vcount; i++)
+    {
+      h = ctx->hdrs[ctx->v2r[i]];
+      if (h->tagged)
+	mutt_add_to_reference_headers (env, h->env, &p, &q);
+    }
+  }
+  else
+    mutt_add_to_reference_headers (env, curenv, NULL, NULL);
 }
 
 static int
@@ -648,28 +708,8 @@ envelope_defaults (ENVELOPE *env, CONTEXT *ctx, HEADER *cur, int flags)
       return (-1);
     }
 
-    mutt_fix_reply_recipients (env);
     mutt_make_misc_reply_headers (env, ctx, cur, curenv);
-    
-    if(tag)
-    {
-      HEADER *h;
-      LIST **p;
-
-      env->references = NULL;
-      p = &env->references;
-
-      for(i = 0; i < ctx->vcount; i++)
-      {
-	while(*p) p = &(*p)->next;
-	h = ctx->hdrs[ctx->v2r[i]];
-	if(h->tagged)
-	  *p = mutt_make_references(h->env);
-      }
-    }
-    else
-      env->references = mutt_make_references(curenv);
-
+    mutt_make_reference_headers (tag ? NULL : curenv, env, ctx);
   }
   else if (flags & SENDFORWARD)
     mutt_make_forward_subject (env, ctx, cur);
@@ -719,7 +759,7 @@ generate_body (FILE *tempfp,	/* stream for outgoing message */
   }
   else if (flags & SENDFORWARD)
   {
-    if ((i = query_quadoption (OPT_MIMEFWD, _("Forward MIME encapsulated?"))) == M_YES)
+    if ((i = query_quadoption (OPT_MIMEFWD, _("Forward as attachment?"))) == M_YES)
     {
       BODY *last = msg->content;
 
@@ -766,20 +806,18 @@ generate_body (FILE *tempfp,	/* stream for outgoing message */
     else if (i == -1)
       return -1;
   }
-
-
-
-#ifdef HAVE_PGP
-  else if (flags & SENDKEY) 
+  /* if (WithCrypto && (flags & SENDKEY)) */
+  else if ((WithCrypto & APPLICATION_PGP) && (flags & SENDKEY)) 
   {
     BODY *tmp;
-    if ((tmp = pgp_make_key_attachment (NULL)) == NULL)
+
+    if ((WithCrypto & APPLICATION_PGP)
+        && (tmp = crypt_pgp_make_key_attachment (NULL)) == NULL)
       return -1;
 
     tmp->next = msg->content;
     msg->content = tmp;
   }
-#endif
 
   mutt_clear_error ();
 
@@ -821,7 +859,9 @@ void mutt_set_followup_to (ENVELOPE *e)
 
     if (e->mail_followup_to && !mutt_is_list_recipient (0, e->to, e->cc))
     {
-      if (e->from)
+      if (e->reply_to)
+	from = rfc822_cpy_adr (e->reply_to);
+      else if (e->from)
 	from = rfc822_cpy_adr (e->from);
       else
 	from = mutt_default_from ();
@@ -892,7 +932,7 @@ ADDRESS *mutt_default_from (void)
   {
     adr = rfc822_new_address ();
     adr->mailbox = safe_malloc (mutt_strlen (Username) + mutt_strlen (fqdn) + 2);
-    sprintf (adr->mailbox, "%s@%s", NONULL(Username), NONULL(fqdn));
+    sprintf (adr->mailbox, "%s@%s", NONULL(Username), NONULL(fqdn));	/* __SPRINTF_CHECKED__ */
   }
   else
   {
@@ -951,14 +991,12 @@ static int send_message (HEADER *msg)
 static void encode_descriptions (BODY *b, short recurse)
 {
   BODY *t;
-  char tmp[LONG_STRING];
 
   for (t = b; t; t = t->next)
   {
     if (t->description)
     {
-      rfc2047_encode_string (tmp, sizeof (tmp), (unsigned char *) t->description);
-      mutt_str_replace (&t->description, tmp);
+      rfc2047_encode_string (&t->description);
     }
     if (recurse && t->parts)
       encode_descriptions (t->parts, recurse);
@@ -969,15 +1007,12 @@ static void encode_descriptions (BODY *b, short recurse)
 static void decode_descriptions (BODY *b)
 {
   BODY *t;
-  char tmp[LONG_STRING];
   
   for (t = b; t; t = t->next)
   {
     if (t->description)
     {
-      /* this should really have the same interface as rfc2047_encode_string. */
-      rfc2047_decode (tmp, t->description, sizeof (tmp));
-      mutt_str_replace (&t->description, tmp);
+      rfc2047_decode (&t->description);
     }
     if (t->parts)
       decode_descriptions (t->parts);
@@ -1006,15 +1041,15 @@ ci_send_message (int flags,		/* send mode */
   FILE *tempfp = NULL;
   BODY *pbody;
   int i, killfrom = 0;
+  int fcc_error = 0;
+  int free_clear_content = 0;
 
-#ifdef HAVE_PGP
   BODY *save_content = NULL;
   BODY *clear_content = NULL;
   char *pgpkeylist = NULL;
   /* save current value of "pgp_sign_as" */
   char *signas = NULL;
-  char *signmic = NULL;
-#endif
+  char *tag = NULL, *err = NULL;
 
   int rv = -1;
   
@@ -1032,19 +1067,15 @@ ci_send_message (int flags,		/* send mode */
   }
   
   
-#ifdef HAVE_PGP
-  if (flags & SENDPOSTPONED)
-  {
+  if ((WithCrypto & APPLICATION_PGP) && (flags & SENDPOSTPONED))
     signas = safe_strdup(PgpSignAs);
-    signmic = safe_strdup(PgpSignMicalg);
-  }
-#endif /* HAVE_PGP */
 
-  if (msg)
-  {
-    mutt_expand_aliases_env (msg->env);
-  }
-  else
+  /* Delay expansion of aliases until absolutely necessary--shouldn't
+   * be necessary unless we are prompting the user or about to execute a
+   * send-hook.
+   */
+
+  if (!msg)
   {
     msg = mutt_new_header ();
 
@@ -1073,11 +1104,12 @@ ci_send_message (int flags,		/* send mode */
     pbody->next = msg->content; /* don't kill command-line attachments */
     msg->content = pbody;
     
-    msg->content->type = TYPETEXT;
-    msg->content->subtype = safe_strdup ("plain");
+    mutt_parse_content_type (ContentType, msg->content);
     msg->content->unlink = 1;
     msg->content->use_disp = 0;
     msg->content->disposition = DISPINLINE;
+    if (option (OPTTEXTFLOWED) && msg->content->type == TYPETEXT && !ascii_strcasecmp (msg->content->subtype, "plain"))
+      mutt_set_parameter ("format", "flowed", &msg->content->parameter);
     
     if (!tempfile)
     {
@@ -1105,7 +1137,15 @@ ci_send_message (int flags,		/* send mode */
     /* we shouldn't have to worry about freeing `msg->env->from' before
      * setting it here since this code will only execute when doing some
      * sort of reply.  the pointer will only be set when using the -H command
-     * line option */
+     * line option.
+     *
+     * We shouldn't have to worry about alias expansion here since we are
+     * either replying to a real or postponed message, therefore no aliases
+     * should exist since the user has not had the opportunity to add
+     * addresses to the list.  We just have to ensure the postponed messages
+     * have their aliases expanded.
+     */
+
     msg->env->from = set_reverse_name (cur->env);
   }
 
@@ -1120,6 +1160,7 @@ ci_send_message (int flags,		/* send mode */
       process_user_recips (msg->env);
       process_user_header (msg->env);
     }
+    mutt_expand_aliases_env (msg->env);
   }
   else if (! (flags & (SENDPOSTPONED|SENDRESEND)))
   {
@@ -1130,6 +1171,9 @@ ci_send_message (int flags,		/* send mode */
     if (option (OPTHDRS))
       process_user_recips (msg->env);
 
+    /* Expand aliases and remove duplicates/crossrefs */
+    mutt_fix_reply_recipients (msg->env);
+
     if (! (flags & SENDMAILX) &&
 	! (option (OPTAUTOEDIT) && option (OPTEDITHDRS)) &&
 	! ((flags & SENDREPLY) && option (OPTFASTREPLY)))
@@ -1149,9 +1193,30 @@ ci_send_message (int flags,		/* send mode */
       killfrom = 1;
     }
 
+    if ((flags & SENDREPLY) && cur)
+    {
+      /* change setting based upon message we are replying to */
+      mutt_message_hook (ctx, cur, M_REPLYHOOK);
+
+      /*
+       * set the replied flag for the message we are generating so that the
+       * user can use ~Q in a send-hook to know when reply-hook's are also
+       * being used.
+       */
+      msg->replied = 1;
+    }
+
     /* change settings based upon recipients */
     
-    mutt_send_hook (msg);
+    mutt_message_hook (NULL, msg, M_SENDHOOK);
+
+    /*
+     * Unset the replied flag from the message we are composing since it is
+     * no longer required.  This is done here because the FCC'd copy of
+     * this message was erroneously get the 'R'eplied flag when stored in
+     * a maildir-style mailbox.
+     */
+    msg->replied = 0;
 
     if (killfrom)
     {
@@ -1163,61 +1228,85 @@ ci_send_message (int flags,		/* send mode */
       process_user_header (msg->env);
 
 
-
-#ifdef HAVE_PGP
-    if (! (flags & SENDMAILX))
-    {
-      if (option (OPTPGPAUTOSIGN))
-	msg->pgp |= PGPSIGN;
-      if (option (OPTPGPAUTOENCRYPT))
-	msg->pgp |= PGPENCRYPT;
-      if (option (OPTPGPREPLYENCRYPT) && cur && cur->pgp & PGPENCRYPT)
-	msg->pgp |= PGPENCRYPT;
-      if (option (OPTPGPREPLYSIGN) && cur && cur->pgp & PGPSIGN)
-	msg->pgp |= PGPSIGN;
-      if (option (OPTPGPREPLYSIGNENCRYPTED) && cur && cur->pgp & PGPENCRYPT)
-	msg->pgp |= PGPSIGN;
-    }
-#endif /* HAVE_PGP */
-
-
+    if (option (OPTSIGONTOP) && (! (flags & (SENDMAILX | SENDKEY)) && Editor && mutt_strcmp (Editor, "builtin") != 0))
+      append_signature (tempfp);
 
     /* include replies/forwarded messages, unless we are given a template */
     if (!tempfile && (ctx || !(flags & (SENDREPLY|SENDFORWARD)))
 	&& generate_body (tempfp, msg, flags, ctx, cur) == -1)
       goto cleanup;
 
-    if (! (flags & (SENDMAILX | SENDKEY)) && Editor && mutt_strcmp (Editor, "builtin") != 0)
+    if (!option (OPTSIGONTOP) && (! (flags & (SENDMAILX | SENDKEY)) && Editor && mutt_strcmp (Editor, "builtin") != 0))
       append_signature (tempfp);
+
+    /* 
+     * this wants to be done _after_ generate_body, so message-hooks
+     * can take effect.
+     */
+
+    if (WithCrypto && !(flags & SENDMAILX))
+    {
+      if (option (OPTCRYPTAUTOSIGN))
+	msg->security |= SIGN;
+      if (option (OPTCRYPTAUTOENCRYPT))
+	msg->security |= ENCRYPT;
+      if (option (OPTCRYPTREPLYENCRYPT) && cur && (cur->security & ENCRYPT))
+	msg->security |= ENCRYPT;
+      if (option (OPTCRYPTREPLYSIGN) && cur && (cur->security & SIGN))
+	msg->security |= SIGN;
+      if (option (OPTCRYPTREPLYSIGNENCRYPTED) && cur && (cur->security & ENCRYPT))
+	msg->security |= SIGN;
+    }      
+
+    if (WithCrypto && msg->security)
+    {
+      /* 
+       * When reypling / forwarding, use the original message's
+       * crypto system.  According to the documentation,
+       * smime_is_default should be disregarded here.
+       * 
+       * Problem: At least with forwarding, this doesn't really
+       * make much sense. Should we have an option to completely
+       * disable individual mechanisms at run-time?
+       */
+      if (cur)
+      {
+	if ((WithCrypto & APPLICATION_PGP) && option (OPTCRYPTAUTOPGP) 
+	    && (cur->security & APPLICATION_PGP))
+	  msg->security |= APPLICATION_PGP;
+	else if ((WithCrypto & APPLICATION_SMIME) && option (OPTCRYPTAUTOSMIME)
+		 && (cur->security & APPLICATION_SMIME))
+	  msg->security |= APPLICATION_SMIME;
+      }
+      
+      /*
+       * No crypto mechanism selected? Use availability + smime_is_default
+       * for the decision. 
+       */
+      if (!(msg->security & (APPLICATION_SMIME | APPLICATION_PGP)))
+      {
+	if ((WithCrypto & APPLICATION_SMIME) && option (OPTCRYPTAUTOSMIME) 
+	    && option (OPTSMIMEISDEFAULT))
+	  msg->security |= APPLICATION_SMIME;
+	else if ((WithCrypto & APPLICATION_PGP) && option (OPTCRYPTAUTOPGP))
+	  msg->security |= APPLICATION_PGP;
+	else if ((WithCrypto & APPLICATION_SMIME) && option (OPTCRYPTAUTOSMIME))
+	  msg->security |= APPLICATION_SMIME;
+      }
+    }
+    
+    /* No permissible mechanisms found.  Don't sign or encrypt. */
+    if (!(msg->security & (APPLICATION_SMIME|APPLICATION_PGP)))
+      msg->security = 0;
   }
   /* wait until now to set the real name portion of our return address so
      that $realname can be set in a send-hook */
-  if (msg->env->from && !msg->env->from->personal && !(flags & (SENDRESEND|SENDPOSTPONED)))
+  if (msg->env->from && !msg->env->from->personal
+      && !(flags & (SENDRESEND|SENDPOSTPONED)))
     msg->env->from->personal = safe_strdup (Realname);
 
-
-
-#ifdef HAVE_PGP
-  
-  if (! (flags & SENDKEY) && tempfp)
-  {
-
-#endif
-    
-
-
-    fclose (tempfp);
-    tempfp = NULL;
-    
-
-
-#ifdef HAVE_PGP
-    
-  }
-
-#endif
-
-
+  if (!((WithCrypto & APPLICATION_PGP) && (flags & SENDKEY)))
+    safe_fclose (&tempfp);
 
   if (flags & SENDMAILX)
   {
@@ -1227,22 +1316,36 @@ ci_send_message (int flags,		/* send mode */
   else if (! (flags & SENDBATCH))
   {
     struct stat st;
-    time_t mtime;
-
-    stat (msg->content->filename, &st);
-    mtime = st.st_mtime;
-
+    time_t mtime = mutt_decrease_mtime (msg->content->filename, NULL);
+    
     mutt_update_encoding (msg->content);
 
-    /* If the this isn't a text message, look for a mailcap edit command */
-    if(! (flags & SENDKEY))
+    /*
+     * Select whether or not the user's editor should be called now.  We
+     * don't want to do this when:
+     * 1) we are sending a key/cert
+     * 2) we are forwarding a message and the user doesn't want to edit it.
+     *    This is controled by the quadoption $forward_edit.  However, if
+     *    both $edit_headers and $autoedit are set, we want to ignore the
+     *    setting of $forward_edit because the user probably needs to add the
+     *    recipients.
+     */
+    if (! (flags & SENDKEY) &&
+	((flags & SENDFORWARD) == 0 ||
+	 (option (OPTEDITHDRS) && option (OPTAUTOEDIT)) ||
+	 query_quadoption (OPT_FORWEDIT, _("Edit forwarded message?")) == M_YES))
     {
+      /* If the this isn't a text message, look for a mailcap edit command */
       if (mutt_needs_mailcap (msg->content))
 	mutt_edit_attachment (msg->content);
       else if (!Editor || mutt_strcmp ("builtin", Editor) == 0)
 	mutt_builtin_editor (msg->content->filename, msg, cur);
       else if (option (OPTEDITHDRS))
+      {
+	mutt_env_to_local (msg->env);
 	mutt_edit_headers (Editor, msg->content->filename, msg, fcc, sizeof (fcc));
+	mutt_env_to_idna (msg->env, NULL, NULL);
+      }
       else
 	mutt_edit_file (Editor, msg->content->filename);
     }
@@ -1252,7 +1355,7 @@ ci_send_message (int flags,		/* send mode */
       if (stat (msg->content->filename, &st) == 0)
       {
 	/* if the file was not modified, bail out now */
-	if (mtime == st.st_mtime &&
+	if (mtime == st.st_mtime && !msg->content->next &&
 	    query_quadoption (OPT_ABORT, _("Abort unmodified message?")) == M_YES)
 	{
 	  mutt_message _("Aborted unmodified message.");
@@ -1267,7 +1370,7 @@ ci_send_message (int flags,		/* send mode */
   /* specify a default fcc.  if we are in batchmode, only save a copy of
    * the message if the value of $copy is yes or ask-yes */
 
-  if (!fcc[0] && !(flags & (SENDRESEND|SENDPOSTPONED)) && (!(flags & SENDBATCH) || (quadoption (OPT_COPY) & 0x1)))
+  if (!fcc[0] && !(flags & (SENDPOSTPONED)) && (!(flags & SENDBATCH) || (quadoption (OPT_COPY) & 0x1)))
   {
     /* set the default FCC */
     if (!msg->env->from)
@@ -1291,6 +1394,7 @@ ci_send_message (int flags,		/* send mode */
   {
 main_loop:
 
+    fcc_error = 0; /* reset value since we may have failed before */
     mutt_pretty_mailbox (fcc);
     i = mutt_compose_menu (msg, fcc, sizeof (fcc), cur);
     if (i == -1)
@@ -1313,6 +1417,7 @@ main_loop:
 
       encode_descriptions (msg->content, 1);
       mutt_prepare_envelope (msg->env, 0);
+      mutt_env_to_idna (msg->env, NULL, NULL);	/* Handle bad IDNAs the next time. */
 
       if (!Postponed || mutt_write_fcc (NONULL (Postponed), msg, (cur && (flags & SENDREPLY)) ? cur->env->message_id : NULL, 1, fcc) < 0)
       {
@@ -1341,6 +1446,16 @@ main_loop:
     }
   }
 
+  if (mutt_env_to_idna (msg->env, &tag, &err))
+  {
+    mutt_error (_("Bad IDN in \"%s\": '%s'"), tag, err);
+    FREE (&err);
+    if (!(flags & SENDBATCH))
+      goto main_loop;
+    else 
+      goto cleanup;
+  }
+  
   if (!msg->env->subject && ! (flags & SENDBATCH) &&
       (i = query_quadoption (OPT_SUBJECT, _("No subject, abort sending?"))) != M_NO)
   {
@@ -1361,36 +1476,51 @@ main_loop:
 
   encode_descriptions (msg->content, 1);
   
-#ifdef HAVE_PGP
-  if (msg->pgp)
-  {
-    /* save the decrypted attachments */
-    clear_content = msg->content;
-
-    if ((pgp_get_keys (msg, &pgpkeylist) == -1) ||
-	(pgp_protect (msg, pgpkeylist) == -1))
-    {
-      msg->content = mutt_remove_multipart (msg->content);
-      
-      if (pgpkeylist)
-	FREE (&pgpkeylist);
-      
-      decode_descriptions (msg->content);
-      goto main_loop;
-    }
-    encode_descriptions (msg->content, 0);
-  }
-
-  /* 
-   * at this point, msg->content is one of the following three things:
-   * - multipart/signed.  In this case, clear_content is a child.
-   * - multipart/encrypted.  In this case, clear_content exists
-   *   independently
-   * - application/pgp.  In this case, clear_content exists independently.
-   * - something else.  In this case, it's the same as clear_content.
+  /*
+   * Make sure that clear_content and free_clear_content are
+   * properly initialized -- we may visit this particular place in
+   * the code multiple times, including after a failed call to
+   * mutt_protect().
    */
-
-#endif /* HAVE_PGP */
+  
+  clear_content = NULL;
+  free_clear_content = 0;
+  
+  if (WithCrypto)
+  {
+    if (msg->security)  
+    {
+      /* save the decrypted attachments */
+      clear_content = msg->content;
+  
+      if ((crypt_get_keys (msg, &pgpkeylist) == -1) ||
+          mutt_protect (msg, cur, pgpkeylist) == -1)
+      {
+        msg->content = mutt_remove_multipart (msg->content);
+        
+	FREE (&pgpkeylist);
+        
+        decode_descriptions (msg->content);
+        goto main_loop;
+      }
+      encode_descriptions (msg->content, 0);
+    }
+  
+    /* 
+     * at this point, msg->content is one of the following three things:
+     * - multipart/signed.  In this case, clear_content is a child.
+     * - multipart/encrypted.  In this case, clear_content exists
+     *   independently
+     * - application/pgp.  In this case, clear_content exists independently.
+     * - something else.  In this case, it's the same as clear_content.
+     */
+  
+    /* This is ugly -- lack of "reporting back" from mutt_protect(). */
+    
+    if (clear_content && (msg->content != clear_content)
+        && (msg->content->parts != clear_content))
+      free_clear_content = 1;
+  }
 
   if (!option (OPTNOCURSES) && !(flags & SENDMAILX))
     mutt_message _("Sending message...");
@@ -1419,26 +1549,22 @@ main_loop:
   if (*fcc && mutt_strcmp ("/dev/null", fcc) != 0)
   {
     BODY *tmpbody = msg->content;
-#ifdef HAVE_PGP
     BODY *save_sig = NULL;
     BODY *save_parts = NULL;
-#endif /* HAVE_PGP */
 
-#ifdef HAVE_PGP
-    if (msg->pgp && option (OPTFCCCLEAR))
+    if (WithCrypto && msg->security && option (OPTFCCCLEAR))
       msg->content = clear_content;
-#endif
 
     /* check to see if the user wants copies of all attachments */
     if (!option (OPTFCCATTACH) && msg->content->type == TYPEMULTIPART)
     {
-#ifdef HAVE_PGP
-      if (mutt_strcmp (msg->content->subtype, "encrypted") == 0 ||
-	  mutt_strcmp (msg->content->subtype, "signed") == 0)
+      if (WithCrypto
+          && (mutt_strcmp (msg->content->subtype, "encrypted") == 0 ||
+              mutt_strcmp (msg->content->subtype, "signed") == 0))
       {
 	if (clear_content->type == TYPEMULTIPART)
 	{
-	  if (!(msg->pgp & PGPENCRYPT) && (msg->pgp & PGPSIGN))
+	  if(!(msg->security & ENCRYPT) && (msg->security & SIGN))
 	  {
 	    /* save initial signature and attachments */
 	    save_sig = msg->content->parts->next;
@@ -1448,7 +1574,7 @@ main_loop:
 	  /* this means writing only the main part */
 	  msg->content = clear_content->parts;
 
-	  if (pgp_protect (msg, pgpkeylist) == -1)
+	  if (mutt_protect (msg, cur, pgpkeylist) == -1)
 	  {
 	    /* we can't do much about it at this point, so
 	     * fallback to saving the whole thing to fcc
@@ -1462,57 +1588,75 @@ main_loop:
 	}
       }
       else
-#endif /* HAVE_PGP */
 	msg->content = msg->content->parts;
     }
 
-#ifdef HAVE_PGP
 full_fcc:
-#endif /* HAVE_PGP */
     if (msg->content)
-      mutt_write_fcc (fcc, msg, NULL, 0, NULL);
+    {
+      /* update received time so that when storing to a mbox-style folder
+       * the From_ line contains the current time instead of when the
+       * message was first postponed.
+       */
+      msg->received = time (NULL);
+      if (mutt_write_fcc (fcc, msg, NULL, 0, NULL) == -1)
+      {
+	/*
+	 * Error writing FCC, we should abort sending.
+	 */
+	fcc_error = 1;
+      }
+    }
 
     msg->content = tmpbody;
 
-#ifdef HAVE_PGP
-    if (save_sig)
+    if (WithCrypto && save_sig)
     {
       /* cleanup the second signature structures */
-      mutt_free_body (&save_content->parts->next);
-      save_content->parts = NULL;
+      if (save_content->parts)
+      {
+	mutt_free_body (&save_content->parts->next);
+	save_content->parts = NULL;
+      }
       mutt_free_body (&save_content);
 
       /* restore old signature and attachments */
       msg->content->parts->next = save_sig;
       msg->content->parts->parts->next = save_parts;
     }
-    else if (save_content)
+    else if (WithCrypto && save_content)
     {
       /* destroy the new encrypted body. */
       mutt_free_body (&save_content);
     }
-      
-#endif /* HAVE_PGP */
+
   }
 
 
-  if ((i = send_message (msg)) == -1)
+  /*
+   * Don't attempt to send the message if the FCC failed.  Just pretend
+   * the send failed as well so we give the user a chance to fix the
+   * error.
+   */
+  if (fcc_error || (i = send_message (msg)) == -1)
   {
     if (!(flags & SENDBATCH))
     {
-#ifdef HAVE_PGP
-      if ((msg->pgp & PGPENCRYPT) || 
-	  ((msg->pgp & PGPSIGN) && msg->content->type == TYPEAPPLICATION))
+      if (!WithCrypto)
+        ;
+      else if ((msg->security & ENCRYPT) || 
+               ((msg->security & SIGN)
+                && msg->content->type == TYPEAPPLICATION))
       {
 	mutt_free_body (&msg->content); /* destroy PGP data */
 	msg->content = clear_content;	/* restore clear text. */
       }
-      else if ((msg->pgp & PGPSIGN) && msg->content->type == TYPEMULTIPART)
+      else if ((msg->security & SIGN) && msg->content->type == TYPEMULTIPART)
       {
-	mutt_free_body (&msg->content->parts->next);		/* destroy sig */
-	msg->content = mutt_remove_multipart (msg->content);	/* remove multipart */
+	mutt_free_body (&msg->content->parts->next);	     /* destroy sig */
+	msg->content = mutt_remove_multipart (msg->content); 
       }
-#endif
+
       msg->content = mutt_remove_multipart (msg->content);
       decode_descriptions (msg->content);
       mutt_unprepare_envelope (msg->env);
@@ -1527,14 +1671,11 @@ full_fcc:
   else if (!option (OPTNOCURSES) && ! (flags & SENDMAILX))
     mutt_message (i == 0 ? _("Mail sent.") : _("Sending in background."));
 
-#ifdef HAVE_PGP
-  if (msg->pgp & PGPENCRYPT)
-  {
-    /* cleanup structures from the first encryption */
-    mutt_free_body (&clear_content);
+  if (WithCrypto && (msg->security & ENCRYPT))
     FREE (&pgpkeylist);
-  }
-#endif /* HAVE_PGP */
+  
+  if (WithCrypto && free_clear_content)
+    mutt_free_body (&clear_content);
 
   if (flags & SENDREPLY)
   {
@@ -1553,30 +1694,19 @@ full_fcc:
   
 cleanup:
 
-
-
-#ifdef HAVE_PGP
-  if (flags & SENDPOSTPONED)
+  if ((WithCrypto & APPLICATION_PGP) && (flags & SENDPOSTPONED))
   {
-    
     if(signas)
     {
-      safe_free((void **) &PgpSignAs);
+      FREE (&PgpSignAs);
       PgpSignAs = signas;
     }
-    
-    if(signmic)
-    {
-      safe_free((void **) &PgpSignMicalg);
-      PgpSignMicalg = signmic;
-    }
   }
-#endif /* HAVE_PGP */
    
-  if (tempfp)
-    fclose (tempfp);
+  safe_fclose (&tempfp);
   mutt_free_header (&msg);
   
   return rv;
-
 }
+
+/* vim: set sw=2: */
