@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 1996-2000 Michael R. Elkins <me@cs.hmc.edu>
+ * Copyright (C) 1996-2000 Michael R. Elkins <me@mutt.org>
  * 
  *     This program is free software; you can redistribute it and/or modify
  *     it under the terms of the GNU General Public License as published by
@@ -40,9 +40,13 @@ RETSIGTYPE exit_handler (int sig)
 #else
 #if (__sun__ && __svr4__)
   printf(_("Caught %s...  Exiting.\n"), _sys_siglist[sig]);
+#else
+#if (__alpha && __osf__)
+  printf(_("Caught %s...  Exiting.\n"), __sys_siglist[sig]);
 #else
   printf(_("Caught signal %d...  Exiting.\n"), sig);
 #endif
+#endif
 #endif
   exit (0);
 }
@@ -111,12 +115,19 @@ void mutt_signal_init (void)
   /* we want to avoid race conditions */
   sigaddset (&act.sa_mask, SIGTSTP);
 
+  act.sa_handler = sighandler;
+
+  /* we want SIGALRM to abort the current syscall, so we do this before
+   * setting the SA_RESTART flag below.  currently this is only used to
+   * timeout on a connect() call in a reasonable amout of time.
+   */
+  sigaction (SIGALRM, &act, NULL);
+
   /* we also don't want to mess with interrupted system calls */
 #ifdef SA_RESTART
   act.sa_flags = SA_RESTART;
 #endif
 
-  act.sa_handler = sighandler;
   sigaction (SIGCONT, &act, NULL);
   sigaction (SIGTSTP, &act, NULL);
   sigaction (SIGINT, &act, NULL);
@@ -218,3 +229,16 @@ void mutt_unblock_signals_system (int catch)
     unset_option (OPTSYSSIGNALSBLOCKED);
   }
 }
+
+void mutt_allow_interrupt (int disposition)
+{
+  struct sigaction sa;
+  
+  memset (&sa, 0, sizeof sa);
+  sa.sa_handler = sighandler;
+#ifdef SA_RESTART
+  if (disposition == 0)
+    sa.sa_flags |= SA_RESTART;
+#endif
+  sigaction (SIGINT, &sa, NULL);
+}
