@@ -29,11 +29,11 @@
  *           gcc -DTEST_SNPRINTF -o snprintf snprintf.c -lm
  *    and run snprintf for results.
  * 
- *  Thomas Roessler <roessler@guug.de> 01/27/98 for mutt 0.89i
+ *  Thomas Roessler <roessler@does-not-exist.org> 01/27/98 for mutt 0.89i
  *    The PGP code was using unsigned hexadecimal formats. 
  *    Unfortunately, unsigned formats simply didn't work.
  *
- *  Michael Elkins <me@cs.hmc.edu> 03/05/98 for mutt 0.90.8
+ *  Michael Elkins <me@mutt.org> 03/05/98 for mutt 0.90.8
  *    The original code assumed that both snprintf() and vsnprintf() were
  *    missing.  Some systems only have snprintf() but not vsnprintf(), so
  *    the code is now broken down under HAVE_SNPRINTF and HAVE_VSNPRINTF.
@@ -44,7 +44,8 @@
 
 #if !defined(HAVE_SNPRINTF) || !defined(HAVE_VSNPRINTF)
 
-#include <ctype.h>
+#include <string.h>
+# include <ctype.h>
 #include <sys/types.h>
 
 /* Define this as a fall through, HAVE_STDARG_H is probably already set */
@@ -115,6 +116,7 @@ static void dopr_outch (char *buffer, size_t *currlen, size_t maxlen, char c );
 #define DP_C_LDOUBLE 3
 
 #define char_to_int(p) (p - '0')
+#undef MAX
 #define MAX(p,q) ((p >= q) ? p : q)
 
 static void dopr (char *buffer, size_t maxlen, const char *format, va_list args)
@@ -178,7 +180,7 @@ static void dopr (char *buffer, size_t maxlen, const char *format, va_list args)
       }
       break;
     case DP_S_MIN:
-      if (isdigit(ch)) 
+      if (isdigit((unsigned char)ch)) 
       {
 	min = 10*min + char_to_int (ch);
 	ch = *format++;
@@ -202,7 +204,7 @@ static void dopr (char *buffer, size_t maxlen, const char *format, va_list args)
 	state = DP_S_MOD;
       break;
     case DP_S_MAX:
-      if (isdigit(ch)) 
+      if (isdigit((unsigned char)ch)) 
       {
 	if (max < 0)
 	  max = 0;
@@ -760,7 +762,7 @@ int main (void)
       sprintf (buf2, fp_fmt[x], fp_nums[y]);
       if (strcmp (buf1, buf2))
       {
-	printf("snprintf doesn't match Format: %s\n\tsnprintf = %s\n\tsprintf  = %s\n", 
+	printf("snprintf doesn't match Format: %s\n\tsnprintf = %s\n\tsprintf  = %s\n",	/* __SPRINTF_CHECKED__ */
 	    fp_fmt[x], buf1, buf2);
 	fail++;
       }
@@ -774,7 +776,7 @@ int main (void)
       sprintf (buf2, int_fmt[x], int_nums[y]);
       if (strcmp (buf1, buf2))
       {
-	printf("snprintf doesn't match Format: %s\n\tsnprintf = %s\n\tsprintf  = %s\n", 
+	printf("snprintf doesn't match Format: %s\n\tsnprintf = %s\n\tsprintf  = %s\n",	/* __SPRINTF_CHECKED__ */
 	    int_fmt[x], buf1, buf2);
 	fail++;
       }
